/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Vector;
import java.awt.Point;

/** A CompoundVertex represents a subgraph which we have chosen to see
    as a unit. */

public class CompoundVertex extends MusicVertex {

    /** The {@link Subgraph} contained in this CompoundVertex. */
    public Subgraph subgraph=null;

    /** The graph production rule which arises from the substitution
        of the subgraph with this CompoundVertex; in graph-grammar
        terms, this CompoundVertex may be said to *produce* its inner
        Subgraph. */
    public GraphProductionRule productionRule=null;
    
    /** The similarity measure result which describes the quality of
        the {@link SimilarityStatement} in which the inner Subgraph of
        this CompoundVertex was found to be similar to something
        else. This is used in the {@link SimSegmenter}. From the
        {@link SimMeasureResult}, we can know which {@link
        SimilarityMeasure} was used. */
    public SimMeasureResult smr=null;

    /* Please do remember to change this length to -1 if the inner
    subgraph is changed; then it is automatically recalculated.*/
    private double length=-1.0;

    /* Constructor. Compound vertices are only created in the
       MotherGraph.compoundSubstitution() method. */
    public CompoundVertex(Subgraph s, 
			  double start, 
			  long time, 
			  GraphProductionRule gprule,
			  SimMeasureResult smresult) {
	inEdges = new Vector();
	outEdges = new Vector();
	subgraph=s;
	starttime=start;
	objectTimeStamp=time;
	productionRule=gprule;
	smr = smresult;
    }

    /** The length of this CompoundVertex is the largest end-time of
    MusicVertices (be they simple or compound) inside the inner
    Subgraph of this CompoundVertex minus the smallest start-time
    inside the inner subgraph. */
    public double getLength() {
	if (length== -1.0) {
	    length= subgraph.getMaxEndtime()-subgraph.getMinStarttime();
	}
	return length;
    }

    /** Returns a CompoundVertex whose inner subgraph has been
        deepCloned. The Vectors of in- and out-edges are not cloned,
        they need to be set from the outside, since deepCloning is
        intended to create a new CompoundVertex used in a new music
        graph, so the new CompoundVertex should point to edges of this
        new graph instead of those from the original music graph. This
        is a tricky part of the MusicGraph.deepClone() method.*/
    public MusicVertex deepClone() {
	CompoundVertex cv = new CompoundVertex((Subgraph)subgraph.deepClone().clone(),
					       this.starttime,
					       FuncLib.getTimeStamp(),
					       productionRule /* deepClone pr?*/,
					       smr /*deepClone smr? */);
	/* inEdges and outEdges cannot be set here, do it from outside */
	return cv;
    }

    /** Writes a one-line description of this CompoundVertex. */
    public String toString() {
	String result = "CV of "+productionRule.nt.toString()+", start="+starttime
	    +", end="+(starttime+getLength());
	return result;
    }
    
    /** Writes a longer description of this CompoundVertex, including
        also a toString of the inner subgraph.  */
    public String toStringWithSubgraph() {
	String result = "CV of "+productionRule.nt.toString()+", start="+starttime
	    +", end="+(starttime+getLength())+"\n";
	result+= subgraph.toString();
	return result;
    }
    
    /** Anchor resolution on a CompoundVertex recursively finds the
        SimpleVertex determined by the Vector of Integers that is
        input as a parameter. */
    public SimpleVertex resolveAnchor(Vector anchor /*of Integer */) {
	if (anchor.size()==0) {
	    System.out.println("resolveAnchor() called with empty anchor vector in "+toString());
	    return null;
	}

	subgraph.sortVerticesByStarttime();
	/* We need to remove an index from the anchor vector every time we peel
	 * away an onion layer of Compounds, therefore clone it, so that we don't
	 * change the original anchor that is owned by the edge. */
	Vector anch = (Vector) anchor.clone();

	int index = ((Integer)anch.get(0)).intValue();
	MusicVertex mv = (MusicVertex) subgraph.vertices.get(index);
	anch.remove(0);
	return mv.resolveAnchor(anch);
    }

    /** Returns the index of MusicVertex mv if it is contained in the
        inner subgraph of this CompoundVertex. */
    public int getIndexOf(MusicVertex mv) {
	subgraph.sortVerticesByStarttime();
	return subgraph.vertices.indexOf(mv);
    }
}

