/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.*;
import java.io.*;

/** This class is a program used to give an impression of the fitness
landscape for sequential similarity measures. It reads a midi or
MuseData file into a graph, constructs a number of sequential
subgraphs, and slides them to the right, registering after each slide
the fitness of a similarity statement with the sliding sequential
subgraph in question as one half and another fixed, randomly chosen
sequential subgraph as the other half. The registered fitness values
are written to an output file in a format ready to use in GnuPlot. */

public class FitnessLandscape {

    /** The sequential similarity measure used to evaluate the
     sequential similarity statements in
     writeNeighbourFitnessToFile(). You should change the value of
     simMeasure if you want to use another similarity measure.  */
    public static SeqSimMeasure simMeasure=SimLibDiffCount.seqDFIAndLength_Grp;

    /** Effects the main call to writeNeighbourFitnessToFile().
	Accepts command line parameters input file, output file,
	subgraphs size, and number of subgraphs. */
    public static void main(String[] args) {

	if (args.length<4) {
	    System.out.println("Usage: java FitnessLandscape <filename> <output filename> <subgraph size> <num of subgraphs>");
	    System.exit(0);
	}
	MotherGraph piece=null;
	try {
	    piece = MotherGraph.getMomFromFile(args[0],true,true,false);
	} catch(Exception e) {
	    System.out.println("FitnessLandscape: problem loading mothergraph from file "+args[0]);
	    System.exit(0);
	}
	int size = new Integer(args[2]).intValue();
	int numOfSubgraphs = new Integer(args[3]).intValue();

	// Vector subs = piece.getNSubgraphsOfSize(numOfSubgraphs,size);
	// writeAllDistancesToFile(subs,args[1],size);
	writeNeighbourFitnessToFile(piece,args[1],size);
    }

    /** Creates sequential subgraphs of size "size", slides them
through the graph representing the piece and writes the registered
fitness values to an output file. This is where the main action in
class FitnessLandscape happens. */
    public static void writeNeighbourFitnessToFile(MotherGraph piece,
						   String outfilename,
						   int size) {
	try {
	    Vector subgraphs = piece.getBeginningSubgraphSet(size);

	    Vector allThreads = new Vector();
	    for (int i=0; i<subgraphs.size(); i++) {
		allThreads.add(new Vector());
	    }

	    // counterparts is a vector of SeqSubgraph
	    Vector counterparts=new Vector(); // the other halves of 
	                                      // the SimStmts
	    for (int i=0; i<subgraphs.size(); i++) {
		counterparts.add(piece.getRandomSeqSubgraph(size));
	    }

	    FileWriter outfile = new FileWriter(outfilename);
	    outfile.write("# Outfile from FitnessLandscape.writeNeighbourFitnessToFile, created "+new Date().toString()+"\n"); 
	    outfile.write("# Shows distance changes from sliding size "+size+" subgraphs in "+allThreads.size()+" threads\n");

	    SeqSubgraph s1=piece.getRandomSeqSubgraph(2);
	    SeqSubgraph s2=piece.getRandomSeqSubgraph(2);
	    /* OBS: ingen SimFinder tilknyttet simstmt*/
	    SeqSimStmt dummy = new SeqSimStmt(s1,s2,simMeasure,null);
	    outfile.write("# slideNum "+dummy.gnuplotHeader()+"\n"); 

	    boolean[] done = new boolean[subgraphs.size()];
	    for (int i=0; i<done.length; i++) done[i]=false;

	    SeqSubgraph[] subs = new SeqSubgraph[subgraphs.size()];
	    for (int i=0; i<subgraphs.size(); i++) {
		subs[i]=(SeqSubgraph)subgraphs.get(i);
	    }
	    int generation=0;
	    while (numOfFalse(done)>0) {
		for (int i=0; i<subs.length; i++) {
		    if (!done[i] && subs[i]!=null) {
			SeqSubgraph sub = subs[i];
			SeqSubgraph subClone = (SeqSubgraph) sub.clone();
			try {
			    SeqSubgraph other = (SeqSubgraph) counterparts.get(i);
			    /* OBS: ingen SimFinder tilknyttet simstmt*/
			    SimilarityStatement ss = new SeqSimStmt(sub,other,simMeasure,null);
			    String line = ""+generation+" "+ss.toGnuplotString();
			    Vector thread = (Vector) allThreads.get(i);
			    thread.add(line);

			    subClone.slideRight(1);
			    subs[i]=subClone;
			} catch(Exception e) {
			    done[i]=true;
			    subs[i]=null;
			}
		    }
		}
		generation++;
		System.out.print(".");
	    }
	    for (int i=0; i<allThreads.size(); i++) {
		Vector thread = (Vector) allThreads.get(i);
		outfile.write("\n#Thread "+i+"\n");
		for (int j=0; j<thread.size(); j++) {
		    String str = (String) thread.get(j);
		    outfile.write(str+"\n");
		}
		outfile.write("\n\n");
	    }
	    outfile.flush();
	    outfile.close();
	} catch(Exception e) {
	    System.out.println(e.getMessage());
	}
    }

    /** Counts and returns the number of false entries in the input
     boolean array.  */
    public static int numOfFalse(boolean[] done) {
	int result =0;
	for (int i=0; i<done.length; i++) {
	    if (!done[i]) result++;
	}
	return result;
    }

    /**
       @deprecated This method isn't used any more. Use
       writeNeighbourFitnessToFile instead.  */
    public static void writeAllDistancesToFile(Vector subs 
					       /* of SeqSubgraph */, 
					       String outfilename,
					       int size) {
	int totalTodo=(int)Math.pow(subs.size(),2);
	int onePercent = totalTodo/100;
	System.out.println("onePercent="+onePercent);
	try {
	    FileWriter outfile = new FileWriter(outfilename);
	    outfile.write("# Outfile from FitnessLandscape.java, created "+new Date().toString()+"\n"); 
	    outfile.write("# Shows distance vs. similarities between first "+subs.size()+" SimStmts of size "+size+"\n");

	    SeqSubgraph dseqsub1 = (SeqSubgraph) subs.get(0);
	    SeqSubgraph dseqsub2 = (SeqSubgraph) subs.get(1);
	    /* OBS: ingen SimFinder tilknyttet simstmt*/
	    SimilarityStatement dummy = new SeqSimStmt(dseqsub1,dseqsub2,simMeasure,null);
	    dummy.value();
	    outfile.write("# overallDist value maxStart minStart maxEnd minEnd "+dummy.smResult.toGnuplotHeader()+"\n");
	    



	    for (int i=0; i<subs.size(); i++) {
		SeqSubgraph mg1 = (SeqSubgraph) subs.get(i);
		int[][] dist1 = mg1.getDistanceArrays();

		for (int j=0; j<subs.size(); j++) {
		    if (i!=j) {
			SeqSubgraph mg2 = (SeqSubgraph) subs.get(j);
			int[][] dist2 = mg2.getDistanceArrays();
			double maxstart = euclideanDistance(dist1[0],dist2[0]);
			double minstart = euclideanDistance(dist1[1],dist2[1]);
			double maxend = euclideanDistance(dist1[2],dist2[2]);
			double minend = euclideanDistance(dist1[3],dist2[3]);

			double overallDist = (maxstart+minstart+maxend+minend)/4;
			/* OBS: ingen SimFinder tilknyttet simstmt*/
			SimilarityStatement ss = new SeqSimStmt(mg1,mg2,simMeasure,null);
			String data = ""+overallDist; /* 1 */
			data+=" "+ss.value();         /* 2 */
			data+=" "+maxstart;           /* 3 */
			data+=" "+minstart;           /* 4 */
			data+=" "+maxend;             /* 5 */
			data+=" "+minend;             /* 6 */

			data+=" "+ss.smResult.toGnuplotString()+"\n";
			
			outfile.write(data);
			if (((i*subs.size()+j)% onePercent)==0)
			    System.out.print("#");
		    }
		}
	    }

	    outfile.flush();
	    outfile.close();
	    System.out.println("FitnessLandscape results written to file "+outfilename);
	} catch(Exception e) {
	    System.out.println(e.getMessage());
	}
    }

    /**
       Returns the Euclidean distance between two integer arrays.
     */
    public static double euclideanDistance(int[] v1, int[] v2) {
	if (v1.length !=v2.length) return -1.0;

	double result =0.0;
	for (int i=0; i<v1.length; i++) {
	    result += Math.pow(v1[i]-v2[i],2);
	}
	result = Math.sqrt(result);
	return result;
    }

}
