/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Vector;

/** GraphProductionRule describes a production rule in a graph
    grammar. These production rules have the general form: the
    left-hand side is a subgraph composed of a {@link CompoundVertex}
    and a number of other surrounding music vertices that are
    connected to the compound vertex. The right-hand side is a
    subgraph composed of these same surrounding vertices now connected
    not to the compound vertex but to the inner subgraph that was
    contained in the compound. */

public class GraphProductionRule {

    /** A counter used for the labeling of rules. */
    public static int globalRuleCounter=0;

    /** The number labeling this GraphProductionRule. */
    public int ruleNumber=0;

    /** The subgraph representing the left-hand side of the production rule. */
    Subgraph leftside;
    /** The subgraph representing the left-hand side of the production rule. */
    Subgraph rightside;

    /** The non-terminal which is associated with this production rule. */
    Nonterminal nt;

    /** The similarity measure according to which the inner subgraph
        of the compound vertex contained in the left-hand side of this
        GraphProductionRule was found to be similar to some other
        subgraph. */
    SimilarityMeasure measure;

    /** If the motif described in this GraphProductionRule is a
        derivation of motif, this variable points to the
        GraphProductionRule describing the original motif. */
    GraphProductionRule original=null;

    /** If the motif described in this GraphProductionRule is an
        original motif, this variable points to a Vector of
        GraphProductionRules describing derivations of this original
        motif. */
    Vector derivations = new Vector(); // of GraphProductionRules


    /** The full constructor. */
    public GraphProductionRule(Subgraph left, Subgraph right, 
			       Nonterminal nterm, SimilarityMeasure meas) {
	leftside=left;
	rightside=right;
	nt=nterm;
	measure=meas;
	globalRuleCounter++;
	ruleNumber=globalRuleCounter;
    }
    /** Constructs a GraphProductionRule without specifying the left
        and right subgraphs. */
    public GraphProductionRule(Nonterminal nterm, SimilarityMeasure meas) {
	leftside=null;
	rightside=null;
	nt=nterm;
	measure=meas;
	globalRuleCounter++;
	ruleNumber=globalRuleCounter;
    }

    public String toString() {
	String result="PR"+ruleNumber+" "+nt.name+" "+measure.name;
	if (original==null) {
	    result+=" Orig("+derivations.size()+")";
	}
	else {
	    result+=" Deriv(PR"+original.ruleNumber+")";
	}
	return result;
    }
    
}
