/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */


/** This class implements the MAXPARRING algorithm studied in the
      traditional compulsory first year dADS assignment.  In the
      incidence matrix, 0 means no edge, 1 means edge from left to
      right and -1 means edge from right to left.  In the color
      arrays, 0 means white, 1 means gray, 2 means black.  */
public class Matching {

    public static void main(String[] args) {
	int[][] inc=new int[4][4];
	inc[0][0]=1;
	inc[0][1]=1;
	inc[0][2]=1;
	inc[1][1]=1;

	inc[2][2]=1;
	inc[2][3]=1;
	inc[3][2]=1;
	inc[3][3]=1;

	int[] match = bipartiteMaxMatch(inc);
	if (match!=null) {
	    for (int i=0; i<match.length; i++) {
		System.out.println("match["+i+"]="+match[i]);
	    }
	} else System.out.println("no matching found");
    }

    private static final boolean DEBUG=false;


    public static final int NOEDGE   =  0;
    public static final int RIGHTEDGE=  1;
    public static final int LEFTEDGE = -1;

    private static final int WHITE=0;
    private static final int GRAY =1;
    private static final int BLACK=2;

    /** Returns an array in which the indices are indices to the left
     * side and the values are indices to the right side of the
     * bipartite graph represented in incidenceMatrix. A value of -1
     * means that the vertex at the index in question has not been
     * matched. */
    public static int[] bipartiteMaxMatch(int[][] incMat) {
	int[][] incidenceMatrix = copyIntMatrix(incMat);

	int[] leftColor = new int[incidenceMatrix.length];
	int[] rightColor = new int[incidenceMatrix[0].length];

	debug("first inc:");
	debug(printinc(incidenceMatrix));
	colorGraph(incidenceMatrix,leftColor,rightColor);

	debug("first graph coloring done:");
	debug(printinc(incidenceMatrix));
	int[] path= findChangePath(incidenceMatrix,leftColor,rightColor);
	while (path!=null) {
	    debug("---mainloop---");
	    changePath(path,incidenceMatrix);
	    debug("changed a path:"+pathtostr(path));
	    debug(printinc(incidenceMatrix));
	    colorGraph(incidenceMatrix,leftColor,rightColor);
	    path=findChangePath(incidenceMatrix,leftColor,rightColor);
	}
	debug("done pathfinding and coloring");
	debug(printinc(incidenceMatrix));

	// the matching is now composed of all LEFTEDGEs.
	int[] result = new int[incidenceMatrix.length];
	for (int i=0; i<result.length; i++) result[i]=-1;
	for (int i=0; i<incidenceMatrix.length; i++) {
	    for (int j=0; j<incidenceMatrix[0].length; j++) {
		if (incidenceMatrix[i][j]==LEFTEDGE) {
		    result[i]=j;
		}
	    }
	}
	debug("match returned:");
	debug(pathtostr(result));

	return result;
    }

    private static void colorGraph(int[][] inc,
				   int[] lcolor, 
				   int[] rcolor) {
	// color all white
	for (int i=0; i<lcolor.length; i++) {
	    lcolor[i]=WHITE;
	}
	for (int i=0; i<rcolor.length; i++) {
	    rcolor[i]=WHITE;
	}

	// color the graph
	boolean change=true;
	while (change) {
	    change = false;
	    // try transition 1
	    for (int i=0; i<lcolor.length; i++) {
		if (lcolor[i]==WHITE && inDegreeLeft(i,inc)==0) {
		    lcolor[i]=BLACK;
		    change = true;
		}
	    }
	    // try transition 2 or 3
	    for (int i=0; i<lcolor.length; i++) {
		if (lcolor[i]==BLACK || lcolor[i]==GRAY) {
		    // does it have an edge to a white vertex?
		    int rightWhiteIndex= getIndexOfWhiteRightSonOf(i,inc,rcolor);
		    if (rightWhiteIndex!=-1) {
			int leftWhiteIndex = getIndexOfWhiteLeftSonOf(rightWhiteIndex,inc,lcolor);
			if (leftWhiteIndex!=-1) {
			    lcolor[leftWhiteIndex]=GRAY;
			    rcolor[rightWhiteIndex]=GRAY;
			    change=true;
			}

		    }
		}
	    }
	    // try transition 4 or 5
	    for (int i=0; i<lcolor.length; i++) {
		if (lcolor[i]==BLACK || lcolor[i]==GRAY) {
		    for (int j=0; j<inc[i].length; j++) {
			if (inc[i][j]==RIGHTEDGE && rcolor[j]==WHITE) {
			    if (outDegreeRight(j,inc)==0) {
				rcolor[j]=BLACK;
				change=true;
			    }
			}
		    }
		}
	    }
	} // end while (graph coloring)

    }
    private static int inDegreeLeft(int index, int[][] inc) {
	int result=0;
	for (int j=0; j<inc[0].length; j++) {
	    if (inc[index][j]==LEFTEDGE) result++;
	}
	return result;
    }
    private static int outDegreeRight(int index, int[][] inc) {
	int result=0;
	for (int i=0; i<inc.length; i++) {
	    if (inc[i][index]==LEFTEDGE) result++;
	}
	return result;
    }
    private static int getIndexOfWhiteRightSonOf(int father,
						 int[][] inc,
						 int[] rcolor) {
	for (int j=0; j<rcolor.length; j++) {
	    if (rcolor[j]==WHITE) {
		if (inc[father][j]==RIGHTEDGE) {
		    return j;
		}
	    }
	}
	return -1;
    }
    private static int getIndexOfWhiteLeftSonOf(int father,
						 int[][] inc,
						 int[] lcolor) {
	for (int i=0; i<lcolor.length; i++) {
	    if (lcolor[i]==WHITE) {
		if (inc[i][father]==LEFTEDGE) {
		    return i;
		}
	    }
	}
	return -1;
    }

    private static int[] findChangePath(int[][] inc, int[] lcolor, 
					int[] rcolor) {
	for (int i=0; i<inc.length; i++) {
	    // try to find a path starting at i
	    if (lcolor[i]==BLACK) {
		boolean[] rtaken = new boolean[rcolor.length];
		boolean[] ltaken = new boolean[lcolor.length];
		for (int j=0; j<rtaken.length; j++) rtaken[j]=false;
		for (int j=0; j<ltaken.length; j++) ltaken[j]=false;
		int[] cp = findCPfrom(i,inc,rtaken,ltaken,lcolor,rcolor);
		if (cp!=null) return cp;
	    }
	}
	return null;
    }
    private static int[] findCPfrom(int index, int[][] inc,
				    boolean[] righttaken, boolean[] lefttaken,
				    int[] lcolor, int[] rcolor) {
	for (int j=0; j<inc[0].length; j++) {
	    if (inc[index][j]==RIGHTEDGE && !righttaken[j]) {
		if (rcolor[j]==BLACK) {
		    int[] result= new int[2];
		    result[0]=index;
		    result[1]=j;
		    return result;
		} else {
		    for (int i=0; i<inc.length; i++) {
			if (inc[i][j]==LEFTEDGE && !lefttaken[i]) {
			    righttaken[j]=true;
			    lefttaken[i]=true;
			    int[] tmpresult=findCPfrom(i,inc,righttaken,lefttaken,
						       lcolor,rcolor);
			    if (tmpresult!=null) {
				int[] result= new int[tmpresult.length+2];
				result[0]=index;
				result[1]=j;
				for (int k=0; k<tmpresult.length; k++) {
				    result[k+2]=tmpresult[k];
				}
				return result;
			    } else {
				righttaken[j]=false;
				lefttaken[i]=false;
			    }
			}
		    }
		}
	    }
	}
	return null;
    }

    private static void changePath(int[] path,int[][] inc) {
	for (int i=0; i<path.length-1; i++) {
	    if (i%2==0) {
		inc[path[i]][path[i+1]]=LEFTEDGE;
	    } else {
		inc[path[i+1]][path[i]]=RIGHTEDGE;
	    }
	}
    }

    private static int[][] copyIntMatrix(int[][] orig) {
	int[][] result = new int[orig.length][orig[0].length];
	for (int i=0; i<orig.length; i++) {
	    for (int j=0; j<orig[0].length; j++) {
		result[i][j]=orig[i][j];
	    }
	}
	return result;
    }

    private static void debug(String s) { 
	if (DEBUG)
	    System.out.println(s);
    }

    private static String printinc(int[][] inc) {
	String result="";
	for (int i=0; i<inc.length; i++) {
	    for (int j=0; j<inc[0].length; j++) {
		result+="["+inc[i][j]+"]";
	    }
	    result+="\n";
	}
	return result;
    }
    private static String pathtostr(int[] path) {
	String result="";
	for (int i=0; i<path.length; i++) {
	    result+="["+path[i]+"]";
	}
	return result;
    }

}
