/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Vector;
import jm.music.data.*;
import jm.JMC;
import jm.util.*;
import jm.midi.*;

/** A MotherGraph is a {@link MusicGraph} which owns its vertices and
edges. {@link Subgraph}s only point to vertices and edges owned by
their respective mother graphs. All subgraphs must have a mother
graph, whereas a mother graph is its own, so to speak. Subgraphs can
be changed in any way you like, but changing a mother graph is a
serious matter because it affects all subgraphs that may be pointing
to its vertices and edges. Proceed with caution. The only such
manipulation is the compoundSubsitution() method. */

public class MotherGraph extends MusicGraph {

    /** The Score object that this MotherGraph originated from. */
    public Score theScore=null; 

    /** A neutral constructor. */
    public MotherGraph(Vector vert /* of MusicVertex */, 
		       Vector edg /* of MusicEdge */,
		       String t,
		       Score sc) {
	vertices=vert;
	edges=edg;
	theScore=sc;
	title=t;
    }

    /** Constructs a MotherGraph from a midifile. */
    public MotherGraph(String filename, boolean includeInterPartEdges) {
	System.out.println("Building mothergraph...");
	long lasttime=0;

	// load file
	Score score=new Score(filename);
	Read.midi(score,filename);
	theScore = score;
	Part[] parts = score.getPartArray();
	title="Graph for file "+filename;
	double rmin= Double.MAX_VALUE;
	double rmax= Double.MIN_VALUE;
	/* Merge all notes in different Parts and Phrases into a vector of 
	 * NoteInfos that is sorted by begintime of the notes. */
	Vector allNoteInfo = new Vector(); // of NoteInfo
	for (int i=0; i<parts.length; i++) {
	    Phrase[] phr = parts[i].getPhraseArray();
	    for (int j=0; j<phr.length; j++) {
		double begintime=phr[j].getStartTime(); // begintime of this phrase
		double deltatime=0.0; // sum of rhythmValue of earlier notes in this phrase
		Note[] notes = phr[j].getNoteArray();
		Vector phraseNoteInfo = new Vector(); // of NoteInfo
		/* to be merged with allNoteInfo */

		for (int k=0; k<notes.length; k++) {
		    Note curnote = notes[k];
		    NoteInfo curNoteInfo= new NoteInfo(parts[i],
						       phr[j],
						       curnote,
						       k,
						       begintime+deltatime);
		    phraseNoteInfo.add(curNoteInfo);
		    deltatime+=curnote.getRhythmValue();
		    // debug:
		    if (curnote.getRhythmValue()>20) {
			System.out.println("long note "+curnote.getRhythmValue()+" at "+deltatime);
		    }
		    if (rmin>curnote.getRhythmValue()) rmin=curnote.getRhythmValue();
		    if (rmax<curnote.getRhythmValue()) rmax=curnote.getRhythmValue();

		}
		allNoteInfo= mergeNoteInfoVectors(allNoteInfo,phraseNoteInfo);
	    }
	}

	// debug:
	System.out.println("debug: rmin="+rmin+", rmax="+rmax);
	/* shows min and max rhythm values found in file. */


	/* Now all notes in the midi file are represented by a NoteInfo
	 * object in the vector allNoteInfo, sorted by begintime. 
	 * Build the graph! */
	vertices = new Vector();
	edges = new Vector();

	// add notes to graph as vertices in groups of simultaneous notes
	int i=0;
	Vector lastVertexGroup = new Vector();
	while (i<allNoteInfo.size()) {
	    Vector vertexGroupToBeAdded = new Vector();
	    NoteInfo ninf = (NoteInfo) allNoteInfo.get(i);
	    double begintime = ninf.starttime;

	    vertexGroupToBeAdded.add(new SimpleVertex(ninf,FuncLib.getTimeStamp()));
	    // add all the NoteInfos that are simultaneous with ninf, to the group
	    for (int k=i+1; 
		 k<allNoteInfo.size() && ((NoteInfo) allNoteInfo.get(k)).starttime==begintime; 
		 k++) {
		vertexGroupToBeAdded.add(new SimpleVertex((NoteInfo) allNoteInfo.get(k),FuncLib.getTimeStamp()));
		i=k;
	    }
	    addVertexGroup(lastVertexGroup,vertexGroupToBeAdded);
	    i++;
	}

	if (!includeInterPartEdges) {
	    removeInterPartEdges();
	}
	System.out.println("Done building mothergraph!");

	calculateVertexDistances();

	System.out.println("Vertices: "+vertices.size()+", edges: "+edges.size());
    }
    
    Vector partsFromMusedata;
    /** Vector of Vectors with notes and rests and boundaries. */ 
    Vector partsWithBoundaries;

    /** Constructs a graph from a Musedata file */
    public MotherGraph(String filename, boolean includeInterPartEdges, 
		       boolean combineTies, boolean combineRests){
	//************
	MusedataTokenizer mdt = new MusedataTokenizer(filename);
	partsFromMusedata = mdt.getGraph(combineTies,combineRests);

	System.out.println("Number of parts: "+partsFromMusedata.size());
	
	for(int i=0;i<partsFromMusedata.size();i++){
	    Vector pa=(Vector)partsFromMusedata.get(i);
	    System.out.println("part"+i+".size():"+pa.size());
	}
	theScore=mdt.dummyScore;



	//************

	System.out.println("Building graph...");
	title="Graph for file "+filename;
	Vector allNoteInfo = new Vector();
	for(int i=0;i<partsFromMusedata.size();i++){
	    Vector part = (Vector)partsFromMusedata.get(i);
	    allNoteInfo = mergeNoteInfoVectors(allNoteInfo,part);
	}

	/* Now all notes in the midi file are represented by a NoteInfo
	 * object in the vector allNoteInfo, sorted by begintime. 
	 * Build the graph! */
	vertices = new Vector();
	edges = new Vector();

	// add notes to graph as vertices in groups of simultaneous notes
	int i=0;
	Vector lastVertexGroup = new Vector();
	while (i<allNoteInfo.size()) {
	    Vector vertexGroupToBeAdded = new Vector();
	    NoteInfo ninf = (NoteInfo) allNoteInfo.get(i);
	    double begintime = ninf.starttime;
	    
	    vertexGroupToBeAdded.add(new SimpleVertex(ninf,FuncLib.getTimeStamp()));
	    // add all the NoteInfos that are simultaneous with ninf, to the group
	    for (int k=i+1; 
		 k<allNoteInfo.size() && ((NoteInfo) allNoteInfo.get(k)).starttime==begintime; 
		 k++) {
		vertexGroupToBeAdded.add(new SimpleVertex((NoteInfo) allNoteInfo.get(k),FuncLib.getTimeStamp()));
		i=k;
	    }
	    addVertexGroup(lastVertexGroup,vertexGroupToBeAdded);
	    i++;
	}

	if (!includeInterPartEdges) {
	    removeInterPartEdges();
	}
	System.out.println("Done building graph!");

	try {calculateVertexDistances();}
	catch(Exception e) {System.out.println(e.getMessage());}

	System.out.println("Vertices: "+vertices.size()+", edges: "+edges.size());
    }


    /** Used in the construction of the graph. */
    void addVertexGroup(Vector last /* of SimpleVertex */, 
			Vector next /* of SimpleVertex */) {

	Vector toBeRemoved = new Vector();
	// add a full (directed) connection between layers last and next
	for (int i=0; i<last.size(); i++) {
	    SimpleVertex lastmv = (SimpleVertex) last.get(i);
	    boolean remove = false;
	    for (int j=0; j<next.size(); j++) {
		SimpleVertex nextmv = (SimpleVertex) next.get(j);
		if (lastmv.starttime + 
		    lastmv.getLength() <= nextmv.starttime) {
		    remove = true;
		    MusicEdge me = new MusicEdge(lastmv,nextmv,MusicEdge.FOLLOWS);
		    lastmv.outEdges.add(me);
		    nextmv.inEdges.add(me);
		    edges.add(me);
		}
	    }
	    if (remove) toBeRemoved.add(lastmv);
	}

	for (int i=0; i<toBeRemoved.size(); i++) {
	    last.remove(toBeRemoved.get(i));
	}

	// add new vertices to graph and interconnect these new vertices with
	// SIMULTANEOUS edges
	for (int i=0; i<next.size(); i++) {
	    SimpleVertex mv1 = (SimpleVertex) next.get(i);
	    last.add(mv1);
	    vertices.add(mv1);
	    
	    // interconnect with SIMULTANEOUS relation edges
	    for (int j=0; j<next.size(); j++) {
		if (i!=j) {
		    SimpleVertex mv2 = (SimpleVertex) next.get(j);
		    MusicEdge me=new MusicEdge(mv1,mv2,MusicEdge.SIMULTANEOUS);
		    mv1.outEdges.add(me);
		    mv2.inEdges.add(me);
		    edges.add(me);
		}
	    }
	}
    }

    /** Used in the construction of the graph. */
    Vector mergeNoteInfoVectors(Vector v1, Vector v2) {
	if (v1.size()==0) return (Vector) v2.clone();
	if (v2.size()==0) return (Vector) v1.clone();

	Vector result = new Vector(); // of NoteInfo

	// Both vectors are non-empty
	NoteInfo addme;
	int elems = v1.size() + v2.size();
	int p1=0, p2=0;
	for (int i=0; i<elems; i++) {
	    if (p1>=v1.size()) {
		// no more elems in v1
		addme=(NoteInfo) v2.get(p2);
		p2++;
	    }
	    else if (p2>=v2.size()) {
		// no more elems in v2
		addme=(NoteInfo) v1.get(p1);
		p1++;
	    }
	    else {
		if (((NoteInfo) v1.get(p1)).starttime <= ((NoteInfo) v2.get(p2)).starttime) {
		    addme = (NoteInfo) v1.get(p1);
		    p1++;
		}
		else {
		    addme = (NoteInfo) v2.get(p2);
		    p2++;
		}
	    }
	    result.add(addme);
	}

	return result;
    }

    /** Returns a *shallow* clone of this MotherGraph, which is a {@link
       NonSeqSubgraph} -- Call MusicGraph.deepClone() instead if you really
       really want to do the entire deep cloning. */
    public Object clone() {
	NonSeqSubgraph cl = new NonSeqSubgraph("NonSeqSubgraph("+title+")",this);
	cl.vertices = (Vector) this.vertices.clone();
	cl.edges = (Vector) this.edges.clone();
	return cl;
    }

    /** Returns a *shallow* clone of this MotherGraph which is a
        {@link SeqSubgraph}, if this MotherGraph is in fact
        sequential; if not, null is returned. */
    public SeqSubgraph sequentialClone() {
	NonSeqSubgraph cl = new NonSeqSubgraph("NonSeqSubgraph("+title+")",this);
	cl.vertices = (Vector) this.vertices.clone();
	cl.edges = (Vector) this.edges.clone();

	if (cl.isStrongSequential()) {
	    SeqSubgraph result = new SeqSubgraph("SeqSubgraph("+title+")",this);
	    result.vertices = (Vector) this.vertices.clone();
	    result.edges = (Vector) this.edges.clone();
	    return result;
	}

	return null;
    }

    public String toString() {
	String result = "MotherGraph [title="+title+"]\n";
	result += super.toString();
	return result;
    }


    /** This method is used to run through the graph and once and for
       all set the max/minDistanceFromStart/End variables in all
       vertices. These distances are used to get an idea about whether
       it is possible from a given vertex to construct a random
       sequential subgraph of a given length. In case it isn't, it
       would take exponential time in the desired length (or,
       actually, in the max length possible in the real graph, which
       is less than the desired length when it isn't possible to
       construct it) of the sequential subgraph to try all
       possibilities and realise that it is impossible. NB: this
       method should be called if the topology of a MotherGraph is
       changed, e.g. after compound substitution. The
       compoundSubstitution method does this automatically. */
    public void calculateVertexDistances() {
	/* do breadth-first search for distances from start */
	System.out.print("Calculating vertex distances...");
	// reset vertex distance info and graphic layout info
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    mv.maxDistanceFromStart=0;
	    mv.minDistanceFromStart=Integer.MAX_VALUE;
	    mv.maxDistanceFromEnd=0;
	    mv.minDistanceFromEnd=Integer.MAX_VALUE;

	    mv.upperLeftPosition=null;
	    mv.lowerRightPosition=null;
	}

	Vector wave = getStrongSources();
	for (int i=0; i<wave.size(); i++) {
	    MusicVertex mv = (MusicVertex) wave.get(i);
	    mv.maxDistanceFromStart=0;
	    mv.minDistanceFromStart=0;
	}
	while (wave.size()>0) {
	    Vector nextwave = new Vector();
	    for (int i=0; i<wave.size(); i++) {
		MusicVertex mv = (MusicVertex) wave.get(i);
		for (int j=0; j<mv.outEdges.size(); j++) {
		    MusicEdge me = (MusicEdge)mv.outEdges.get(j);
		    if (me.isStrongFollowEdge()) {
			MusicVertex to = me.to;
			int maxdist = mv.maxDistanceFromStart +1;
			int mindist = mv.minDistanceFromStart +1;
			// System.out.println("("+maxdist+","+mindist+")");
			if (to.maxDistanceFromStart<maxdist)
			    to.maxDistanceFromStart=maxdist;
			if (to.minDistanceFromStart>mindist)
			    to.minDistanceFromStart=mindist;
			if (!nextwave.contains(to))
			    nextwave.add(to);
		    }
		}
	    }
	    wave=nextwave;
	}
	
	/* do breadth-first search for distances from end */
	wave = getStrongSinks();
	for (int i=0; i<wave.size(); i++) {
	    MusicVertex mv = (MusicVertex) wave.get(i);
	    mv.maxDistanceFromEnd=0;
	    mv.minDistanceFromEnd=0;
	}
	while (wave.size()>0) {
	    Vector nextwave = new Vector();
	    for (int i=0; i<wave.size(); i++) {
		MusicVertex mv = (MusicVertex) wave.get(i);
		for (int j=0; j<mv.inEdges.size(); j++) {
		    MusicEdge me = (MusicEdge)mv.inEdges.get(j);
		    if (me.isStrongFollowEdge()) {
			MusicVertex dad = me.from;
			int maxdist = mv.maxDistanceFromEnd +1;
			int mindist = mv.minDistanceFromEnd +1;
			if (dad.maxDistanceFromEnd<maxdist)
			    dad.maxDistanceFromEnd=maxdist;
			if (dad.minDistanceFromEnd>mindist)
			    dad.minDistanceFromEnd=mindist;
			if (!nextwave.contains(dad))
			    nextwave.add(dad);
		    }
		}
	    }
	    wave=nextwave;
	}
	System.out.println("done!");
    }
    
    /** Constructs a random non-sequential subgraph of this
        MotherGraph. */
    public NonSeqSubgraph getRandomNonSeqSubgraph(int size) throws Exception {
	NonSeqSubgraph result=null;
	int[] randomSeq = makeRandomNsequence(vertices.size());
	for (int j=0; j<randomSeq.length; j++) {
	    MusicVertex mv = (MusicVertex) vertices.get(randomSeq[j]);
	    result = new NonSeqSubgraph(mv,this);
	    try { 
		result.extend(size-1);
		if (!result.graphIsConnected(true/*include sim edges*/,
					     true/*include weak edges*/)) {
		    System.out.println("getrandomnonseqsubgraph returns unconnected graph:\n"+result.toString());
		}
		return result;
	    }
	    catch(Exception e) {
		System.out.println("getrandomnonseqsubgraph caught an exception: "+e.getMessage());
		throw new Exception(e.getMessage());
	    }
	}
	return null;
    }

    /* Subgraph SAFE */
    /** Constructs a random sequential subgraph of this
        MotherGraph. */
    public SeqSubgraph getRandomSeqSubgraph(int size) {
	//System.out.println("size="+size);
	SeqSubgraph result;

	//	System.out.println("grss "+size);
	int[] randomSeq = makeRandomNsequence(vertices.size());

	for (int i=0; i<randomSeq.length; i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(randomSeq[i]);

	    /* check if at all possible to build subgraph of this size
	     * from here */
	    if (mv.maxDistanceFromStart+mv.maxDistanceFromEnd+1>=size) {
		// yes, possible
		int left = rand.nextInt(size-1);
		// size-1 because mv is neither to the left or right of itself.
		int right = size-left-1;
		while (left>mv.maxDistanceFromStart) {
		    left--;
		    right++;
		}
		while (right>mv.maxDistanceFromEnd) {
		    left++;
		    right--;
		}

		/* now we know for certain that it's possible to expand
		 * a subgraph 'left' times to the left and 'right' times
		 * to the right from mv	*/
		result = new SeqSubgraph(mv,this);
		
		try {
		    result.extendRight(right);
		    result.extendLeft(left);
		} catch(Exception e) {
		    System.out.println("getRandomSeqSubgraph: exception caught when trying to extend. "+e.getMessage());
		}
		// boolean success = tryBuildingNsizeSeqSub(result,mv,size);
		return result;
	    } else {
		// no, impossible. try another vertex (restart forloop)
	    }
	}
	System.out.println("MotherGraph.randomSeqSubgraph: Couldn't find a seq subgraph of size="+size);
	// failure to find subgraph of given size.
	return null;
    }

    /* Subgraph SAFE */
    boolean tryBuildingNsizeSeqSub(SeqSubgraph mg, MusicVertex mv, int n) {
	if (n==0) return true;
	int[] randomSeq = makeRandomNsequence(mv.outEdges.size());
	
	for (int i=0; i<randomSeq.length; i++) {
	    MusicEdge me = (MusicEdge) mv.outEdges.get(randomSeq[i]);
	    if (me.isStrongFollowEdge()
		&& edges.contains(me) /* subgraph safety */) {
		if (tryBuildingNsizeSeqSub(mg,me.to,n-1)){
		    mg.vertices.add(mv);
		    if (n>1) mg.edges.add(me);
		    return true;
		}
	    }
	}
	return false;
    }

    /* CAN ONLY BE CALLED from the special constructor for
     * constructing graphs without edges between parts, so we may
     * assume that the graph does not contain any Compound
     * Vertices. */
    private void removeInterPartEdges() {
	Vector toBeRemoved = new Vector(); // of MusicEdge
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    if (me.isStrongFollowEdge() &&
		me.from instanceof SimpleVertex &&
		me.to instanceof SimpleVertex) {
		SimpleVertex sv1=(SimpleVertex)me.from;
		SimpleVertex sv2=(SimpleVertex)me.to;
		if (sv1.info.thePart!=sv2.info.thePart) {
		    // remove!
		    toBeRemoved.add(me);
		}
	    }
	}
	for (int i=0; i<toBeRemoved.size(); i++) {
	    MusicEdge me = (MusicEdge) toBeRemoved.get(i);
	    edges.remove(me);
	    me.from.outEdges.remove(me);
	    me.to.inEdges.remove(me);
	}
    }

    // REDESIGN SMARTER (THIS IS SAME AS SUBGRAPH VERSION BUT NEED NOT BE)
    /** Returns a Vector of Vector of MusicEdge. */
    public Vector getAllProperFollowInEdges(boolean includeWeakEdges) {
	Vector inEdgeVectors = new Vector();

	/* find subgraph 1's edges */	
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    Vector ownedInEdges = new Vector();
	    for (int j=0; j<mv.inEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.inEdges.get(j);
		if (me.isStrongFollowEdge() && edges.contains(me)) {
		    ownedInEdges.add(me);
		}
	    }
	    inEdgeVectors.add(ownedInEdges);
	}
	return inEdgeVectors;
    } 
    // REDESIGN SMARTER (THIS IS SAME AS SUBGRAPH VERSION BUT NEED NOT BE)
    /** Returns a Vector of Vector of MusicEdge. */
    public Vector getAllProperFollowOutEdges(boolean includeWeakEdges) {
	Vector outEdgeVectors = new Vector();

	/* find subgraph 1's edges */	
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    Vector ownedOutEdges = new Vector();
	    for (int j=0; j<mv.outEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.outEdges.get(j);
		if (me.isStrongFollowEdge() && edges.contains(me)) {
		    ownedOutEdges.add(me);
		}
	    }
	    outEdgeVectors.add(ownedOutEdges);
	}
	return outEdgeVectors;
    } 

    // REDESIGN SMARTER (THIS IS SAME AS SUBGRAPH VERSION BUT NEED NOT BE)
    /** Returns a Vector of Vector of MusicEdge. */
    public Vector getAllProperInEdges(boolean includeWeakEdges) {
	Vector inEdgeVectors = new Vector();
	/* find subgraph 1's edges */	
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    Vector ownedInEdges = new Vector();
	    for (int j=0; j<mv.inEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.inEdges.get(j);
		if (edges.contains(me)) {
		    ownedInEdges.add(me);
		}
	    }
	    inEdgeVectors.add(ownedInEdges);
	}
	return inEdgeVectors;
    } 
    // REDESIGN SMARTER (THIS IS SAME AS SUBGRAPH VERSION BUT NEED NOT BE)
    /** Returns a Vector of Vector of MusicEdge. */
    public Vector getAllProperOutEdges(boolean includeWeakEdges) {
	Vector outEdgeVectors = new Vector();

	/* find subgraph 1's edges */	
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    Vector ownedOutEdges = new Vector();
	    for (int j=0; j<mv.outEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.outEdges.get(j);
		if (edges.contains(me)) {
		    ownedOutEdges.add(me);
		}
	    }
	    outEdgeVectors.add(ownedOutEdges);
	}
	return outEdgeVectors;
    } 

    /** Sets the usage counters of all vertices in this MotherGraph to
        zero. */
    public void resetUsageInfo() {
	    for (int i=0; i<vertices.size(); i++) {
		MusicVertex mv = (MusicVertex) vertices.get(i);
		mv.useCounter=0;
	    }
    }

    /** Returns the originating Score object; if this MotherGraph was
        created from a midifile, the conversion using the jMusic
        package created such a Score object; if this MotherGraph was
        created from a MuseData file, the Score is a dummy object. */
    public Score getScore() {
	return theScore;
    }

    /** Returns the size of the maximal sequential subgraph that could
        possibly be constructed from this MotherGraph. */
    public int getMaximalSequentialSubgraphSize() {
	Vector sons = getStrongSources();
	int count=0;
	while (sons.size()>0) {
	    count++;
	    sons=getAllSons(sons);
	}
	return count;
    }

    /** Returns a Vector of {@link MusicVertex} which follow upon any
        of the music vertices in the input vector, i.e. there is a
        FOLLOWS {@link MusicEdge} from the vertex in the input vector
        to the one in the output vector.  */
    public Vector getAllSons(Vector vert /* of MusicVertex */) {
	Vector result = new Vector();
	for (int i=0; i<vert.size(); i++) {
	    MusicVertex mv = (MusicVertex) vert.get(i);
	    for (int j=0; j<mv.outEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.outEdges.get(j);
		if (me.isStrongFollowEdge()) {
		    if (!result.contains(me.to))
			result.add(me.to);
		}
	    }
	}
	return result;
    }

    /** Definition of abstract method from class {@link
        MusicGraph}. */
    public MotherGraph getMotherGraph() { return this; }
    
    /** Definition of abstract method from class {@link MusicGraph};
        simply returns the originating Score just as getScore(). */
    public Score convertToScore() { return theScore; }

    /** Substitutes a subgraph of this MotherGraph with a newly
        created {@link CompoundVertex}. */
    public CompoundVertex compoundSubstitution(Subgraph sg, 
					       Nonterminal nt,
					       SimilarityMeasure measure,
					       SimMeasureResult smresult) {
	GraphProductionRule gpr = new GraphProductionRule(nt,measure);
	/* build right side of prod.rule to be sg with its context */
	/* include the subgraph and all outside edges connected to it,
           and their vertices */
	if (!contains(sg)) {
	    new Exception().printStackTrace();
	    
	    System.out.println("ALARM: MotherGraph.compoundSubstitution called with subgraph not contained in it: \n"+sg.toString());
	    GraphWindow gw = new GraphWindow(this,200,200,true/*seq*/);
	    // gw.setAndShowHighlight(sg);
	    try {Thread.sleep(1800000);} catch(Exception e) {}
	    System.exit(0);
	}
	Subgraph right = (Subgraph) sg.clone();

	for (int i=0; i<edges.size(); i++) {
	    // find all edges with exactly one vertex in Subgraph sg
	    MusicEdge me = (MusicEdge) edges.get(i);
	    if (sg.vertices.contains(me.from) &&
		!sg.vertices.contains(me.to)) {
		// edge from sg to outside
		right.edges.add(me);
		if (!right.vertices.contains(me.to)) {
		    // we only want to include vertices once
		    right.vertices.add(me.to);
		}
	    } else if (!sg.vertices.contains(me.from) &&
		       sg.vertices.contains(me.to)) {
		// edge from outside to sg
		right.edges.add(me);
		if (!right.vertices.contains(me.from)) {
		    // we only want to include vertices once
		    right.vertices.add(me.from);
		}
	    }
	}
	// make sure all mothergraph edges between vertices in right
	// are included
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge)edges.get(i);
	    if (right.vertices.contains(me.from) &&
		right.vertices.contains(me.to)){
		// both endpoints are in right
		if (!right.edges.contains(me)) {
		    right.edges.add(me);
		}
	    }
	}
	
	gpr.rightside=(Subgraph)(right.deepClone().clone());
	NonSeqSubgraph left = new NonSeqSubgraph("LeftGPR("+title+")",this);

	// find earliest occurring vertex in sg
	double starttime=Double.MAX_VALUE;
	for (int i=0; i<sg.vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex)sg.vertices.get(i);
	    if (mv.starttime< starttime) starttime = mv.starttime;
	}

	CompoundVertex cv = new CompoundVertex(sg, starttime,
					       FuncLib.getTimeStamp(),
					       gpr,
					       smresult);
	Vector edgesToBeRemoved = new Vector(); // of MusicEdges
	Vector edgesToBeAdded = new Vector(); // of MusicEdges

	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);

	    if (cv.subgraph.vertices.contains(me.from)) {
		if (cv.subgraph.vertices.contains(me.to)) {
		    // CASE 1: edge inside the compound subgraph being
		    // abstracted. Remove it
		    edgesToBeRemoved.add(me);
		} else {
		    // CASE 2: edge from the compound to the mothergraph
		    boolean partwise = !hasInterpartEdges();
		    if (!partwise || (partwise && 
				      me.fromAnchor().info.thePart== 
				      me.toAnchor().info.thePart )){ 
			// decide to add weak or strong edge
			int edgetype=0;
			if (me.isFollowEdge()) {
			    if (me.to.starttime==cv.starttime+cv.getLength()) {
				edgetype=MusicEdge.FOLLOWS;
			    } else {
				edgetype=MusicEdge.WEAKFOLLOWS;
			    }
			} else if (me.isSimultaneousEdge()) {
			    if (me.to.starttime==cv.starttime) {
				edgetype=MusicEdge.SIMULTANEOUS;
			    } else {
				edgetype=MusicEdge.WEAKSIMULTANEOUS;
			    }
			}
			
			Vector frAnch = (Vector) me.fromAnchorVector.clone();
			Vector toAnch = (Vector) me.toAnchorVector.clone();
			
			int inx = cv.getIndexOf(me.from);
			frAnch.insertElementAt(new Integer(inx),0);
			MusicEdge meNew=new MusicEdge(cv,me.to,edgetype,
						      frAnch,toAnch);
			edgesToBeAdded.add(meNew);
			cv.outEdges.add(meNew);
			me.to.inEdges.add(meNew);
		    }
		    edgesToBeRemoved.add(me);
		    me.to.inEdges.remove(me);
		    me.from.outEdges.remove(me);
		    
		}
	    } else {
		if (cv.subgraph.vertices.contains(me.to)) {
		    // CASE 3: edge from the mothergraph to the compound

		    boolean partwise = !hasInterpartEdges();
		    if (!partwise || (partwise && 
				      me.fromAnchor().info.thePart== 
				      me.toAnchor().info.thePart )){ 
			// decide to add weak or strong edge
			int edgetype=0;
			if (me.isFollowEdge()) {
			    if (me.to.starttime==cv.starttime) {
				edgetype=MusicEdge.FOLLOWS;
			    } else {
				edgetype=MusicEdge.WEAKFOLLOWS;
			    }
			} else if (me.isSimultaneousEdge()) {
			    if (me.to.starttime==cv.starttime) {
				edgetype=MusicEdge.SIMULTANEOUS;
			    } else {
				edgetype=MusicEdge.WEAKSIMULTANEOUS;
			    }
			}
			
			Vector frAnch = me.cloneFromAnchorVector();
			Vector toAnch = me.cloneToAnchorVector();
			toAnch.insertElementAt(new Integer(cv.getIndexOf(me.to)),0);
			MusicEdge meNew=new MusicEdge(me.from,cv,edgetype,
						      frAnch,toAnch);
			edgesToBeAdded.add(meNew);
			cv.inEdges.add(meNew);
			me.from.outEdges.add(meNew);
		    }
		    
		    edgesToBeRemoved.add(me);
		    me.to.inEdges.remove(me);
		    me.from.outEdges.remove(me);
		} else {
		    // CASE 4: edge from mothergraph to mothergraph -
		    // not concerned by the compound substitution
		}
	    }
	} // end for all edges

	// int cvsize = cv.subgraph.nestedSize();
	cv.subgraph = (Subgraph) cv.subgraph.deepClone().clone();
	cv.subgraph.sortVerticesByStarttime();
	// int cvsize2 = cv.subgraph.nestedSize();

	/*if (cvsize!=cvsize2) {
	    System.out.println("Mothergraph.compoundSubstitution(): cvsize changed from "+cvsize+" to "+cvsize2+" during deepclone.");
	    }*/

	// construct left side of gpr as the compound with its context
	left.vertices.add(cv);
	left.edges.addAll(edgesToBeAdded);
	for (int i=0; i<left.edges.size(); i++) {
	    MusicEdge me = (MusicEdge) left.edges.get(i);
	    MusicVertex mv = me.from;
	    if (!left.vertices.contains(mv)) left.vertices.add(mv);
	    mv = me.to;
	    if (!left.vertices.contains(mv)) left.vertices.add(mv);
	}
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    if (left.vertices.contains(me.from) &&
		left.vertices.contains(me.to)) {
		if (!left.edges.contains(me)) {
		    left.edges.add(me);
		}
	    }
	}

	gpr.leftside=(Subgraph)left.deepClone().clone();
	
	edges.removeAll(edgesToBeRemoved);
	edges.addAll(edgesToBeAdded);
	vertices.removeAll(sg.vertices);

	vertices.add(cv);
	sortVerticesByStarttime();
	// we must update all vertex distances, they are probably
	// changed by the substitution
	calculateVertexDistances();
	
	nt.prodRules.add(gpr);
	return cv;
    }

    //    public int numberOfLeftMostRests();

    /** This static method is used to create a MotherGraph from any
        (supported) file type. It checks if the file is a midi file or
        a MuseData file and calls the appropriate constructor. */
    public static MotherGraph getMomFromFile(String filename,
				       boolean interpartEdges,
				       boolean combineTies,
				       boolean combineRests) throws Exception{
	MotherGraph mother=null;
	if (filename.endsWith("md") ||
	    filename.endsWith("md2") ||
	    filename.endsWith("MD") ||
	    filename.endsWith("MD2")) {
	    // call musedata constructor
	    mother = new MotherGraph(filename,interpartEdges,combineTies,combineRests);
	} else if (filename.endsWith("mid") ||
		   filename.endsWith("MID")) {
	    // call midi constructor
	    mother = new MotherGraph(filename,interpartEdges);
	} else {
	    throw new Exception("Unrecognised file: "+filename+" -- filename must end with 'md', 'md2' or 'mid'");
	}
	return mother;
    }



}
