/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Vector;
import java.awt.Point;

/** A MusicEdge is one of the two most basic components of a {@link
    MusicGraph}, the other being the {@link MusicVertex}. Music edges
    may have different kinds, represented using the very important
    integer attribute 'edgetype'. Different edge types could have been
    modeled as subclasses of an abstract class MusicEdge, but we have
    chosen the primitive way to represent the different types. A
    MusicEdge is directed, pointing both to a 'from'- and a
    'to'-MusicVertex. */

public class MusicEdge {


    /** The attribute edgetype is a way to distinguish between
     * different kinds of edges, eg. edges that represent temporal
     * 'follows' relations between notes, and edges that represent
     * notes that are played simultaneously (temporal 'equals'
     * relation).  */
    public int edgetype;

    /** Edge type FOLLOWS (=strong FOLLOW)*/
    public static final int FOLLOWS=0;
    /** Edge type SIMULTANEOUS (=strong SIMULTANEOUS)*/
    public static final int SIMULTANEOUS=1;
    /** Edge type MATCH, used for the matching of {@link MusicVertex}
        pairs in the allMatches Vector, which is an attribute in class
        {@link NSSMVertexResult}. */
    public static final int MATCH=2;
    /** Edge type WEAKFOLLOWS*/
    public static final int WEAKFOLLOWS=3;
    /** Edge type WEAKSIMULTANEOUS */
    public static final int WEAKSIMULTANEOUS=4;

    /** The point in the graphical layout, where this edges starts. */
    public Point fromPoint=null;
    /** The point in the graphical layout, where this edges ends. */
    public Point toPoint = null;
    /** The MusicVertex that this MusicEdge departs from. */
    public MusicVertex from;
    /** The MusicVertex that this MusicEdge arrives to. */
    public MusicVertex to;

    /** A Vector of Integers describing the indices of MusicVertices
        inside CompoundVertices that we have to peel through to get to
        the SimpleVertex which is the anchor at the 'from'-end of this
        MusicEdge. */
    public Vector fromAnchorVector; // of Integer
    /** A Vector of Integers describing the indices of MusicVertices
        inside CompoundVertices that we have to peel through to get to
        the SimpleVertex which is the anchor at the 'to'-end of this
        MusicEdge. */
    public Vector toAnchorVector; // of Integer

    /** Resolves and returns the from-anchor of this MusicEdge. */
    public SimpleVertex fromAnchor() {
	return from.resolveAnchor(fromAnchorVector);
    }
    /** Resolves and returns the to-anchor of this MusicEdge. */
    public SimpleVertex toAnchor() {
	return to.resolveAnchor(toAnchorVector);
    }

    /** Returns true if this MusicEdge has been given graphical
        layout. */
    public boolean graphicLayoutDone() {
	if (fromPoint==null) return false;
	if (toPoint==null) return false;
	return true;
    }

    /** Constructor for edges between simple vertices. Anchor
        information is generated automatically. */
    public MusicEdge(SimpleVertex f, SimpleVertex t, int kind) {
	edgetype=kind;
	from=f;
	to=t;
	fromAnchorVector=new Vector();
	fromAnchorVector.add(new Integer(0));
	toAnchorVector=new Vector();
	toAnchorVector.add(new Integer(0));
    }

    /** Constructor edges between simple and/or compound
        vertices. Includes anchor information.*/
    public MusicEdge(MusicVertex f, MusicVertex s, int kind,
		     Vector fanch /* of Integer */, 
		     Vector tanch /* of Integer*/) {
	edgetype=kind;
	from=f;
	to=s;
	fromAnchorVector = fanch;
	toAnchorVector = tanch;
    }
    public String toString() {
	String result = "";
	if (isStrongSimultaneousEdge()) result+= "Simultaneous ";
	else if (edgetype==WEAKSIMULTANEOUS) result+="Weak sim ";
	else if (edgetype==WEAKFOLLOWS) result+="Weak ";
	result += "Edge from ["+from.toString()+"] to ["+to.toString()+"]";
	return result;
    }

    /** Includes toStrings of the anchors. */
    public String toStringWithAnchors() {
	String result = "";
	if (isStrongSimultaneousEdge()) result+= "Simultaneous ";
	else if (edgetype==WEAKSIMULTANEOUS) result+="Weak sim ";
	else if (edgetype==WEAKFOLLOWS) result+="Weak ";
	result += "Edge from anchor ["+fromAnchor().toString()+"] to anchor["+toAnchor().toString()+"]";
	return result;
    }

    public boolean isStrongFollowEdge() {
	if (edgetype==FOLLOWS) return true;
	else return false;
    }

    public boolean isStrongSimultaneousEdge() {
	if (edgetype==SIMULTANEOUS) return true;
	else return false;
    }
    
    public boolean isFollowEdge() {
	if (edgetype==FOLLOWS || edgetype==WEAKFOLLOWS) return true;
	else return false;
    }

    public boolean isSimultaneousEdge() {
	if (edgetype==SIMULTANEOUS || edgetype==WEAKSIMULTANEOUS) return true;
	else return false;
    }

    /** Returns true if edgetype==WEAKFOLLOWS or edgetype==WEAKSIMULTANEOUS. */
    public boolean isWeak() {
	if (edgetype==WEAKFOLLOWS || edgetype==WEAKSIMULTANEOUS) return true;
	else return false;
    }

    
    /** As yet, matching edges should not occur in graphs. We may
     * decide to let them later on. They are used as a data structure
     * for MusicVertex pairwise matching. */
    public boolean isMatchingEdge() {
	if (edgetype==MATCH) return true;
	else return false;
    }

    /** Returns a clone of the Vector describing the from-anchor. */
    public Vector cloneFromAnchorVector() {
	Vector result = new Vector();
	for (int i=0; i<fromAnchorVector.size(); i++) {
	    result.add(new Integer(((Integer) fromAnchorVector.get(i)).intValue()));
	}
	return result;
    }

    /** Returns a clone of the Vector describing the to-anchor. */
    public Vector cloneToAnchorVector() {
	Vector result = new Vector();
	for (int i=0; i<toAnchorVector.size(); i++) {
	    result.add(new Integer(((Integer) toAnchorVector.get(i)).intValue()));
	}
	return result;
    }

}
