/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Vector;
import java.util.Random;
import java.awt.Point;
import jm.music.data.*;
import jm.JMC;
import jm.util.*;
import jm.midi.*;

/** MusicGraph is the abstract superclass of {@link MotherGraph} and
 {@link Subgraph}. This class contains the fundamental attributes
 (vertices and edges) of a graph and a number of general methods to
 manipulate a graph.  */
public abstract class MusicGraph {

    /** A Vector of {@link MusicVertex}. */
    public Vector vertices; // of MusicVertex
    /** A Vector of {@link MusicEdge}. */
    public Vector edges; // of MusicEdge
    /** The title is just a name (a label) that makes it easier for
        the user to distinguish between MusicGraphs. */
    public String title;
    static Random rand = new Random();

    /** Returns the number of simple or compound vertices in this
        graph; see also method nestedSize(). */
    public int size() {
	return vertices.size();
    }
    
    /* public int numberOfRests(){
	int res=0;
	for(int i=0;i<vertices.size();i++){
	    MusicVertex mv=(MusicVertex)vertices.get(i);
	    if(mv.info.theNote.getPitch()==Integer.MIN_VALUE) {res++;}
	}
	return res;
	}*/ 
    
    /** Recursive definition: Returns the number of vertices in this
        graph plus the nested size of any compound vertices contained
        in this graph. It thus returns the number of {@link
        SimpleVertex} contained in this MusicGraph and all of its
        inner {@link CompoundVertex}'s. */
    public int nestedSize() {
	int result=0;
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    if (mv instanceof CompoundVertex) {
		result+= ((CompoundVertex)mv).subgraph.nestedSize();
	    }
	    else result++;
	}
	return result;
    }

    /** Returns a Vector of all MusicVertices that have no inEdges of
	type FOLLOWS. */
    /* Subgraph SAFE */
    public Vector getStrongSources() {
	Vector result = new Vector();
	// for all vertices, add it to result if it has 0 inEdges
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    int occurrenceCount = 0;
	    for (int j=0; j<edges.size(); j++) {
		MusicEdge me=(MusicEdge) edges.get(j);
		if (me.isStrongFollowEdge() && mv.inEdges.contains(me)) {
		    occurrenceCount++;
		}
	    }
	    if (occurrenceCount==0)
		result.add(mv);
	}
	return result;
    }

    /** Returns a Vector of all MusicVertices that have no outEdges of
	type FOLLOWS */
    /* Subgraph SAFE */
    public Vector getStrongSinks() {
	Vector result = new Vector();
	// for all vertices, add it to result if it has 0 outEdges
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    int occurrenceCount = 0;
	    for (int j=0; j<edges.size(); j++) {
		MusicEdge me=(MusicEdge) edges.get(j);
		if (me.isStrongFollowEdge() && mv.outEdges.contains(me)) {
		    occurrenceCount++;
		}
	    }
	    if (occurrenceCount==0)
		result.add(mv);
	}
	return result;
    }

    /** If this graph has exactly one sink, that sink is returned. If
        it has zero or more than one sink, null is returned. This is
        useful mostly for sequential subgraphs. */
    /* Subgraph SAFE */
    MusicVertex getSoleStrongSink() {
	Vector sinks = getStrongSinks();
	if (sinks.size()!=1) return null;
	else return (MusicVertex) sinks.get(0);
    }

    private int nextVertex=0;
    private double vertexDistance=0.0;

    /** For debugging. */
    public void checkSorting() {
	int stimeErr=0;
	int stampErr=0;
	for (int i=0; i<vertices.size()-1; i++) {
	    MusicVertex mv1 = (MusicVertex) vertices.get(i);
	    MusicVertex mv2 = (MusicVertex) vertices.get(i+1);
	    if (mv1.starttime>mv2.starttime) {
		stimeErr++;
	    }
	    if (mv1.starttime==mv2.starttime &&
		mv1.getLength()==mv2.getLength() &&
		mv1.objectTimeStamp>mv2.objectTimeStamp) {
		stampErr++;
	    }
	}
	System.out.println("starttime errors: "+stimeErr);
	System.out.println("timestamp errors: "+stampErr);
    }

    /* Subgraph SAFE */
    /** The group tokenizer is a mechanism built into this class which
       allows us to group nodes that are simultaneous (d=0.0) or
       within some distance d. Groups are found and returned by the
       getNextGroup() method, but some variables need to be
       initialized first. That initialization is accomplished by this
       method. */
    public boolean initGroupTokenizer(double d) {
	nextVertex=0;
	vertexDistance=d;
	sortVerticesByStarttime();
	return true;
    }

    /* Subgraph SAFE */
    /** This method is the heart of the group tokenizer built into the
        MusicGraph class; it finds and returns a Vector of
        MusicVertex. initGroupTokenizer() must have been called before
        using this method. Returns a Vector of size=0 when no more
        vertices in graph. */
    public Vector getNextGroup() {
	Vector result = new Vector(); // of MusicVertex
	if (nextVertex>=vertices.size()) return result;
	MusicVertex mv=null;
	try {  mv= (MusicVertex) vertices.get(nextVertex);}
	catch(Exception e) { 
	    System.out.println(e.getMessage()+"\nsize="+vertices.size());
	}
	double groupstart = mv.starttime;
	result.add(mv);
	nextVertex++;
	if (nextVertex>= vertices.size()) return result;
	mv=(MusicVertex) vertices.get(nextVertex);
	while (mv.starttime-groupstart <= vertexDistance && nextVertex<vertices.size()) {
	    result.add(mv);
	    nextVertex++;
	    if (nextVertex<vertices.size()) {
		try {
		    mv=(MusicVertex) vertices.get(nextVertex);
		}
		catch (Exception e) {
		    System.out.println(e.getMessage()+"\nsize="+vertices.size());
		}
	    }
	}
	return result;
    }

    /** Uses the transitive closure of the graph to determine if
       removing the vertex at the specified index in the "vertices"
       Vector would split this MusicGraph into two unconnected
       halves. */
    public boolean vertexAtIndexSplitsSubgraph(int index, 
					       boolean[][] originalClosure) {
	// a graph of size 0, 1 or 2 cannot be split by removal of 1
	// vertex
	if (vertices.size()<3) return false;

	/* find out if vertex at 'index' in vertices can be
	 * removed without splitting the subgraph */
	boolean[][] modified=copyBoolMatrix(originalClosure);

	// remove 'index' from the closure matrix
	for (int j=0; j<modified.length; j++) {
	    modified[index][j]=false;
	    modified[j][index]=false;
	}
	modified = transitiveClosureOf(modified);
	
	/* if removing the vertex at 'index' splits the graph, there
	 * will be more 'false' values in a row of the transitive
	 * closure matrix a row number !='index' than simply the value
	 * for 'index'. Find some other row number -- it doesn't
	 * matter which one: */
	int anotherIndex=index;
	while (anotherIndex==index) {
	    anotherIndex=rand.nextInt(vertices.size());
	}
	for (int k=0; k<modified.length; k++) {
	    /* of course removing 'index' will make it unreachable,
	     * don't test for that */
	    if (k!=index) {
		/* are there other unreachable vertices? */
		if (!modified[k][anotherIndex]) {
		    return true;
		}
	    }
	}
	return false;
    }

    /* Subgraph SAFE */
    /** Returns a sequence of numbers from 0 to n-1 in random order. */
    int[] makeRandomNsequence(int n) {
	int[] result = new int[n];

	// fill with numbers
	for (int i=0; i<n; i++) {
	    result[i]=i;
	}
	// mix it up by swapping around
	for (int i=0; i<2*n; i++) {
	    int inx1 = rand.nextInt(n);
	    int inx2 = rand.nextInt(n);
	    int tmp = result[inx1];
	    result[inx1] = result[inx2];
	    result[inx2] = tmp;
	}
	return result;
    }

    /* Subgraph SAFE */
    /** Returns a string with the toString's of all vertices and edges
        in this MusicGraph. */
    public String toString() {
	String result= "Vertices:\n";
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    result+= "\t"+mv.toString()+"\n";
	}
	result+= "Edges:\n";
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    result+= "\t"+me.toString()+"\n";
	}
	return result;
    }

    /* Subgraph SAFE*/
    /** If this MusicGraph contains() the other MusicGraph, and the
         other MusicGraph contains() this MusicGraph, true is
         returned, otherwise false. */
    public boolean equals(MusicGraph other) {
	if (size()!=other.size()) {
	    // System.out.println("-EQ-");
	    return false;
	}
	if (contains(other) && other.contains(this)) {
	    // System.out.println("+EQ+");
	    return true;
	}
	else 	    { 
	    // System.out.println("-EQ-");
	    return false;
	}
    }

    /* Subgraph SAFE*/
    /** Checks if all vertices and all edges in the other MusicGraph
     are also contained in this MusicGraph. */
    public boolean contains(MusicGraph other) {
	for (int i=0; i<other.vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) other.vertices.get(i);
	    if (!vertices.contains(mv)) return false;
	}
	for (int i=0; i<other.edges.size(); i++) {
	    MusicEdge me = (MusicEdge) other.edges.get(i);
	    if (!edges.contains(me)) return false;
	}
	return true;
    }

    /* Subgraph SAFE*/
    /** Returns the sum of the number of vertices and the number of
        edges that this MusicGraph has in common with the other; this
        is mostly useful for comparing {@link Subgraph}s. "Having in
        common" means that both MusicGraphs are pointing to the same
        {@link MusicVertex} or {@link MusicEdge} objects. A
        MusicVertex or MusicEdge object is always owned by some {@link
        MotherGraph}. */
    public int overlap(MusicGraph other) {
	int result=0;
	for (int i=0; i<other.vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) other.vertices.get(i);
	    if (vertices.contains(mv)) result++;
	}
	for (int i=0; i<other.edges.size(); i++) {
	    MusicEdge me = (MusicEdge) other.edges.get(i);
	    if (edges.contains(me)) result++;
	}
	return result;
    }

    /* Subgraph SAFE*/
    /** A helper function that calls the sortVertexVectorByStarttime()
        with the 'vertices' Vector as a parameter. */
    public void sortVerticesByStarttime() {
	sortVertexVectorByStarttime(vertices);
    }

    /** Sorts the input Vector of {@link MusicVertex} by start
        time. */
    public void sortVertexVectorByStarttime(Vector vert) {
	// bubblesort
	for (int i=0; i<vert.size(); i++) {
	    boolean changed = false;
	    for (int j=0; j<vert.size()-1; j++) {
		MusicVertex mv1 = (MusicVertex) vert.get(j);
		MusicVertex mv2 = (MusicVertex) vert.get(j+1);
		boolean swap=false;
		if (mv1.starttime>mv2.starttime) {
		    swap=true;
		} else if (mv1.starttime==mv2.starttime ) {
		    if (mv1.getLength()> mv2.getLength()) {
			swap = true;
		    } else if (mv1.getLength()== mv2.getLength()) {
			if (mv1.objectTimeStamp>mv2.objectTimeStamp) {
			    swap=true;
			}
		    }
		}
		if (swap) {
		    // swap
		    vert.set(j,mv2);
		    vert.set(j+1,mv1);
		    changed=true;
		}
	    }
	    if (!changed) return;
	}
    }

    /* Subgraph SAFE*/
    /** @deprecated Not used any more. Use sortVerticesByStarttime()
        instead. */
    public void sortEdges() {
	// bubblesort
	for (int i=0; i<edges.size(); i++) {
	    boolean changed = false;
	    for (int j=0; j<edges.size()-1; j++) {
		MusicEdge me1 = (MusicEdge) edges.get(j);
		MusicEdge me2 = (MusicEdge) edges.get(j+1);
		if (me1.from.starttime>me2.from.starttime ||
		    (me1.from.starttime==me2.from.starttime &&
		     me1.from.objectTimeStamp>me2.from.objectTimeStamp) ||
		    (me1.from.starttime==me2.from.starttime &&
		     me1.from.objectTimeStamp==me2.from.objectTimeStamp &&
		     me1.to.starttime>me2.to.starttime) ||
		    (me1.from.starttime==me2.from.starttime &&
		     me1.from.objectTimeStamp==me2.from.objectTimeStamp &&
		     me1.to.starttime==me2.to.starttime &&
		     me1.to.objectTimeStamp>me2.to.objectTimeStamp)
		    ) {
		    // swap
		    edges.set(j,me2);
		    edges.set(j+1,me1);
		    changed=true;
		}

	    }
	    if (!changed) return;
	}
    }

    public abstract Object clone();

    /** Returns a Vector of {@link SeqSubgraph}. CAREFUL!! The
     * complexity of this method easily exceeds anything computable in
     * realistic time! Consider using its little brother,
     * getNSequentialSubgraphsOfSize() instead. */
    public Vector getAllStrongSequentialSubgraphsOfSize(int size) {
	Vector allsubs = new Vector(); // of SeqSubgraph
	MusicGraph clone = (MusicGraph) clone();
	
	if(size==2){
	    for(int i=0;i<clone.edges.size();i++){
		SeqSubgraph ss=new SeqSubgraph("Subgraph("+title+")",getMotherGraph());
		MusicEdge e=(MusicEdge)clone.edges.get(i);
		if(e.isStrongFollowEdge()){
		    ss.edges.add(e);
		    ss.vertices.add(e.from);
		    ss.vertices.add(e.to);
		    allsubs.add(ss);
		}
	    }
	}else{
	    // for all groups, find all subgraphs of length 'size'
	    // beginning with a vertex in the group
	    initGroupTokenizer(0.0);
	    Vector group = getNextGroup();
	    
	    while (group.size()>0) {
		allsubs.addAll(clone.getBeginningSubgraphSet(size));
		clone.vertices.removeAll(group);
		Vector edgesToBeRemoved = new Vector();
		for (int i=0; i<clone.edges.size(); i++) {
		    MusicEdge me = (MusicEdge) clone.edges.get(i);
		    if (group.contains(me.from)) {
			edgesToBeRemoved.add(me);
		    }
		}
		clone.edges.removeAll(edgesToBeRemoved);
		group = getNextGroup();
	    }
	}
	return allsubs;
    }
    
    /** Like getAllSequentialSubgraphsOfSize(int size), but includes
       weak FOLLOW edges. */
    public Vector getAllSequentialSubgraphsOfSize(int size) {
	Vector allsubs = new Vector(); // of SeqSubgraph
	MusicGraph clone = (MusicGraph) clone();
	
	if(size==2){
	    for(int i=0;i<clone.edges.size();i++){
		SeqSubgraph ss=new SeqSubgraph("Subgraph("+title+")",getMotherGraph());
		MusicEdge e=(MusicEdge)clone.edges.get(i);
		if(e.isFollowEdge()){
		    ss.edges.add(e);
		    ss.vertices.add(e.from);
		    ss.vertices.add(e.to);
		    allsubs.add(ss);
		}
	    }
	}else{
	    // for all groups, find all subgraphs of length 'size'
	    // beginning with a vertex in the group
	    initGroupTokenizer(0.0);
	    Vector group = getNextGroup();
	    
	    while (group.size()>0) {
		allsubs.addAll(clone.getBeginningSubgraphSet(size));
		clone.vertices.removeAll(group);
		Vector edgesToBeRemoved = new Vector();
		for (int i=0; i<clone.edges.size(); i++) {
		    MusicEdge me = (MusicEdge) clone.edges.get(i);
		    if (group.contains(me.from)) {
			edgesToBeRemoved.add(me);
		    }
		}
		clone.edges.removeAll(edgesToBeRemoved);
		group = getNextGroup();
	    }
	}
	return allsubs;
    }
 

    /** Returns a Vector of at least n sequential subgraphs of size
       'size'. */
    public Vector getNSequentialSubgraphsOfSize(int n, int size) {
	Vector allsubs = new Vector(); // of SeqSubgraph
	MusicGraph clone = (MusicGraph) clone();

	// for all groups, find all subgraphs of length 'size'
	// beginning with a vertex in the group
	initGroupTokenizer(0.0);
	Vector group = getNextGroup();
		       
	while (group.size()>0) {
	    allsubs.addAll(clone.getBeginningSubgraphSet(size));
	    if (allsubs.size()>=size) {
		return allsubs;
	    }

	    clone.vertices.removeAll(group);
	    Vector edgesToBeRemoved = new Vector();
	    for (int i=0; i<clone.edges.size(); i++) {
		MusicEdge me = (MusicEdge) clone.edges.get(i);
		if (group.contains(me.from)) {
		    edgesToBeRemoved.add(me);
		}
	    }
	    clone.edges.removeAll(edgesToBeRemoved);

	    group = getNextGroup();
	}
	return allsubs;
    }
    public abstract MotherGraph getMotherGraph();

    public abstract int getMaximalSequentialSubgraphSize();

    /* Subgraph SAFE */
    /** Gives a Vector of {@link SeqSubgraph} which are all the
        possible sequential subgraphs of a given length (size) whose
        first MusicVertex is one of the first vertices in this
        MusicGraph. */
    /* TODO this method should be renamed getBeginningSeqSubgraphSet */
    public Vector getBeginningSubgraphSet(int size) {
	Vector sources = getStrongSources(); // of MusicVertex
	Vector subgraphs = new Vector(); // of SeqSubgraph
	for (int i=0; i<sources.size(); i++) {
	    subgraphs.add(new SeqSubgraph((MusicVertex)sources.get(i),getMotherGraph()));
	}
	int curSize=1;
	while (curSize<size) {
	    subgraphs = getAllPossibleRightExtended(subgraphs);
	    curSize++;
	}
	return subgraphs;
    }

    /* Subgraph SAFE */
    /** Gives a Vector of {@link SeqSubgraph} which are all the
       possible right extensions of the SeqSubgraphs in the input
       Vector; considers only right extensions possible inside this
       graph whether its a subgraph or not. */
    public Vector getAllPossibleRightExtended(Vector subs/* of SeqSubgraph */) {
	Vector result = new Vector(); // of SeqSubgraph

	for (int i=0; i<subs.size(); i++) {
	    SeqSubgraph mg = (SeqSubgraph) subs.get(i);
	    Vector sinks = mg.getStrongSinks();
	    if (sinks.size()==1) {
		MusicVertex lastVertex = (MusicVertex) sinks.get(0);
		for (int j=0; j<lastVertex.outEdges.size(); j++) {
		    MusicEdge me = (MusicEdge) lastVertex.outEdges.get(j);
		    if (me.isStrongFollowEdge() && edges.contains(me)) {
			SeqSubgraph mgClone = (SeqSubgraph) mg.clone();
			mgClone.edges.add(me);
			mgClone.vertices.add(me.to);
			mgClone.sortVerticesByStarttime();
			result.add(mgClone);
		    }
		}
	    }
	}
	return result;
    }
    /* Subgraph SAFE */
    /** Gives a Vector of {@link SeqSubgraph} which are all the
       possible right slides of the SeqSubgraphs in the input Vector;
       Considers only right slides possible inside this graph whether
       its a subgraph or not. */
    public Vector getAllPossibleRightSlides(Vector subs /* of SeqSubgraph */) {
	Vector result = new Vector(); // of SeqSubgraph
	Vector tmp = getAllPossibleRightExtended(subs);
	for (int i=0; i<tmp.size(); i++) {
	    SeqSubgraph mg = (SeqSubgraph) tmp.get(i);
	    try {
		mg.shortenLeft(1);
		boolean foundAlready=false;
		for (int j=0; j<result.size(); j++) {
		    SeqSubgraph mg2 = (SeqSubgraph) result.get(j);
		    if (mg.equals(mg2)) {
			foundAlready=true;
			break;
		    }
		}
		if (!foundAlready) result.add(mg);
	    }
	    catch(Exception e) {System.out.println(e.getMessage());}
	}
	return result;
    }

    /** 
     */
    public abstract Vector getAllProperFollowInEdges(boolean includeWeakEdges);
    public abstract Vector getAllProperFollowOutEdges(boolean includeWeakEdges);
    public abstract Vector getAllProperInEdges(boolean includeWeakEdges);
    public abstract Vector getAllProperOutEdges(boolean includeWeakEdges);

    /** Determines whether this graph is connected, with or without
     SIMULTANEOUS, WEAKFOLLOWS and WEAKSIMULTANEOUS edges, according
     to the parameters.  */
    public boolean graphIsConnected(boolean includeSimEdges,
				    boolean includeWeakEdges) {
	boolean[][] closure = getTransitiveClosure(includeSimEdges,
						   includeWeakEdges);
	
	boolean result = true;
	for (int i=0; i<closure.length; i++) {
	    result = result && closure[i][0];
	}
	return result;
    }

    /** Constructs an incidence matrix describing which vertices are
        connected to which other vertices.*/
    public boolean[][] getClosure(boolean includeSimEdges, 
				  boolean includeWeakEdges) {
	boolean[][] result = new boolean[size()][size()];
	Vector properInEdges;
	Vector properOutEdges;

	if (includeSimEdges) {
	    properInEdges = getAllProperInEdges(includeWeakEdges);
	    properOutEdges = getAllProperOutEdges(includeWeakEdges);
	} else {
	    properInEdges = getAllProperFollowInEdges(includeWeakEdges);
	    properOutEdges = getAllProperFollowOutEdges(includeWeakEdges);
	}

	// fill matrix
	for (int i=0; i<vertices.size(); i++) {
	    Vector in = (Vector) properInEdges.get(i);
	    for (int j=0; j<in.size(); j++) {
		MusicEdge me = (MusicEdge) in.get(j);
		int inx = vertices.indexOf(me.from);
		result[i][inx] =true;
		result[inx][i] =true;
	    }
	    Vector out = (Vector) properOutEdges.get(i);
	    for (int j=0; j<out.size(); j++) {
		MusicEdge me = (MusicEdge) out.get(j);
		int inx = vertices.indexOf(me.to);
		result[i][inx] =true;
		result[inx][i] =true;
	    }
	    result[i][i]=true;
	}
	return result;
    }

    /** Constructs a matrix describing which vertices can be reached
        from which other vertices.*/
    public boolean[][] getTransitiveClosure(boolean includeSimEdges,
					    boolean includeWeakEdges) {
	boolean[][] result = getClosure(includeSimEdges, includeWeakEdges);
	return transitiveClosureOf(result);
    }

    /** Helper function. */
    public static boolean[][] transitiveClosureOf(boolean[][] res) {
	boolean[][] tmp=boolMatrixMultiply(res,res);
	while (!boolMatrixEqual(tmp,res)) {
	    tmp=res;
	    res=boolMatrixMultiply(tmp,tmp);
	}
	return res;
    }

    /** Helper function. */
    public static boolean[][] copyBoolMatrix(boolean[][] m) {
	boolean[][] result = new boolean[m.length][m[0].length];

	for (int i=0; i<m.length; i++) {
	    for (int j=0; j<m[0].length; j++) {
		result[i][j]=m[i][j];
	    }
	}
	return result;
    }

    /** Helper function. */
    public static boolean boolMatrixEqual(boolean[][] m1, boolean[][] m2) {
	for (int i=0; i<m1.length; i++) {
	    for (int j=0; j<m1[0].length; j++) {
		if (m1[i][j]!=m2[i][j]) {
		    return false;
		}
	    }
	}
	return true;
    }
    /** Helper function. */
    public static boolean[][] boolMatrixMultiply(boolean[][] m1, boolean[][] m2) {
	if (m1.length!=m2[0].length || m1[0].length!=m2.length) return null;

	boolean[][] result = new boolean[m1.length][m1[0].length];
	for (int i=0; i<m1.length; i++) {
	    for (int j=0; j<m1[0].length; j++) {
		boolean tmp=false;;
		for (int k=0; k<m1[0].length; k++) {
		    tmp = tmp || (m1[i][k] && m2[k][j]);
		}
		result[i][j]=tmp;
	    }
	}
	return result;
    }

    /** For debugging. */
    public static void prettyPrintBoolMatrix(boolean[][] b) {
	System.out.println();
	for (int i=0; i<b.length; i++) {
	    for (int j=0; j<b[0].length; j++) {
		if (b[i][j]) System.out.print("T");
		else System.out.print(".");
	    }
	    System.out.println();
	}
    }

    /** GetScore returns a pointer to the original Score that this
       MusicGraph's mothergraph was created from. */
    public abstract Score getScore();
    /** ConvertToScore returns a Score representing the music in this
       MusicGraph. For subgraphs, it is a newly instantiated and built
       Score, for MotherGraphs it is a pointer to the original Score
       and thus returns the same as getScore(). */
    public abstract Score convertToScore();

    /** Returns the nestedNumberOfRests divided by the nestedSize. */
    public double getNestedRestPercentage(){
	return (double)nestedNumberOfRests()/(double)nestedSize();
    }

    double getRestPercentage() {
	int rests=0;
	for(int i=0;i<vertices.size();i++){
	    MusicVertex mv =(MusicVertex) vertices.get(i);
	    if (mv instanceof SimpleVertex) {
		SimpleVertex sv = (SimpleVertex)mv;
		if(sv.info.theNote.getPitch()==Integer.MIN_VALUE) rests++;
	    }
	}
	return rests;
    }

    /** Returns the number of vertices in this MusicGraph that
        represent a rest, plus the nestedNumberOfRests of the
        subgraphs inside any {@link CompoundVertex}'s in this
        MusicGraph. */
    public int nestedNumberOfRests() {
	int result=0;
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    if (mv instanceof CompoundVertex) {
		result+= ((CompoundVertex)mv).subgraph.nestedNumberOfRests();
	    }
	    else if (mv instanceof SimpleVertex) {
		SimpleVertex sv = (SimpleVertex)mv;
		if (sv.info.theNote.getPitch()==Integer.MIN_VALUE)
		    result++;
	    }
	}
	return result;
    }

    /** Returns the maximal value of the usage counter in every vertex
        in this MusicGraph. */
    public int maxUsage() {
	int result=0;
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    if (mv.useCounter>result) result= mv.useCounter;
	}
	return result;
    }

    /** Returns the least start time found in the 'vertices' Vector. */
    public double getMinStarttime() {
	double result=Double.MAX_VALUE;
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex)vertices.get(i);
	    if (mv.starttime<result) result=mv.starttime;
	}
	return result;
    }

    /** Returns the greatest end time found in the 'vertices'
        Vector. */
    public double getMaxEndtime() {
	double result=Double.MIN_VALUE;
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex)vertices.get(i);
	    if (mv.getEndtime()>result) result=mv.getEndtime();
	}
	return result;
    }

    /** Returns true if there are no {@link CompoundVertex} in this
        graph. */
    public boolean isSimpleGraph() {
	if (numOfCompoundVertices()>0) return false;
	else return true;
    }

    /** Returns the number of {@link CompoundVertex} in this graph. */
    public int numOfCompoundVertices() {
	int result =0;
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex)vertices.get(i);
	    if (mv instanceof CompoundVertex) result++;
	}
	return result;
    }
    /** The nested version of numOfCompoundVertices(). */
    public int nestedNumOfCompoundVertices() {
	int result =0;
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex)vertices.get(i);
	    if (mv instanceof CompoundVertex) result++;
	    result+= ((CompoundVertex)mv).subgraph.nestedNumOfCompoundVertices();
	}
	return result;
    }

    /** Returns how deep a tree we may descend into if we look inside
        the compound vertices contained in this graph; in other words,
        how many compound vertices are contained inside each other in
        this graph. */
    public int nestingDepth() {
	int result =0;
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex)vertices.get(i);
	    if (mv instanceof CompoundVertex) {
		int dep = ((CompoundVertex)mv).subgraph.nestingDepth()+1;
		if (dep>result) result = dep;
	    }
	}
	return result;
    }

    /** Returns a vector of {@link SimpleVertex}'s all contained in
     * this musicgraph or inside {@link CompoundVertex}'s inside
     * it. This in effect is a flattening of the hierarchical
     * structure, but without edges. */
    public Vector getNestedVertexList() {
	Vector result=new Vector();
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    if (mv instanceof SimpleVertex) result.add(mv);
	    else if (mv instanceof CompoundVertex) {
		CompoundVertex cv = (CompoundVertex) mv;
		result.addAll(cv.subgraph.getNestedVertexList());
	    }
	}
	return result;
    }

    /** Returns true if there are any edges in this graph that connect
        two vertices belonging to different Parts; if there are, this
        is a non-partwise graph, otherwise it is a partwise graph. */
    public boolean hasInterpartEdges() {
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    SimpleVertex sv1 = me.fromAnchor();
	    SimpleVertex sv2 = me.toAnchor();
	    if (sv1.info.thePart!=sv2.info.thePart) return true;
	}
	return false;
    }

    /** Returns true if there is information originating from a
        MuseData file in any vertices in this graph; in particular,
        pitch spelling. */
    public boolean hasMusedata() {
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    if (mv instanceof SimpleVertex) {
		SimpleVertex sv = (SimpleVertex) mv;
		if (sv.info.hasMusedata()) return true;
	    }
	}
	return false;
    }

    /** A deep clone not only makes a clone of the 'vertices' and
        'edges' Vectors of references to {@link MusicVertex} and
        {@link MusicEdge} objects but also clones these objects; the
        resulting graph thus points to objects that are not owned by
        any other graphs and thus must be a MotherGraph.*/
    public MotherGraph deepClone() {
	sortVerticesByStarttime();
	sortEdges();

	Vector edg= new Vector();
	Vector vert = new Vector();
	
	for (int i=0; i<this.vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) this.vertices.get(i);
	    vert.add(mv.deepClone());
	}
	for (int i=0; i<this.edges.size(); i++) {
	    MusicEdge me = (MusicEdge) this.edges.get(i);
	    /*if (!this.vertices.contains(me.from)) {
		System.out.println("me.from not in this.vertices:"+me.from.toString());
		System.out.println("-------Vertex list:");
		for (int k=0; k<this.vertices.size(); k++) {
		    MusicVertex mv = (MusicVertex) this.vertices.get(k);
		    System.out.println(mv.toString());
		    if (!this.vertices.contains(mv)) {
			System.out.println("NOT in this.vertices");
		    }
		}
	    }
	    if (!this.vertices.contains(me.to)) {
		System.out.println("me.to not in this.vertices:"+me.to.toString());
		}*/
	    try {
	    int fromIndex = this.vertices.indexOf(me.from);
	    int toIndex = this.vertices.indexOf(me.to);
	    MusicVertex from = (MusicVertex) vert.get(fromIndex);
	    MusicVertex to = (MusicVertex) vert.get(toIndex);
	    MusicEdge meClone = new MusicEdge(from,to,me.edgetype,
					      me.cloneFromAnchorVector(),
					      me.cloneToAnchorVector());
	    edg.add(meClone);
	    from.outEdges.add(meClone);
	    to.inEdges.add(meClone);
	    } catch(Exception e) {
		System.out.println(e.getMessage());
		e.printStackTrace();
		System.out.println("testIntegrity: "+testIntegrity()+" errors");
		/* GraphGui gg = new GraphGui(getMotherGraph(),400,300);
		try {Thread.sleep(1800000);}
		catch(Exception e2) {System.out.println(e2.getMessage());}*/
		System.exit(0);
	    }
	}
	
	MotherGraph result = new MotherGraph(vert,edg,
					     "deepClone("+title+")",getScore());
	return result;
    }

    /** A method used in connection with the GUI part of the SimFinder
        system; it returns true if all vertices and all edges have
        been assigned a position in the graphical layout. */
    public boolean graphicLayoutDone() {
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    if (!mv.graphicLayoutDone()) {
		System.out.println("No layout: "+mv.toString());
		return false;
	    }
	}
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    if (!me.graphicLayoutDone()) {
		System.out.println("No layout: "+me.toString());
		return false;
	    }
	}
	return true;
    }

    /** Searches for a vertex based on its graphic layout - if it
       clips p, it is returned. */
    public MusicVertex findVertexAt(Point p) {
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    if (p.x>mv.upperLeftPosition.x &&
		p.x<mv.lowerRightPosition.x &&
		p.y>mv.upperLeftPosition.y &&
		p.y<mv.lowerRightPosition.y)
	    {
		return mv;
	    }
	}
	return null;
    }

    /** For debugging. */ 
    public int testIntegrity() {
	int result=0;
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    if (!vertices.contains(me.from)) {
		result++;
		System.out.println("\""+title+"\" says");
		System.out.println("me.from not in vertices:"+me.from.toString());
		System.out.println("-------Vertex list:");
		for (int k=0; k<vertices.size(); k++) {
		    MusicVertex mv = (MusicVertex) vertices.get(k);
		    System.out.println(mv.toString());
		}
	    }
	    if (!vertices.contains(me.to)) {
		result++;
		System.out.println("\""+title+"\" says");
		System.out.println("me.to not in vertices:"+me.to.toString());
	    
		System.out.println("-------Vertex list:");
		for (int k=0; k<vertices.size(); k++) {
		    MusicVertex mv = (MusicVertex) vertices.get(k);
		    System.out.println(mv.toString());
		}
	    }
	}
	return result;
    }

    /** For debugging. */
    public void printAllDist() {
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    System.out.println(mv.toString()+" "+mv.distToString());
	}
    }


}
