/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Vector;
import java.awt.Point;

/** A MusicVertex is one of the two basic components of a {@link
    MusicGraph}, the other being the {@link MusicEdge}. A MusicVertex
    must be either a {@link SimpleVertex} or a {@link CompoundVertex},
    and therefore this class is abstract, describing their common
    features. A SimpleVertex represents a note from a Score, a
    CompoundVertex represents a subgraph. */
public abstract class MusicVertex {

    /** Vector of {@link MusicEdge} going into this MusicVertex. */
    public Vector inEdges;
    /** Vector of {@link MusicEdge} going out of this MusicVertex. */
    public Vector outEdges; // of MusicEdge

    /** The upper left point ofthe bounding box of this MusicVertex
        in the GraphGui canvas. */
    public Point upperLeftPosition = null;
    /** The lower right point of the bounding box of this MusicVertex
        in the GraphGui canvas. */
    public Point lowerRightPosition = null;

    /** A variable used for counting the number of subgraphs this
        MusicVertex has been included in for a given generation of the
        SimFinder.*/
    public int useCounter=0;

    /** The starttime of this MusicVertex.*/
    public double starttime=0.0;

    /** A time stamp which is only set at the creation of this
        MusicVertex object, it is used to construct a total order of
        music vertices, allowing us to unambgiuously sort Vectors of
        music vertices.*/
    public long objectTimeStamp=0;

    // these are distances along both STRONG follow-edges only 

    /** The maximal distance from the first MusicVertex in the mother
        graph, calculated along ONLY strong FOLLOW-edges where each
        edge has a weight of 1.*/
    public int maxDistanceFromStart=0;
    /** The minimal distance from the first MusicVertex in the mother
        graph, calculated along ONLY strong FOLLOW-edges where each
        edge has a weight of 1.*/
    public int minDistanceFromStart=Integer.MAX_VALUE;
    /** The maximal distance to the last MusicVertex in the mother
        graph, calculated along ONLY strong FOLLOW-edges where each
        edge has a weight of 1.*/
    public int maxDistanceFromEnd=0;
    /** The minmal distance to the last MusicVertex in the mother
        graph, calculated along ONLY strong FOLLOW-edges where each
        edge has a weight of 1.*/
    public int minDistanceFromEnd=Integer.MAX_VALUE;

    /** Returns the length of this MusicVertex, which is equal to the
        endtime minus the starttime. */
    public abstract double getLength();

    /** Returns the end time of this MusicVertex. */
    public double getEndtime() {
	return starttime+getLength();
    }

    /** Returns true if this MusicVertex has been assigned a position
        in the graphical layout, otherwise false.*/
    public boolean graphicLayoutDone() {
	if (upperLeftPosition==null) return false;
	if (lowerRightPosition==null) return false;
	return true;
    }

    // beware: only necessarily true for mother graph
    /** Returns true if this MusicVertex is a sink IN THE MOTHERGRAPH. */
    public boolean isSink() {
	for (int i=0; i<outEdges.size(); i++) {
	    MusicEdge me = (MusicEdge) outEdges.get(i);
	    if (me.isStrongFollowEdge()) return false;
	}
	return true;
    }

    // beware: only necessarily true for mother graph
    /** Returns true if this MusicVertex is a source IN THE MOTHERGRAPH. */
    public boolean isSource() {
	for (int i=0; i<inEdges.size(); i++) {
	    MusicEdge me = (MusicEdge) inEdges.get(i);
	    if (me.isStrongFollowEdge()) return false;
	}
	return true;
    }

    /** Picks out the strong FOLLOW edges in the input vector; input
        either inEdges or outEdges as parameter.  Returns a Vector of
        {@link MusicEdge}. */
    public Vector getStrongFollowsEdges(Vector edg) {
	Vector result = new Vector();
	for (int i=0; i<edg.size(); i++) {
	    MusicEdge me = (MusicEdge) edg.get(i);
	    if (me.isStrongFollowEdge())
		result.add(me);
	}
	return result;
    }

    /** Picks out the FOLLOW or WEAKFOLLOW edges in the input vector;
        input either inEdges or outEdges as parameter.  Returns a
        Vector of {@link MusicEdge}. */
    public Vector getFollowsEdges(Vector edg) {
	Vector result = new Vector();
	for (int i=0; i<edg.size(); i++) {
	    MusicEdge me = (MusicEdge) edg.get(i);
	    if (me.isFollowEdge())
		result.add(me);
	}
	return result;
    }

    /** Picks out the strong SIMULTANEOUS edges in the input vector;
        input either inEdges or outEdges as parameter.  Returns a
        Vector of {@link MusicEdge}. */
    public Vector getStrongSimultaneousEdges(Vector edg) {
	Vector result = new Vector();
	for (int i=0; i<edg.size(); i++) {
	    MusicEdge me = (MusicEdge) edg.get(i);
	    if (me.isStrongSimultaneousEdge())
		result.add(me);
	}
	return result;
    }

    /** Picks out the SIMULTANEOUS edges in the input vector; input
        either inEdges or outEdges as parameter. Returns a Vector of
        {@link MusicEdge}. */
    public Vector getSimultaneousEdges(Vector edg) {
	Vector result = new Vector();
	for (int i=0; i<edg.size(); i++) {
	    MusicEdge me = (MusicEdge) edg.get(i);
	    if (me.isSimultaneousEdge())
		result.add(me);
	}
	return result;
    }

    /** Returns a new vector of those inEdges of type FOLLOW or
     * WEAKFOLLOW to this vertex that are also in the edgelist for the
     * subgraph given as input parameter. */
    public Vector getProperFollowInEdges(Subgraph sg) {
	Vector result = getFollowsEdges(inEdges);
	result.retainAll(sg.edges);
	return result;
    }
    /** Returns a new vector of those inEdges of type WEAKFOLLOW to this
     * vertex that are also in the edgelist for the subgraph given as
     * input parameter. */
    public Vector getProperWeakFollowInEdges(Subgraph sg) {
	Vector result = getFollowsEdges(inEdges);
	result.removeAll(getStrongFollowsEdges(inEdges));
	result.retainAll(sg.edges);
	return result;
    }
    /** Returns a new vector of those inEdges of type FOLLOW to this
     * vertex that are also in the edgelist for the subgraph given as
     * input parameter. */
    public Vector getProperStrongFollowInEdges(Subgraph sg) {
	Vector result = getStrongFollowsEdges(inEdges);
	result.retainAll(sg.edges);
	return result;
    }

    /** Returns a new vector of those outEdges of type FOLLOW or
     * WEAKFOLLOW to this vertex that are also in the edgelist for the
     * subgraph given as input parameter */
    public Vector getProperFollowOutEdges(Subgraph sg) {
	Vector result = getFollowsEdges(outEdges);
	result.retainAll(sg.edges);
	return result;
    }
    /** Returns a new vector of those outEdges of type WEAKFOLLOW to
     * this vertex that are also in the edgelist for the subgraph
     * given as input parameter */
    public Vector getProperWeakFollowOutEdges(Subgraph sg) {
	Vector result = getFollowsEdges(outEdges);
	result.removeAll(getStrongFollowsEdges(outEdges));
	result.retainAll(sg.edges);
	return result;
    }
    /** Returns a new vector of those outEdges of type FOLLOW to this
     * vertex that are also in the edgelist for the subgraph given as
     * input parameter */
    public Vector getProperStrongFollowOutEdges(Subgraph sg) {
	Vector result = getStrongFollowsEdges(outEdges);
	result.retainAll(sg.edges);
	return result;
    }

    /** Returns a new vector of those inEdges of type SIMULTANEOUS or
     * WEAKSIMULTANEOUS to this vertex that are also in the edgelist
     * for the subgraph given as input parameter */
    public Vector getProperSimultaneousInEdges(Subgraph sg) {
	Vector result = getSimultaneousEdges(inEdges);
	result.retainAll(sg.edges);
	return result;
    }
    /** Returns a new vector of those inEdges of type WEAKSIMULTANEOUS
     * to this vertex that are also in the edgelist for the subgraph
     * given as input parameter */
    public Vector getProperWeakSimultaneousInEdges(Subgraph sg) {
	Vector result = getSimultaneousEdges(inEdges);
	result.removeAll(getStrongSimultaneousEdges(inEdges));
	result.retainAll(sg.edges);
	return result;
    }
    /** Returns a new vector of those inEdges of type SIMULTANEOUS to
     * this vertex that are also in the edgelist for the subgraph
     * given as input parameter */
    public Vector getProperStrongSimultaneousInEdges(Subgraph sg) {
	Vector result = getStrongSimultaneousEdges(inEdges);
	result.retainAll(sg.edges);
	return result;
    }

    /** Returns a new vector of those outEdges of type SIMULTANEOUS or
     * WEAKSIMULTANEOUS to this vertex that are also in the edgelist
     * for the subgraph given as input parameter */
    public Vector getProperSimultaneousOutEdges(Subgraph sg) {
	Vector result = getSimultaneousEdges(outEdges);
	result.retainAll(sg.edges);
	return result;
    }
    /** Returns a new vector of those outEdges of type
     * WEAKSIMULTANEOUS to this vertex that are also in the edgelist
     * for the subgraph given as input parameter */
    public Vector getProperWeakSimultaneousOutEdges(Subgraph sg) {
	Vector result = getSimultaneousEdges(outEdges);
	result.removeAll(getStrongSimultaneousEdges(outEdges));
	result.retainAll(sg.edges);
	return result;
    }
    /** Returns a new vector of those outEdges of type SIMULTANEOUS to
     * this vertex that are also in the edgelist for the subgraph
     * given as input parameter */
    public Vector getProperStrongSimultaneousOutEdges(Subgraph sg) {
	Vector result = getStrongSimultaneousEdges(outEdges);
	result.retainAll(sg.edges);
	return result;
    }

    /** Used for user interface purposes. */
    public String graphPosToString() {
	int x=0,y=0;
	if (upperLeftPosition!=null) {
	    x=upperLeftPosition.x;
	    y=upperLeftPosition.y;
	}
	String result = "Pos=("+x+","+y+")";
	return result;
    }

    /** Used for user interface purposese. */
    public String distToString() {
	String result = "DIST=[+"+
	    maxDistanceFromStart+",-"+ 
	    minDistanceFromStart+";+"+
	    maxDistanceFromEnd+",-"+
	    minDistanceFromEnd+"]";
	return result;
    }

    /** Resolve the anchor inside this MusicVertex that is pointed to
        by the parameter Vector. */
    public abstract SimpleVertex resolveAnchor(Vector anchor /*of Integer */);

    /** Returns a deep clone of this MusicVertex. */
    public abstract MusicVertex deepClone();

}

