/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Random;
import java.util.Vector;


/** A non-sequential similarity statement is a statement about the
       similarity of two non-sequential subgraphs. */
public class NonSeqSimStmt extends SimilarityStatement {

    public NonSeqSimStmt(NonSeqSubgraph sub1, NonSeqSubgraph sub2,
			 NonSeqSimMeasure sm, SimFinder sf) {
	super(sub1,sub2,sf);
	simMeasure=sm;
    }

    public String toString() {
	String result = "NonSeq"+super.toString();
	return result;
    }

    public String toStringWithSubgraphs() {
	String result = super.toStringWithSubgraphs();
	return result;
    }

    /** Mutates this similarity statement by either substituting one
        subgraph with an entirely new randomly built non-sequential
        subgraph, or extending, shortening or sliding both
        subgraphs.*/
    public void mutate(double mutrate) throws Exception {
	// chance of substituting one subgraph by an entirely new one
	double freshBloodChance = myOwner.freshBloodChance; 

	int times = 1; // (int)(10*mutrate); //1
	int muttimes =1;
	for (int i=0; i<times; i++) {
	    int mutation = rand.nextInt(3);
	    if (myOwner.maxSubSize>0) {
		if (myOwner.maxSubSize<Math.max(subgraph1.size(),subgraph2.size())) {
		    mutation=1;
		} else if (myOwner.maxSubSize==Math.max(subgraph1.size(),subgraph2.size())) {
		    mutation=2;
		}
	    }
	    if(myOwner.occurrenceSearch()) mutation=2;//only slide allowed!
	    
	    if (rand.nextFloat()<freshBloodChance) {
		if (rand.nextFloat()<0.5) {
		    if(!myOwner.occurrenceSearch()){
			subgraph1 = subgraph1.getMotherGraph().getRandomNonSeqSubgraph(subgraph2.size());
		    }
		} else {
		    subgraph2 = subgraph2.getMotherGraph().getRandomNonSeqSubgraph(subgraph1.size());
		}
	    } else { 
		switch (mutation) {
		case 0: // enlarge
		    try {
			extendSubgraphs(muttimes);
		    }
		    catch (Exception e) {
			System.out.println(e.getMessage()); 
			// System.out.println(toStringWithSubgraphs());
			throw e;
		    }
		
		    break;

		case 1: // shrink
		    if (subgraph1.vertices.size()>=2+muttimes) {
			subgraph1.shorten(muttimes);
			subgraph2.shorten(muttimes);
		    } else { 
			/* we need to enlarge more often than shrink, 
			 * otherwise we run aground on size 2 windows 
			 * with perfect match */
			extendSubgraphs(muttimes);
		    }
		    break;
		case 2: // slide
		    subgraph1.slide(muttimes);
		    subgraph2.slide(muttimes);
		    break;
		default:
		}
	    }
	}
	smResult=null;
    }

    /** Make an intelligent subgraph extension. */
    private void extendSubgraphs(int muttimes) throws Exception {
	if (smResult==null) { value(); }
	if (smResult instanceof NSSMVertexResult) {
	    NSSMVertexResult vr = (NSSMVertexResult) smResult;
	    // find a random pair of matched vertices
	    Random r = new Random();
	    int randChoice=r.nextInt(vr.allMatches.size());
	    MusicEdge me = (MusicEdge) vr.allMatches.get(randChoice);
	    if (extendSubgraphsOnceFromMatchedVertices(me.from, me.to)) {
		// System.out.println("extended from matched vertices!");
		return;
	    }
	}
	subgraph1.extend(muttimes); 
	subgraph2.extend(muttimes);
    }

    /** Returns a NonSeqSimStmt with clones of the two subgraphs. */
    public Object clone() {
	NonSeqSimStmt ss = new NonSeqSimStmt((NonSeqSubgraph)subgraph1.clone(),
					     (NonSeqSubgraph)subgraph2.clone(),
					     (NonSeqSimMeasure)simMeasure,
					     myOwner);
	ss.smResult=null;
	return ss;
    }

    /** The method extendSubgraphsOnceFromVertices can be used to
       mutate intelligently from a pair of matched notes instead of
       two notes at random. If this method throws an exception, it is
       from the deep innards of the extend methods in
       NonSeqSubgraph. So don't catch it, let it fly along to be
       caught in the SimFinder. */
    private boolean extendSubgraphsOnceFromMatchedVertices(MusicVertex mv1,
							   MusicVertex mv2) 
	throws Exception {

	float r=rand.nextFloat();
	if (r <0.33) {
	    // try to extend mv1 and mv2 along a follow in-edge
	    if (((NonSeqSubgraph)subgraph1).extend_followIn(mv1)) {
		if (((NonSeqSubgraph)subgraph2).extend_followIn(mv2)) {
		    return true;
		} else {
		    ((NonSeqSubgraph)subgraph2).extend(1);
		}
	    } else {
		if (((NonSeqSubgraph)subgraph1).extend_followOut(mv1)) { 
		    if (((NonSeqSubgraph)subgraph2).extend_followOut(mv2)) {
			return true;
		    } else {
			((NonSeqSubgraph)subgraph2).extend(1);
		    }
		}
		else { 
		    if (((NonSeqSubgraph)subgraph1).extend_simult(mv1)) {
			if (((NonSeqSubgraph)subgraph2).extend_simult(mv2)) {
			    return true;
			} else {
			    ((NonSeqSubgraph)subgraph2).extend(1);
			}
		    } 
		} 
	    }
	} else if (r<0.66) {
	    if (((NonSeqSubgraph)subgraph1).extend_followOut(mv1)) {
		if (((NonSeqSubgraph)subgraph2).extend_followOut(mv2)) {
		    return true;
		} else {
		    ((NonSeqSubgraph)subgraph2).extend(1);
		}
	    } else {
		if (((NonSeqSubgraph)subgraph1).extend_simult(mv1)) { 
		    if (((NonSeqSubgraph)subgraph2).extend_simult(mv2)) {
			return true;
		    } else {
			((NonSeqSubgraph)subgraph2).extend(1);
		    }
		}
		else { 
		    if (((NonSeqSubgraph)subgraph1).extend_followIn(mv1)) {
			if (((NonSeqSubgraph)subgraph2).extend_followIn(mv2)) {
			    return true;
			} else {
			    ((NonSeqSubgraph)subgraph2).extend(1);
			}
		    } 
		} 
	    }
	} else {
	    if (((NonSeqSubgraph)subgraph1).extend_simult(mv1)) {
		if (((NonSeqSubgraph)subgraph2).extend_simult(mv2)) {
		    return true;
		} else {
		    ((NonSeqSubgraph)subgraph2).extend(1);
		}
	    } else {
		if (((NonSeqSubgraph)subgraph1).extend_followIn(mv1)) { 
		    if (((NonSeqSubgraph)subgraph2).extend_followIn(mv2)) {
			return true;
		    } else {
			((NonSeqSubgraph)subgraph2).extend(1);
		    }
		}
		else { 
		    if (((NonSeqSubgraph)subgraph1).extend_followOut(mv1)) {
			if (((NonSeqSubgraph)subgraph2).extend_followOut(mv2)) {
			    return true;
			} else {
			    ((NonSeqSubgraph)subgraph2).extend(1);
			}
		    } 
		} 
	    }
	} // end if
	/* if we have come this far, we never succeeded in extending
           subgraph1. return false*/
	return false;
    }


}
