/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Vector;
import jm.music.data.*;

/** A non-sequential subgraph represents any connected subset of
    vertices inside a mother graph.*/

public class NonSeqSubgraph extends Subgraph {

    /** Neutral constructor for subgraphing. */
    public NonSeqSubgraph(String t, MotherGraph mother) {
	title=t;
	vertices = new Vector();
	edges = new Vector();
	mothergraph=mother;
    }

    /** Constructs a NonSeqSubgraph with only one {@link
        MusicVertex}. */
    public NonSeqSubgraph(MusicVertex mv, MotherGraph mother) {
	title="Subgraph("+mother.title+")";
	vertices = new Vector();
	edges = new Vector();
	mothergraph=mother;
	vertices.add(mv);
    }

    /* WARNING: This method is for non-sequential subgraphs. If
       used on a sequential subgraph, you can't be sure it is
       sequential after this. */
    /** Extends this non-sequential subgraph a number of times. */
    public void extend(int times) throws Exception {
	//	System.out.println("extend begin");
	if (times==0) return;
	int[] randomSeq = makeRandomNsequence(vertices.size());
	for (int i=0; i<randomSeq.length; i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(randomSeq[i]);

	    /* Split the choice of edgekind in 3, because since the
	     * SIMULTANEOUS edges are present both in inEdges and
	     * outEdges, there would be a greater chance of choosing
	     * one of them than a FOLLOW edge if choosing at random
	     * between inEdges and outEdges, not discriminating FOLLOW
	     * and SIMULTANEOUS edges*/
	    float r=rand.nextFloat();
	    if (r <0.33) {
		//System.out.println("extend: follow in");
		if (extend_followIn(mv)) {
		    extend(times-1);
		    return;
		} else {
		    if (extend_followOut(mv)) { extend(times-1); return;}
		    else { if (extend_simult(mv)) {extend(times-1); return;} } 
		}
	    } else if (r<0.66) {
		//System.out.println("extend: follow out");
		if (extend_followOut(mv)) {
		    extend(times-1);
		    return;
		} else {
		    if (extend_simult(mv)) { extend(times-1); return;}
		    else { if (extend_followIn(mv)) {extend(times-1); return;} } 
		}
	    } else {
		//System.out.println("extend: sim");
		if (extend_simult(mv)) {
		    extend(times-1);
		    return;
		} else {
		    if (extend_followIn(mv)) { extend(times-1); return;}
		    else { if (extend_followOut(mv)) {extend(times-1); return;} } 
		}
	    } // end if
	} // end for

	/* if we have come this far without returning, it means that
	 * there are no vertices that we can extend this subgraph from
	 * */
	// System.out.println("extend end");
	throw new Exception("Cannot extend subgraph of size "+size());
    }

    /** Tries to extend this non-sequential subgraph along a FOLLOW
        in-edge of the MusicVertex mv; returns true if the operation
        succeeded, otherwise false. */
    public boolean extend_followIn(MusicVertex mv) throws Exception {
	// try a FOLLOW edge from inEdges
	// System.out.println("x_fIn 1 "+graphIsConnected(true));
	int[] randomEdgeSeq = makeRandomNsequence(mv.inEdges.size());
	for (int j=0; j<mv.inEdges.size(); j++) {
	    MusicEdge me = (MusicEdge) mv.inEdges.get(randomEdgeSeq[j]);
	    if (me.isFollowEdge() && !vertices.contains(me.from)) {
		// we have found a vertex to add
		addVertexAndNecessaryEdges(me.from,me);
		//System.out.println("x_fIn 2 "+graphIsConnected(true));
		return true;
	    }
	}
	return false;
    }

    /** Tries to extend this non-sequential subgraph along a FOLLOW
        out-edge of the MusicVertex mv; returns true if the operation
        succeeded, otherwise false. */
    public boolean extend_followOut(MusicVertex mv) throws Exception {
	// try a FOLLOW edge from outEdges
	// System.out.println("x_fOut 1 "+graphIsConnected(true));
	int[] randomEdgeSeq = makeRandomNsequence(mv.outEdges.size());
	for (int j=0; j<mv.outEdges.size(); j++) {
	    MusicEdge me = (MusicEdge) mv.outEdges.get(randomEdgeSeq[j]);
	    if (me.isFollowEdge() && !vertices.contains(me.to)) {
		// we have found a vertex to add
		addVertexAndNecessaryEdges(me.to,me);
		// System.out.println("x_fOut 2 "+graphIsConnected(true));
		return true;
	    }
	}
	return false;
    }

    /** Tries to extend this non-sequential subgraph along a
        SIMULTANEOUS edge of the MusicVertex mv; returns true if the
        operation succeeded, otherwise false. */
    public boolean extend_simult(MusicVertex mv) throws Exception {
	// Try a SIMULTANEOUS edge. There are SIM inEdges from
	// exactly the same vertices as there are SIM outEdges
	// to, so just pick one side, say inEdges
	// System.out.println("x_sim 1 "+graphIsConnected(true));
	int[] randomEdgeSeq = makeRandomNsequence(mv.inEdges.size());
	for (int j=0; j<mv.inEdges.size(); j++) {
	    MusicEdge me = (MusicEdge) mv.inEdges.get(randomEdgeSeq[j]);
	    if (me.isSimultaneousEdge() && !vertices.contains(me.from)) {
		// we have found a vertex to add
		addVertexAndNecessaryEdges(me.from,me);
		// System.out.println("x_sim 2 "+graphIsConnected(true));
		return true;
	    }
	}
	return false;
    }

    /** Removes a number of {@link MusicVertex}'s from this
        non-sequential subgraph; only vertices that do not split this
        subgraph may be removed. */
    public void shorten(int times) throws Exception {
	if (times==0) return;
	if (size()<times) throw new Exception("Trying to shorten subgraph of size "+size()+" by "+times);

	// go looking for a vertex to remove
	int[] randomSeq = makeRandomNsequence(vertices.size());
	boolean[][] originalClosure=getClosure(true/*include simEdges*/,
					       true/*include weakEdges*/);

	for (int i=0; i<randomSeq.length; i++) {
	    int index = randomSeq[i];

	    if (!vertexAtIndexSplitsSubgraph(index,originalClosure)) {
		// we may safely remove this vertex withou splitting
		// the subgraph
		MusicVertex mv = (MusicVertex) vertices.get(index);
		removeVertexAndConnectedEdges(mv);
		shorten(times-1);
		return;
	    }
	} // end for all randomSeq numbers
 
	// if we have come this far, we didn't find a removable vertex
	throw new Exception("Couldn't find vertex to remove (times="+times+", size="+size()+")");
    }


    /* Yes, yes, this IS a lousy slide function, but go ahead and
       improve it */
    /** It is not as obvious to define sliding for non-sequential
        subgraphs as it is for sequential subgraphs; this
        slide-operation simply calls extend(1) and shorten(1). */
    public void slide(int times) throws Exception {
	for (int i=0; i<times; i++) {
	    extend(1);
	    shorten(1);
	}
    }

    public String toString() {
	String result = "NonSeqSubgraph [title="+title+"]\n";
	result += super.toString();
	return result;
    }

    /** Returns a shallow clone of this NonSeqSubgraph. */
    public Object clone() {
	NonSeqSubgraph cl = new NonSeqSubgraph(title, mothergraph);
	cl.vertices = (Vector) this.vertices.clone();
	cl.edges = (Vector) this.edges.clone();
	return cl;
    }

    /* In a NonSeqSubgraph, the maximal extensibility is not
       necessarily from a source or a sink, so we run through all
       vertices */
    /** Returns how many times this Subgraph may be extended 
        to the left. */
    public int getLeftExtensibility() {
	int maxext=0;
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(0);
	    if (mv.maxDistanceFromStart>maxext) 
		maxext = mv.maxDistanceFromStart;
	}
	return maxext;
    }

    /** Returns how many times this Subgraph may be extended 
        to the right. */
    public int getRightExtensibility() {
	int maxext=0;
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(0);
	    if (mv.maxDistanceFromEnd>maxext) 
		maxext = mv.maxDistanceFromEnd;
	}
	return maxext;
    }

    /** Converts this SeqSubgraph to a Score; includes only notes in
        this subgraph. Assumes that parts are monophonic*/
    public Score convertToScore() {
	sortVerticesByStarttime();
	MusicVertex first = (MusicVertex) vertices.get(0);
	double overallStarttime = first.starttime;
	NonSeqSubgraph[] graphInParts = splitToPartVoices();
	Part[] parts = new Part[graphInParts.length];
	for (int j=0; j<graphInParts.length; j++) {
	    
	    Phrase phr = new Phrase();
	    phr.setAppend(true);
	    if (graphInParts[j].size()>0) {
		MusicVertex mv= (MusicVertex) graphInParts[j].vertices.get(0);
		double partstarttime = mv.starttime;
		phr.setStartTime(partstarttime-overallStarttime);
		
		MusicVertex lastvertex=null;
		for (int i=0; i<graphInParts[j].size(); i++) {
		    mv = (MusicVertex) graphInParts[j].vertices.get(i);
		    if (mv instanceof SimpleVertex) {
			SimpleVertex sv = (SimpleVertex)mv;
			Note n = sv.info.theNote;
			/* Parts are monophonic, so one could think that we
			 * can just add the notes one after the other. But a
			 * non-sequential subgraph does not necessarily
			 * include all consecutive notes of a part in the
			 * interval concerned - i.e. there may be notes that
			 * are left out. These must be added explicitly as
			 * pauses to the Score being constructed here. */
			if (lastvertex!=null) {
			    double gap = sv.starttime-(lastvertex.starttime+lastvertex.getLength());
			    if (gap>0.0) {
				// There is a gap between the last added note
				// and this one. Add it as a rest.
				phr.add(new Note(Integer.MIN_VALUE,gap));
			    }
			}
			phr.add(n.copy());
			lastvertex=sv;
		    }
		}
	    }
	    
	    Part part = new Part(phr);
	    parts[j]=part;
	}
	Score result = new Score(parts, "Score("+title+")");
	return result;
    }


    /* Subgraph SAFE*/
    /** Returns an array of MusicGraphs that are subgraphs of this
       graph, each representing a Part from the originating Score; NB:
       At present, the returned NonSeqSubgraphs may be unconnected
       (i.e. there can be vertices inside them who are not connected
       with the rest of the NonSeqSubgraph).  */
    public NonSeqSubgraph[] splitToPartVoices() {
	Score sc=getScore();
	// System.out.println("score=null"+(sc==null));
	Part[] parts = sc.getPartArray();
	NonSeqSubgraph[] result = new NonSeqSubgraph[parts.length];
	for (int i=0; i<result.length; i++) {
	    result[i] = new NonSeqSubgraph("Part "+i+" of ("+title+")",getMotherGraph());
	}

	Vector nestedVertices = getNestedVertexList();

	/* Distribute vertices to relevant part subgraphs */
	for (int i=0; i<nestedVertices.size(); i++) {
	    SimpleVertex sv = (SimpleVertex) nestedVertices.get(i);
	    for (int j=0; j<parts.length; j++) {
		if (sv.info.thePart==parts[j]) {
		    result[j].vertices.add(sv);
		}
	    }
	}

	/* Distribute edges to relevant subgraphs */
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    for (int j=0; j<parts.length; j++) {
		if (result[j].vertices.contains(me.from) &&
		    result[j].vertices.contains(me.to)) {
		    result[j].edges.add(me);
		}
	    }
	}

	for (int i=0; i<result.length; i++) {
	    result[i].sortVerticesByStarttime();
	}

	return result;
    }


    /* Subgraph safe */
    /** Returns the size of the largest possible sequential subgraph
        of this NonSeqSubgraph. */
    public int getMaximalSequentialSubgraphSize() {
	Vector sons = getStrongSources();
	int count=0;
	while (sons.size()>0) {
	    count++;
	    sons=getAllSons(sons);
	}
	return count;
    }

    /** @deprecated Not used anymore. */
    public double getMothergraphCoverage(){
	return 0.0;
    }

    /* NB: ikke compound safe!*/
    /** @deprecated Not used anymore. */
    public int numberOfLeftmostRests(){
	NonSeqSubgraph copy=(NonSeqSubgraph)this.clone();
	int rests=0;
	boolean cont=true;
	while(cont){
	    cont=false;
	    Vector sources=copy.getStrongSources();
	    for(int i=0;i<sources.size();i++){
		MusicVertex mv = (MusicVertex)sources.get(i);
		if(mv instanceof SimpleVertex){
		    SimpleVertex sv = (SimpleVertex)mv;
		    if(sv.info.theNote.getPitch()==Integer.MIN_VALUE){
			copy.removeVertexAndConnectedEdges(mv);
			rests++;
			cont=true;
		    }
		}
	    }
	}
	return rests;
    }
    
    /* NB: ikke compound safe!*/
    /* tller kun forekomster uden for compounds */
    /** @deprecated Not used anymore. */
    public int numberOfRightmostRests(){
	NonSeqSubgraph copy=(NonSeqSubgraph)this.clone();
	int rests=0;
	boolean cont=true;
	while(cont){
	    cont=false;
	    Vector sinks=copy.getStrongSinks();
	    for(int i=0;i<sinks.size();i++){
		MusicVertex mv=(MusicVertex)sinks.get(i);
		if(mv instanceof SimpleVertex){
		    SimpleVertex sv = (SimpleVertex)mv;
		    if(sv.info.theNote.getPitch()==Integer.MIN_VALUE){
			copy.removeVertexAndConnectedEdges(mv);
			rests++;
			cont=true;
		    }
		}
		//else if(){}
	    }
	}
	return rests;
    }
    
}
