/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Vector;
import jm.music.data.*;

/** A sequential subgraph represents a string of music vertices which
    follow immediately upon each other inside the mother graph of the
    sequential subgraph.*/

public class SeqSubgraph extends Subgraph {

    /** Neutral constructor for subgraphing. */
    public SeqSubgraph(String t, MotherGraph mother) {
	title=t;
	vertices = new Vector();
	edges = new Vector();
	mothergraph=mother;
	// if (mother!=null) theScore= from.theScore;
    }

    /** Constructs a SeqSubgraph with only one {@link MusicVertex} in
        it.*/
    public SeqSubgraph(MusicVertex mv, MotherGraph mother) {
	title="Subgraph("+mother.title+")";
	vertices = new Vector();
	edges = new Vector();
	mothergraph=mother;
	vertices.add(mv);
	// if (from!=null) theScore= from.getScore();
    }

    /** Performs a shallow clone of this sequential subgraph. */
    public Object clone() {
	SeqSubgraph cl = new SeqSubgraph(title, mothergraph);
	cl.vertices = (Vector) this.vertices.clone();
	cl.edges = (Vector) this.edges.clone();
	return cl;
    }

    public String toString() {
	String result = "SeqSubgraph [title="+title+"]\n";
	result += super.toString();
	return result;
    }

    /** Returns the vertices contained in this sequential subgraph in
        an array rather than the Vector 'vertices' in which they are
        contained in the superclass {@link MusicGraph}. */
    public MusicVertex[] getMusicVertexArray(){
	sortVerticesByStarttime();
	Vector nestedVertices = vertices;//contains Simple and Compound vertices
	MusicVertex[] res=new MusicVertex[vertices.size()];
	for(int i=0;i<res.length;i++){
	    res[i]=(MusicVertex)vertices.get(i);
	}
	return res;
    }
    
    /** Returns the edges contained in this sequential subgraph in
        an array rather than the Vector 'edges' in which they are
        contained in the superclass {@link MusicGraph}. */
    public MusicEdge[] getMusicEdgeArray(){
	sortEdges();
	Vector nestedVertices = edges;
	MusicEdge[] res=new MusicEdge[edges.size()];
	for(int i=0;i<res.length;i++){
	    res[i]=(MusicEdge)edges.get(i);
	}
	return res;
    }

    /** Returns the nested noteinfos. */
    public NoteInfo[] getNoteInfoArray() {
	Vector nestedVertices = getNestedVertexList();
	sortVertexVectorByStarttime(nestedVertices);

	NoteInfo [] result = new NoteInfo[nestedVertices.size()];

	for (int i=0; i<nestedVertices.size(); i++) {
	    SimpleVertex sv = (SimpleVertex)nestedVertices.get(i);
	    result[i]=sv.info;
	}
	return result;
    }

    /** Returns an array of the pitch values of notes represented by
        vertices in this sequential subgraph. */
    public int[] getPitchArray() {
	NoteInfo[] ninf=getNoteInfoArray();
	int[] result = new int[ninf.length];

	for (int i=0; i<ninf.length; i++) {
	    result[i]=ninf[i].theNote.getPitch();
	}
	return result;
    }
    
    /** Returns an array of the length of notes represented by
        vertices in this sequential subgraph. */
    public double[] getRhythmArray() {
	NoteInfo[] ninf=getNoteInfoArray();
	double[] result = new double[ninf.length];

	for (int i=0; i<ninf.length; i++) {
	    result[i]=ninf[i].theNote.getRhythmValue();
	}
	return result;
    }

    /** Returns an array of the pitch class values of notes
        represented by vertices in this sequential subgraph. */
    public int[] getPitchClassArray() {
	NoteInfo[] ninf=getNoteInfoArray();
	int[] result = new int[ninf.length];


	for (int i=0; i<ninf.length; i++) {
	    result[i]=ninf[i].theNote.getPitch()%12;
	}
	return result;
    }

    /** Returns arrays of the minimum and maximum distances from and
        to the beginning and end of the MotherGraph of all vertices in
        this sequential subgraph. The first index is in the range
        0<=i<=3, where 0=maxDistanceFromStart, 1=minDistanceFromStart,
        2=maxDistanceFromEnd, 3=minDistanceFromEnd. The second index
        varies over the number of vertices in this subgraph. */
    public int[][] getDistanceArrays() {
	Vector nestedVertices = getNestedVertexList();
	sortVertexVectorByStarttime(nestedVertices);

	int[][] result = new int[4][nestedVertices.size()];
	for (int i=0; i<size(); i++) {
	    MusicVertex mv = (MusicVertex) nestedVertices.get(i);
	    result[0][i] = mv.maxDistanceFromStart;
	    result[1][i] = mv.minDistanceFromStart;
	    result[2][i] = mv.maxDistanceFromEnd;
	    result[3][i] = mv.minDistanceFromEnd;
	}
	return result;
    }

    /** Returns an array of the pitch spelling names of notes
        represented by vertices in this sequential subgraph; NB:
        Should only be used when diatonic data is available */
    public int[] getNameArray(){
	NoteInfo[] n=getNoteInfoArray();
	int[] res=new int[n.length];
	for(int i=0;i<res.length;i++){
	    res[i]=n[i].name;
	}
	return res;
    }

    /** Returns an array of the octaves of notes represented by
        vertices in this sequential subgraph; NB: Should only be used
        when diatonic data is available */
    public int[] getOctaveArray(){
	NoteInfo[] n=getNoteInfoArray();
	int[] res=new int[n.length];
	for(int i=0;i<n.length;i++){
	    res[i]=n[i].octave;
	}
	return res;
    }
    
    /** Slides this sequential subgraph to the left. */
    /* separated from the recursive part to avoid sorting vertices in
     * each reacursive call. */
    public void slideLeft(int times) throws Exception {
    	slideLeft_recursion(times);
	sortVerticesByStarttime();
    }

    /* assume this is a sequential subgraph */
    /* Subgraph SAFE*/
    private void slideLeft_recursion(int times) throws Exception {
	if (times==0) return;
	String BLIND_MSG="blind"; 
	Vector sources = getStrongSources();
	Vector sinks = getStrongSinks();
	if (sources.size()!=1) throw new Exception("Attempted to slide left a subgraph with "+sources.size()+" sources");
	if (sinks.size()!=1) throw new Exception("Attempted to slide left a subgraph with "+sinks.size()+" sinks");

	// mv is the source
	MusicVertex mv = (MusicVertex) sources.get(0);
	if (mv.inEdges.size()==0) throw new Exception(BLIND_MSG);
	else {
	    int[] randomSeq = makeRandomNsequence(mv.inEdges.size());

	    MusicVertex oldsink = (MusicVertex) sinks.get(0);
	    MusicEdge lastEdge=null;
	    // find the in-edge of oldsink that is also in this subgraph
	    boolean done = false;
	    for (int j=0; j<oldsink.inEdges.size() && !done; j++) {
		lastEdge=(MusicEdge) oldsink.inEdges.get(j);
		if (lastEdge.isStrongFollowEdge() && edges.contains(lastEdge)) {
		    done=true;
		}
	    }
	    for (int i=0; i<mv.inEdges.size(); i++) {
		// try another in-edge
		MusicEdge me = (MusicEdge) mv.inEdges.get(randomSeq[i]);

		if (me.isStrongFollowEdge()) {
		    if (me.from.maxDistanceFromStart>=times) {
			// add new leftmost edge and vertex
			vertices.add(me.from);
			edges.add(me);
			// remove old rightmost edge and vertex
			vertices.remove(oldsink);
			edges.remove(lastEdge);
			
			// done sliding subgraph, try recursive call
			// boolean recursiveSuccess=true;
			try { 
			    slideLeft_recursion(times-1);
			    return;
			}
			catch(Exception e) {
			    if (e.getMessage().equals(BLIND_MSG)) {}
				//recursiveSuccess=false; 
			    else throw e; 
			}
			/*
			// if no recursive success, 
			// we should reverse what was done
			// and try another inEdge
			if (recursiveSuccess) break;
			else {
			    vertices.remove(me.from);
			    edges.remove(me);
			    vertices.add(oldsink);
			    edges.add(lastEdge);
			    }*/
		    }
		}	       
		
	    } // end for all inEdges
	} // end elsepart (no dead end found)
    }

    /* separated from the recursive part to avoid sorting vertices in
     * each reacursive call. */
    /** Slides this sequential subgraph to the right. */
    public void slideRight(int times) throws Exception {
    	slideRight_recursion(times);
	sortVerticesByStarttime();
    }

    /* assume this is a sequential subgraph */
    /* Subgraph SAFE*/
    private void slideRight_recursion(int times) throws Exception {
	if (times==0) return;
	String BLIND_MSG="blind"; 
	Vector sources = getStrongSources();
	Vector sinks = getStrongSinks();
	if (sources.size()!=1) throw new Exception("Attempted to slide right a subgraph with "+sources.size()+" sources");
	if (sinks.size()!=1) throw new Exception("Attempted to slide right a subgraph with "+sinks.size()+" sinks");

	MusicVertex mv = (MusicVertex) sinks.get(0);
	if (mv.isSink()) {
	    // System.out.println("Gonna throw blind_msg exception");
	    throw new Exception(BLIND_MSG);
	}
	else {
	    int[] randomSeq = makeRandomNsequence(mv.outEdges.size());

	    MusicVertex oldsource = (MusicVertex) sources.get(0);
	    MusicEdge firstEdge=null;
	    // find the out-edge of oldsource that is also in this subgraph
	    boolean done = false;
	    for (int j=0; j<oldsource.outEdges.size() && !done; j++) {
		firstEdge=(MusicEdge) oldsource.outEdges.get(j);
		if (firstEdge.isStrongFollowEdge() && edges.contains(firstEdge)) {
		    done=true;
		}
	    }
	    for (int i=0; i<mv.outEdges.size(); i++) {
		// try another out-edge
		MusicEdge me = (MusicEdge) mv.outEdges.get(randomSeq[i]);

		if (me.isStrongFollowEdge()) {
		    if (me.to.maxDistanceFromEnd>=times) {
			// add new rightmost edge and vertex
			vertices.add(me.to);
			edges.add(me);
			// remove old rightmost edge and vertex
			vertices.remove(oldsource);
			edges.remove(firstEdge);
			
			// done sliding subgraph, try recursive call
			slideRight_recursion(times-1);
			return;
		   
		    }
		}
	    }
	    // if we have come this far without returning, it is
	    // impossible to slideRight 'times' times.
	    throw new Exception("Exception: cannot slide right");
	}
    }

    /* separated from the recursive part to avoid sorting vertices in
     * each reacursive call. */
    /** Extends this sequential subgraph to the left. */
    public void extendLeft(int times) throws Exception {
    	extendLeft_recursion(times);
	sortVerticesByStarttime();
    }

    /* assume this is a sequential subgraph */
    /* if this subgraph is empty and has a mothergraph, pick random vertex 
     * from mothergraph */
    /* Subgraph SAFE*/
    private void extendLeft_recursion(int times) throws Exception {
	int initSize = size();
	// System.out.println("xLeft");
	if (times==0) return;
	if (vertices.size() == 0) {
	    if (mothergraph!=null) {
		vertices.add(mothergraph.vertices.get(rand.nextInt(mothergraph.vertices.size())));
	    }
	    return;
	}

	String BLIND_MSG="blind"; 
	Vector sources = getStrongSources();
	Vector sinks = getStrongSinks();
	if (sources.size()!=1) throw new Exception("Attempted to extend left a subgraph with "+sources.size()+" sources");
	if (sinks.size()!=1) throw new Exception("Attempted to extend left a subgraph with "+sinks.size()+" sinks");

	MusicVertex mv = (MusicVertex) sources.get(0);
	if (mv.inEdges.size()==0) throw new Exception(BLIND_MSG);
	else {
	    int[] randomSeq = makeRandomNsequence(mv.inEdges.size());

	    for (int i=0; i<mv.inEdges.size(); i++) {
		// try another in-edge
		MusicEdge me = (MusicEdge) mv.inEdges.get(randomSeq[i]);
		if (me.isStrongFollowEdge()) {
		    if (me.from.maxDistanceFromStart>=times-1) {
			// add new leftmost edge and vertex
			vertices.add(me.from);
			edges.add(me);
			
			// done sliding subgraph, try recursive call
			// boolean recursiveSuccess=true;
			try { 
			    extendLeft_recursion(times-1);
			    return;
			}
			catch(Exception e) {
			    //System.out.println("MG.xLeft: "+e.getMessage());
			    if (e.getMessage().equals(BLIND_MSG)) {
				//recursiveSuccess=false; 
			    }
			    else throw e; 
			}
			/* // if no recursive success, we should reverse what was done
			   // and try another inEdge
			   if (recursiveSuccess) break;
			   else {
			   vertices.remove(me.from);
			   edges.remove(me);
			   }*/
		    } else {
			/*			System.out.println("left: maxDist="+me.from.maxDistanceFromStart+" < "+(times-1)+"=times "+i);	
			System.out.println("from="+me.from.toString());
			System.out.println("MV="+mv.toString());*/
		    }
		    
		} else {
		    //System.out.println("simEdge "+i);
		}

	    }
	    if (size()!=initSize+times) {
		System.out.println("ERROR: size="+size()+", desired="+(initSize+times));
		/*
		System.out.println("vertex: "+mv.toString());
		System.out.println(mv.distToString());
		System.out.println("inEdges: "+mv.inEdges.size());
		GraphWindow gw = new GraphWindow(getMotherGraph(),300,400,true);
		gw.setAndShowHighlight(this);
		getMotherGraph().printAllDist();
		try {Thread.sleep(1800000);} catch(Exception e) {
		    System.out.println(e.getMessage());
		    }*/
		throw new Exception("recursive failure");
	    }
	}
    }

    /* separated from the recursive part to avoid sorting vertices in
     * each reacursive call. */
    /** Extends this sequential subgraph to the right. */
    public void extendRight(int times) throws Exception {
    	extendRight_recursion(times);
	
	sortVerticesByStarttime();
    }

    /* assume this is a sequential subgraph */
    /* if this subgraph is empty and has a mothergraph, pick random vertex 
     * from mothergraph */
    /* Subgraph SAFE*/
    private void extendRight_recursion(int times) throws Exception {
	// System.out.println("xRight");
	int initSize = size();
	if (times==0) return;
	if (vertices.size()==0) {
	    if (mothergraph!=null) {
		vertices.add(mothergraph.vertices.get(rand.nextInt(mothergraph.vertices.size())));
	    }
	    return;
	}
	String BLIND_MSG="blind"; 
	Vector sources = getStrongSources();
	Vector sinks = getStrongSinks();
	if (sources.size()!=1) throw new Exception("Attempted to extend right a subgraph with "+sources.size()+" sources");
	if (sinks.size()!=1) throw new Exception("Attempted to extend right a subgraph with "+sinks.size()+" sinks");

	MusicVertex mv = (MusicVertex) sinks.get(0);
	if (mv.outEdges.size()==0) throw new Exception(BLIND_MSG);
	else {
	    int[] randomSeq = makeRandomNsequence(mv.outEdges.size());

	    for (int i=0; i<mv.outEdges.size(); i++) {
		// try another out-edge
		MusicEdge me = (MusicEdge) mv.outEdges.get(randomSeq[i]);
		if (me.isStrongFollowEdge()) {
		    if (me.to.maxDistanceFromEnd>=times-1) {
			// add new rightmost edge and vertex
			vertices.add(me.to);
			edges.add(me);
			
			// done sliding subgraph, try recursive call
			// boolean recursiveSuccess=true;
			try { 
			    extendRight_recursion(times-1);
			    return;
			}
			catch(Exception e) {
			    if (e.getMessage().equals(BLIND_MSG)) {}
				// recursiveSuccess=false; 
			    else throw e;
			}
			/*
			// if no recursive success, we should reverse what was done
			// and try another inEdge
			if (recursiveSuccess) break;
			else {
			    vertices.remove(me.to);
			    edges.remove(me);
			    }*/
		    } else {
			/* System.out.println("right: maxDist="+me.to.maxDistanceFromEnd+" < "+(times-1)+"=times "+i);		
			System.out.println("to="+me.to.toString());
			System.out.println("MV="+mv.toString()); */
		    }

		} else {
		    // System.out.println("simEdge "+i);
		}
	    }
	    if (size()!=initSize+times) {
		System.out.println("size="+size()+", desired="+(initSize+times));
		/*
		System.out.println("vertex: "+mv.toString());
		System.out.println(mv.distToString());
		System.out.println("outEdges: "+mv.outEdges.size());
		GraphWindow gw = new GraphWindow(getMotherGraph(),300,400,true);
		gw.setAndShowHighlight(this);
		getMotherGraph().printAllDist();
		try {Thread.sleep(1800000);} catch(Exception e) {System.out.println(e.getMessage());}
		*/
		throw new Exception("recursive failure");
	    }
	}
    }

    /* assume this is a sequential subgraph */
    /* Subgraph SAFE*/
    /** Shortens this sequential subgraph to the left. */
    public void shortenLeft(int times) throws Exception {
	if (times==0) return;
	if (times==vertices.size()) {
	    vertices.clear();
	    edges.clear();
	    return;
	}
	if (times>vertices.size()) throw new Exception("Can't shortenLeft subgraph of size "+vertices.size()+" by "+times);

	Vector sources = getStrongSources();
	if (sources.size()!=1) throw new Exception("Attempted to shorten left a subgraph with "+sources.size()+" sources");

	MusicVertex oldsource = (MusicVertex) sources.get(0);
	MusicEdge firstEdge=null;
	// find the out-edge of oldsource that is also in this subgraph
	boolean done = false;
	for (int j=0; j<oldsource.outEdges.size() && !done; j++) {
	    firstEdge=(MusicEdge) oldsource.outEdges.get(j);
	    if (firstEdge.isStrongFollowEdge() && edges.contains(firstEdge)) {
		done=true;
	    }
	}
	vertices.remove(oldsource);
	if (firstEdge!=null) edges.remove(firstEdge);

	// done shortening subgraph, try recursive call
	shortenLeft(times-1);
	sortVerticesByStarttime();
    }

    /* assume this is a sequential subgraph */
    /* Subgraph SAFE*/
    /** Shortens this sequential subgraph to the right. */
    public void shortenRight(int times) throws Exception {
	if (times==0) return;
	if (times==vertices.size()) {
	    vertices.clear();
	    edges.clear();
	    return;
	}
	if (times>vertices.size()) throw new Exception("Can't shortenRight subgraph of size "+vertices.size()+" by "+times);

	Vector sinks = getStrongSinks();
	if (sinks.size()!=1) throw new Exception("Attempted to shorten right a subgraph with "+sinks.size()+" sinks");

	MusicVertex oldsink = (MusicVertex) sinks.get(0);
	MusicEdge lastEdge=null;
	// find the out-edge of oldsink that is also in this subgraph
	boolean done = false;
	for (int j=0; j<oldsink.inEdges.size() && !done; j++) {
	    lastEdge=(MusicEdge) oldsink.inEdges.get(j);
	    if (lastEdge.isStrongFollowEdge() && edges.contains(lastEdge)) {
		done=true;
	    }
	}
	vertices.remove(oldsink);
	if (lastEdge!=null) edges.remove(lastEdge);

	// done shortening subgraph, try recursive call
	shortenLeft(times-1);
	sortVerticesByStarttime();
    }

    /** Returns how many times this Subgraph may be extended 
        to the left. */
    public int getLeftExtensibility() {
	Vector sources = getStrongSources();
	if (sources.size()!=1) return 0;
	MusicVertex source = (MusicVertex) sources.get(0);
	return source.maxDistanceFromStart;
    }

    /** Returns how many times this Subgraph may be extended 
        to the right. */
    public int getRightExtensibility() {
	Vector sinks = getStrongSinks();
	if (sinks.size()!=1) return 0;
	MusicVertex sink = (MusicVertex) sinks.get(0);
	return sink.maxDistanceFromEnd;
    }

 
    /* Subgraph SAFE*/
    /** Extends this subgraph by randomly adding to it a {@link
        MusicVertex} which, in the mother graph, is connected to
        either the sink or the source of this sequential subgraph. The
        edge connecting them is also added. */
    public void extend(int times) throws Exception {
	/* check if at all possible to extend to a subgraph of desired
	 * size from here */
	int lex = getLeftExtensibility();
	int rex = getRightExtensibility();

	if (lex+rex>=times) {
	    // yes, possible
	    int left = rand.nextInt(times+1);
	    int right = times-left;
	    while (left>lex) {
		left--;
		right++;
	    }
	    while (right>rex) {
		left++;
		right--;
	    }
	    /* Now we know for certain that it's possible to expand
	     * this subgraph 'left' times to the left from the source
	     * and 'right' times to the right from the sink */
	    try {
		extendRight(right);
		extendLeft(left);
	    } catch(Exception e) {
		/*
		System.out.println("SeqSubgraph extend: exception caught when trying to extend. "+e.getMessage());
		GraphWindow gw=new GraphWindow(getMotherGraph(),200,200,true);
		gw.setAndShowHighlight(this);
		try{
		    Thread.sleep(10000);
		}catch(Exception e2){}
		*/
		throw e;
	    }
	} else {
	    // no, impossible
	    throw new Exception("SeqSubgraph extend cannot possibly extend this graph "+times+" times");
	}
	/*
	//System.out.println("\nrandomExtend");
	int sz=size();
	boolean debug =false;
	if (rand.nextFloat()<0.5) {
	    extendLeft(times);
	    debug=true;
	}
	else {
	    extendRight(times);
	    debug=false; 
	}
	if (sz+times!=size()) {
	    if (debug) System.out.println("Left");
	    else System.out.println("Right");
	    System.out.println(": rEx "+(sz+times)+","+size());
	}
	//System.out.println("randomExtend done");*/ 
    }

    /* Subgraph SAFE*/
    /** Shortens this subgraph by removing either the source or the
        sink of this sequential subgraph and the edge that connected
        it to the rest of this subgraph. */
    public void shorten(int times) throws Exception {
	//System.out.println("\nrandomShorten");
	if (rand.nextFloat()<0.5) {
	    shortenLeft(times);
	}
	else {
	    shortenRight(times);
	}
	//System.out.println("randomShorten done");
    }

    /* Subgraph SAFE */
    /** Slides this sequential subgraph a number of times to the left
        or right. */
    public void slide(int times) throws Exception {
	/* check if at all possible to extend to a subgraph of desired
	 * size from here */
	int lex = getLeftExtensibility();
	int rex = getRightExtensibility();

	if (rand.nextFloat()<0.5) {
	    // prefer left slide
	    if (lex>=times) {
		// yes, possible to slide left
		slideLeft(times);
		return;
	    } else {
		// left slide impossible, try right instead
		if (rex>=times) {
		    slideRight(times);
		} else {
		    // right slide also impossible, throw exception
		    throw new Exception("cannot slide this graph "+times+" times to any side");
		}
	    }
	}
	else {
	    // prefer right slide 
	    if (rex>=times) {
		// yes, possible to slide right
		slideRight(times);
		return;
	    } else {
		// right slide impossible, try left instead
		if (lex>=times) {
		    slideLeft(times);
		} else {
		    // left slide also impossible, throw exception
		    throw new Exception("cannot slide this graph "+times+" times to any side");
		}
	    }
	}
    }
    
    /* Subgraph SAFE*/
    /** Converts this SeqSubgraph to a Score; includes only notes in
        this subgraph.*/
    public Score convertToScore() {
	//	System.out.println("Debug: csts begin");
	Vector nestedVertices=getNestedVertexList();
	sortVertexVectorByStarttime(nestedVertices);

	SimpleVertex sv;
	Phrase phr = new Phrase();
	phr.setAppend(true);

	for (int i=0; i<size(); i++) {
	    sv = (SimpleVertex) nestedVertices.get(i);
	    
	    /*if (i==0) {
		phr.setStartTime(mv.starttime);
		}*/
	    Note n = sv.info.theNote;
	    System.out.println("adding "+n.toString());
	    phr.add(new Note(n.getPitch(),n.getRhythmValue()));
	}

	Part part = new Part(phr);
	Score result = new Score(part, "Score("+title+")");
	/*System.out.println("Debug: csts end");
	  System.out.println(result.toString());*/
	return result;
    }


    /** Simply returns the size of this subgraph, since it is already
        sequential, and its maximal sequential subgraph therefore has
        the same size. */
    public int getMaximalSequentialSubgraphSize() {
	return size();
    }

    /** @deprecated Not used anymore. */
    public double getMothergraphCoverage(){
    
	return 0.0;
    }


    /* NB: not compound safe!*/
    /** @deprecated Not used anymore. */
    public int numberOfLeftmostRests(){
	int i=0;
	while( ((SimpleVertex)vertices.get(i)).info.theNote.getPitch()==Integer.MIN_VALUE ){
	    i++;
	}
	return i;
    }

    /* NB: not compound safe!*/
    /** @deprecated Not used anymore. */
    public int numberOfRightmostRests(){
	int rests=0;
	int i=vertices.size()-1;
	while( ((SimpleVertex)vertices.get(i)).info.theNote.getPitch()==Integer.MIN_VALUE ){
	    rests++;
	    i--;
	}
	return rests;
    }
}
