/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

/** The SeqViewComparator class contains a number of helper methods
    used by sequential view comparators. It is instantiated in the
    SimLibs: {@link SimLibDiffCount} {@link SimLibMeanDiff}. */
public abstract class SeqViewComparator extends ViewComparator {
    
    public SeqViewComparator(String n) { name=n; }

    public ViewDifference compare(SimilarityStatement ss) {
	if (ss instanceof SeqSimStmt) {
	    return seqCompare((SeqSimStmt)ss);
	}
	else { 
	    System.out.println("warning: returning null from SeqViewComparator.getView");
	    return null;
	}
    }

    public abstract ViewDifference seqCompare(SeqSimStmt sss);

    /* helper functions */
    public static int countDifferencesInt(int[] p1, int[] p2) {
	int result=0;
	for (int i=0; i<p1.length; i++) {
	    if (p1[i]!=p2[i]) result++;
	}
	return result;
    }
    public static int countDifferencesDouble(double[] p1, double[] p2) {
	int result=0;
	for (int i=0; i<p1.length; i++) {
	    if (p1[i]!=p2[i]) result++;
	}
	return result;
    }

    /* Contour and interval value assignment assumes that pauses have 
     * same pitch value as preceding note. (a pause is no change in 
     * pitch contour, but the first pitch after the pauses is a contour 
     * change from the last pitch, not from the pause (which HAS no 
     * pitch value)) */
    public static int[] convertPauses(int[] ar) {
	int[] result = new int[ar.length];
	result[0]=ar[0];
	for (int i=1; i<ar.length; i++) {
	    if (ar[i]==Integer.MIN_VALUE) { result[i]=ar[i-1]; }
	    else result[i]=ar[i];
	}
	return result;
    }

    public static double[] getRhythmInterval(double[] p) {
	double[] result = new double[p.length-1];
	for (int i=0; i<p.length-1; i++) {
	    double min = Math.min(p[i],p[i+1]);
	    double max = Math.max(p[i],p[i+1]);
	    int log2=0;
	    while (min<max) {min=min*2; log2++; }
	    result[i]=(double)log2;
	}
	return result;
    }

    public static double[] getRhythmContour(double[] p) {
	double[] result = new double[p.length-1];
	for (int i=0; i<p.length-1; i++) {
	    if (p[i]<p[i+1]) result[i]=1;
	    else if (p[i]==p[i+1]) result[i]=0;
	    else if (p[i]>p[i+1]) result[i]=-1;
	}
	return result;
    }

    public static int[] getPitchIntervalArray(NoteInfo[] infos){
	int[] res=new int[infos.length-1];
	for(int i=0;i<res.length;i++){
	    res[i]=midiJump(infos[i],infos[i+1]);
	}
	return res;	
    }


    // 0:prim, 1:sekund, 2:terts, 3:kvart, 4:kvint, 5:sekst, 6:septim 
    // -1 sekund ned osv.
    public static int[] getDiatonicIntervalArray(int[] nameArray, int[] octaveArray){
	int[] res=new int[nameArray.length-1];
	for(int i=0;i<nameArray.length-1;i++){
	    int firstName=nameArray[i];
	    int nextName=nameArray[i+1];
	    int firstOct=octaveArray[i];
	    int nextOct=octaveArray[i+1];
	    //hvis den ene musik-obj. er en pause indsttes Integer.MIN_VALUE
	    if(firstName==-1 || nextName==-1){
		res[i]=Integer.MIN_VALUE;
	    }else{
		res[i]=(nextName+7*nextOct)-(firstName+7*firstOct);
	    }
	}
	return res;	
    }

    // 0:prim, 1:sekund, 2:terts, 3:kvart, 4:kvint, 5:sekst, 6:septim
    // -1:pause til node, -2:node til pause, -3:pause til pause
    public static int[] getDiatonicForwardArray(int[] nameArray){
	int[] res=new int[nameArray.length-1];
	for(int i=0;i<nameArray.length-1;i++){
	    int firstName=nameArray[i];
	    int nextName=nameArray[i+1];
	    if(firstName==-1 && nextName==-1){
		res[i]=-3;//pause til pause
	    }else if(firstName==-1){
		res[i]=-1;//pause til node
	    }else if(nextName==-1){
		res[i]=-2;//node til pause
	    }else{
		res[i]=(7+firstName-nextName)%7;
	    }
	}
	return res;	
    }

    public static int getDiatonicForwardInterval(MusicEdge me) {
	SimpleVertex sv1 = me.fromAnchor();
	SimpleVertex sv2 = me.toAnchor();
	
	int jump=SeqViewComparator.getDiatonicForwardInterval(sv1.info.name, sv2.info.name);
	return jump;
    }

    public static int getDiatonicInterval(MusicEdge me) {
	SimpleVertex sv1 = me.fromAnchor();
	SimpleVertex sv2 = me.toAnchor();
	
	int jump=SeqViewComparator.diatonicInterval(sv1.info, sv2.info);
	return jump;
    }



    /* udregner dfi mellem 2 givne names */
    public static int getDiatonicForwardInterval(int firstName, int nextName){
	int jump=Integer.MAX_VALUE;
	
	if(firstName==-1 && nextName==-1){
	    jump=-3;//pause til pause
	}else if(firstName==-1){
	    jump=-1;//pause til node
	}else if(nextName==-1){
	    jump=-2;//node til pause
	}else{
	    jump=(7+firstName-nextName)%7;//node til node
	}
	return jump;
    }
    /* udregner midiJump mellem 2 NoteInfo*/
    public static int midiJump(NoteInfo first, NoteInfo last){
	int res=0;
	int firstmidi=first.theNote.getPitch();
	int nextmidi=last.theNote.getPitch();
	if(first.isRest() && last.isRest()){//2 pauser
	    res=-10000;
	}else if(first.isRest()){//pause-node
	    res=-20000;
	}else if(last.isRest()){//node-pause
	    res=-30000;
	}else{//node-node
	    res=nextmidi-firstmidi;
	}
	return res;
    }

    public static int diatonicInterval(NoteInfo first, NoteInfo next){
	int res=0;
	//hvis den ene musik-obj. er en pause indsttes Integer.MIN_VALUE
	if(first.isRest() && next.isRest()){
	    res=-10000;
	}else if(first.isRest()){//pause-node
	    res=-20000;
	}else if(next.isRest()){//node-pause
	    res=-30000;
	}else{
	    res=(next.name+7*next.octave)-(first.name+7*first.octave);
	}
	return res;
    }

    
    
}



