/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.io.*;
import java.util.*;

/** The SimFinder is the GA that we use to search for similarities in
    {@link MusicGraph}s. This is where all the loose ends come
    together. */

public class SimFinder {

    /** Returns true if it is doing an occurrence search.
     */
    public boolean occurrenceSearch(){
	return occurrence!=null;
    }

    final Subgraph occurrence;
    final String POP_FILL_ERROR="Could not fill population";
    /** Creates a SimFinder from a MIDI file, with some hardcoded parameters.*/
    public static void main(String[] args) {
	if (args.length<2) {
	    System.out.println("Usage: java SimFinder <midifilename> <status output filename>");
	    System.exit(0);
	}
	try{
	    SimFinder sf = new SimFinder(MotherGraph.getMomFromFile(args[0],true,true,false),
					 50, // popsize
					 3, // initsize
					 0.0, // crossover
					 0.3, // mutation
					 1, // reportFrequency
					 SimLibMeanDiff.seqMD_APAL,// simMeasure
					 null, // occurrence subgraph
					 0, // maxsize
					 0.2,// freshBloodChance
					 0.9//groupBaseGamma
					 );
	    sf.run(30,args[1]);
	    sf.showPopulation();
	}catch (Exception e){System.out.println("SimFinder: Could not initialise");}
    }

    /** A Vector of {@link SimilarityStatement}s. */
    Vector population = new Vector(); // of SimilarityStatement 
    /** A Vector of {@link StatusInfo}s.*/
    Vector statusHistory = new Vector(); // of StatusInfo
    /** The population size.*/
    public int popsize;
    /** The MotherGraph being search for similarities. */
    public MotherGraph piece=null;
    /** Initial size of random subgraphs. */
    public int initSubSize=3; // initial size of random subgraphs
    /** The upper limit on the size of subgraphs inside similarity
        statements; if 0, there is no upper limit. */
    public int maxSubSize=0; // if 0 there is no upper limit on subgraph sizes
    public double crossoverrate;
    public double mutationrate;
    /** The probability that a mutation of a similarity statement will
        result in a total replacement of one of its subgraphs. */
    public double freshBloodChance;
    Random rand = new Random();
    /** How many generations should pass in the GA before a new
        statusreport is printed on screen. */
    public int reportFrequency;
    /** The similarity measure used to evaluate similarity statements
        in the population. */
    public SimilarityMeasure simMeasure;
    boolean saveTemporaryMidifiles=false;
    SimilarityStatement theBest=null;//bruges af STM i StartMusedataSimFinder
    int totalGenerations=0;// do

    /** A Vector of Subgraphs which we do not want to find again.*/
    public Vector forbiddenSubgraphs=new Vector();
    /** A parameter for the grouping viewpoint. */
    public double groupBaseGamma=-1.0;
    
    /** Returns true, if the population is sequential.
     */
    public boolean seqPop() {
	if (simMeasure==null) return true;
	if (simMeasure instanceof SeqSimMeasure) return true;
	if (simMeasure instanceof NonSeqSimMeasure) return false;
	return true;
    }
 
    /** Returns the number of individuals in the next generation to be
	found by crossover. */
    public int numberPickedByCrossover() {
	int res;
	if (occurrenceSearch()){
	    res=0;// ingen co nr der sges efter forekomster
	} else {
	    res =(int)(popsize*crossoverrate);
	}
	return res;
    }

    /** Returns the number of individuals in the next generation to be
	found by tournament selection. */
    public int numberPickedByTournament() {
	return popsize-numberPickedByMutation()-numberPickedByCrossover();
    }

    /** Returns the number of individuals in the next generation to be
	found by mutation. */
    public int numberPickedByMutation() {
	return (int) (mutationrate*popsize);
    }

    /** Constructs a Simfinder running on a MIDI file. Parameters are
	set for the GA. The {@link MotherGraph} is constructed from a
	MIDI file name.
    */
    public SimFinder(int psize, String fname, int isubsize, double co,
		     double mut, int repFreq, SimilarityMeasure sm, 
		     Subgraph occ, int mxsubsize, double freshBlood) throws Exception{
	if (co+mut>1.0) { 
	    System.out.println("Crossover rate + Mutation rate must be < 1.0");
	    System.exit(0);
	}
	popsize = psize;
	piece = new MotherGraph(fname,true
			       /* CHANGE HERE TO TOGGLE INTERPART EDGES 
				* true = include interpart edges
				* false= exclude interpart edges
				*/);
	maxSubSize=mxsubsize;
	/* maxSubSize must be greater than 2, otherwise it is void */
	if (maxSubSize>2 && maxSubSize<isubsize) isubsize=maxSubSize;
	occurrence=occ;
	if(occurrenceSearch()){
	    initSubSize=occ.size();
	}else{
	    initSubSize=isubsize;
	}
	freshBloodChance=freshBlood;
	crossoverrate = co;
	mutationrate=mut;
	reportFrequency=repFreq;
	simMeasure=sm;

	if (seqPop()) {
	    //try{
		initPopulationSequential();
		//}catch (Exception e){}
	    
	} else {
	    initPopulationNonSequential();
	}
    }

    private void initPopulationSequential() throws Exception{
	for (int i=0; i<popsize; i++) {
	    try {
		SeqSubgraph sub1;
		if(occurrenceSearch()){
		    sub1=(SeqSubgraph)occurrence;
		} else {
		    sub1 = piece.getRandomSeqSubgraph(initSubSize);
		}
		SeqSubgraph sub2 = piece.getRandomSeqSubgraph(initSubSize);
		SeqSimStmt stmt = new SeqSimStmt(sub1,sub2,(SeqSimMeasure)simMeasure,this);
		population.add(stmt);
	    } catch (Exception couldNotInitialise){
		System.out.println("Could not initialise with size "+initSubSize);
		throw new Exception();
		/* the population is then empty */ 
	    }
	}
    }

    private void initPopulationNonSequential() {
	for (int i=0; i<popsize; i++) {
	    try {
		NonSeqSubgraph sub1=null;
		NonSeqSubgraph sub2=null; 
		if(occurrenceSearch()){
		    sub1=(NonSeqSubgraph)occurrence;
		} else {
		    while (sub1==null) {
			sub1= piece.getRandomNonSeqSubgraph(initSubSize);
		    }
		}
		while (sub2==null) {
		    sub2= piece.getRandomNonSeqSubgraph(initSubSize);
		}
		NonSeqSimStmt stmt = new NonSeqSimStmt(sub1,sub2,
						       (NonSeqSimMeasure)simMeasure,this);
		population.add(stmt);
	    } catch (Exception e){
		System.out.println("Could not initialise with size "+initSubSize);
	    }
	}
    }
    
    /** Constructs a Simfinder from any {@link
	MotherGraph}. Parameters are set for the GA.
     @param  mother the {@link MotherGraph}.
     @param psize the GA-population size.
     @param isubsize the initial size of the subgraphs in the population.
     @param co the GA-crossover rate.
     @param  mut the GA-mutation rate. 
     @param repFreq the report frequency. Info is printed for every frequency generarions.
     @param sm the {@link SimilarityMeasure} to be used in this SimFinder.
     @param occ the graph to be found more occurrences of.
     @param mxsubsiz A maximal size can be imposed on the subgraphs in
     the population. If it is 0 there is no limitations.
     @param freshBlood the GA-freshblood chance.
     @param groupBase the grouping structure value.

     @throws Exception throws an exception if it is not able to make an
     initialisation of new SimStmts with the specified isubsize.
    */
    public SimFinder(MotherGraph mother, int psize, int isubsize, 
		     double co, double mut, int repFreq, 
		     SimilarityMeasure sm, Subgraph occ, int mxsubsize,
		     double freshBlood, double groupBase) throws Exception{
	if (co+mut>1.0) { 
	    System.out.println("Crossover rate + Mutation rate must be < 1.0");
	    System.exit(0);
	}
	simMeasure=sm;
	/*
	MusedataTokenizer mdt = new MusedataTokenizer(filename);
	Vector parts = mdt.getGraph(combineTies,combineRests);

	System.out.println("Number of parts: "+parts.size());
	
	for(int i=0;i<parts.size();i++){
	    Vector pa=(Vector)parts.get(i);
	    System.out.println("part"+i+".size():"+pa.size());
	}
	*/
	popsize = psize;
	piece = mother;	
	occurrence=occ;
	maxSubSize=mxsubsize;
	if (maxSubSize>2 && maxSubSize<isubsize) isubsize=maxSubSize;
	if(occurrenceSearch()){
	    initSubSize=occ.size();
	}else{
	    initSubSize=isubsize;
	}
	freshBloodChance=freshBlood;
	crossoverrate = co;
	mutationrate=mut;
	reportFrequency=repFreq;
	groupBaseGamma=groupBase;
	
	if (seqPop()) {
	    System.out.println("The population is Sequential");
	    initPopulationSequential();
	} else {
	    System.out.println("The population is NonSequential");
	    initPopulationNonSequential();
	}
    }

    /** Runs the Simfinder for a number of generations. Output is
	written to a file.
	
	@param generations the number of generations
	@param statusFilename the output file
	@throws Exception if the SimFinder is not properly initialised (popSize is 0)
	
    */
    public void run(int generations, String statusFilename) throws Exception{
	if(population.size()==0) {
	    //if initialization failed
	    throw new Exception("SimFinder.run(): Population size=0");
	}
	long loopTime=-1;
	try {
	    FileWriter outfile=new FileWriter(statusFilename,true);
	    outfile.write("# SimFinder output file created at "+new Date().toString()+"\n\n");
	    // we just need one instance of SimilarityStatement to
	    // print the gnuplotHeader -- pick the best one (might
	    // have chosen any other)
	    SimilarityStatement someSimStmt = findBest();
	    outfile.write(new StatusInfo(someSimStmt).gnuplotHeader()+"\n\n");
	    for (int i=0; i<generations; i++) {	
		totalGenerations++;
		long startTime=(new Date()).getTime();

		countUsageInfo();
		// time begin
		// System.out.print("S");
		if (totalGenerations%reportFrequency==0 &&totalGenerations!=0) {
		    //System.out.println("Total generations:"+generations);//eller total
		    StatusInfo inf = collectStatusInfo(totalGenerations);
		    System.out.println(inf.toScreenString());
		    printMemStatus();
		    statusHistory.add(inf);
		    outfile.write(inf.toGnuplotString()+"\n");
		    outfile.flush();

		    // appendVertexUseStatsToFile("/users/elmer/prog/speciale/similarity/data/vUseStats.data");
		} else System.out.print(".");
		
		Vector nextPop = new Vector(); // of SimilarityStatement
		//System.out.println("B="+countUnconnectedGraphs(nextPop));

		// best always survives -- elitism!
		SimilarityStatement best=(SimilarityStatement)findBest().clone();
		theBest=(SimilarityStatement)best.clone();
		nextPop.add(best);
		
		SimilarityStatement write=(SimilarityStatement)best.clone();
		if(saveTemporaryMidifiles && totalGenerations%reportFrequency==0 &&totalGenerations!=0){
		    write.subgraph1.printTheScore("graph_1.mid");
		    write.subgraph2.printTheScore("graph_2.mid");
		}
		
		//System.out.println("T="+countUnconnectedGraphs(nextPop));
		// tournament selection
		int num = numberPickedByTournament();
		for (int j=0; j<num-1; j++) {
		    nextPop.add(tournament());
		}
			
		//System.out.println("C="+countUnconnectedGraphs(nextPop));
		// crossover
		num = numberPickedByCrossover();
		for (int j=0; j<num; j++) {
		    crossover(nextPop);
		}
		
		//System.out.println("m="+countUnconnectedGraphs(nextPop));
		// mutation
		num = numberPickedByMutation();
		for (int j=0; j<num; j++) {
		    mutation(nextPop);
		}
		//System.out.println("Z="+countUnconnectedGraphs(nextPop));
		
		while (nextPop.size()<population.size()) {
		    //System.out.println("POP ADDITION TO KEEP SIZE");
		    try{
			if (seqPop()) {
			    SeqSubgraph s1;
			    if(occurrenceSearch()) {
				s1=(SeqSubgraph)occurrence;
			    }else{
				s1=piece.getRandomSeqSubgraph(initSubSize);
			}
			    nextPop.add(new SeqSimStmt(s1,piece.getRandomSeqSubgraph(initSubSize),(SeqSimMeasure)simMeasure,this));
			} else {
			    NonSeqSubgraph s1;
			    if(occurrenceSearch()) {
			    s1=(NonSeqSubgraph)occurrence;
			    }else{
				s1=piece.getRandomNonSeqSubgraph(initSubSize);
			    }
			    nextPop.add(new NonSeqSimStmt(s1,piece.getRandomNonSeqSubgraph(initSubSize),(NonSeqSimMeasure)simMeasure,this));		    }
		    }catch (Exception init){
			throw new Exception(POP_FILL_ERROR);
		    }
		}
		population = nextPop;
		//System.out.print("M="+countUnconnectedGraphs(nextPop));
		mutatePercentageOfPop(mutationrate);
		long endTime=(new Date()).getTime();
		loopTime=endTime-startTime;
		if (totalGenerations%reportFrequency==0 &&totalGenerations!=0) {
		    System.out.println("The calculation time for generation "+totalGenerations+" was: "+(loopTime/1000)+" seconds");
		}
	    }
	    outfile.flush();
	    outfile.close();
	    System.out.println("Status history written to file "+statusFilename);
	}
	catch (Exception e) {
	    e.printStackTrace();
	    if(e.getMessage().equals(POP_FILL_ERROR)){
		throw e;
	    }else{
		System.out.println("Exception caught in SimFinder's main run-loop:\n"+e.getMessage());
		e.printStackTrace();
	    }
	}
    }
    
    /** Prints the memory status of the computer.
     */
    public void printMemStatus() {
	Runtime r = Runtime.getRuntime();
	long total = r.totalMemory();
	long free = r.freeMemory();
	System.out.println("MEM FREE: "+free+" of "+total);
    }

    public void writeHistoryToFile(String statusFilename) {
	try {
	    FileWriter outfile=new FileWriter(statusFilename,false);
	    outfile.write("# SimFinder output file created at "+new Date().toString()+"\n\n");
	    outfile.write(new StatusInfo().gnuplotHeader()+"\n\n");

	    for (int i=0; i<statusHistory.size(); i++) {
		StatusInfo inf = (StatusInfo) statusHistory.get(i);
		outfile.write(inf.toGnuplotString()+"\n");
	    }
	    outfile.flush();
	    outfile.close();
	    System.out.println("Status history written to file "+statusFilename);
	}
	catch (Exception e) {
	    System.out.println("Exception while writing status history to file:\n"+e.getMessage()); 
	}
   }

    private int countUnconnectedGraphs(Vector p) {
	int result=0;
	for (int i=0; i<p.size(); i++) {
	    SimilarityStatement ss = (SimilarityStatement) p.get(i);
	    if (!ss.subgraph1.graphIsConnected(true /* include sim */,
					       !seqPop()/*include weak */)) result++;
	    if (!ss.subgraph2.graphIsConnected(true /* include sim */,
					       !seqPop()/*include weak */)) result++;
	}
	return result;
    }
    /** Returns the individual in the population with the best fitness. */
    public SimilarityStatement findBest() {
	double bestfit = Double.MAX_VALUE;
	SimilarityStatement result = null;
	for (int i=0; i<population.size(); i++) {
	    SimilarityStatement ss = (SimilarityStatement) population.get(i);
	    if (ss.value()<=bestfit) {
		bestfit=ss.value();
		result = ss;
	    }
	}
	return result;
    }

    /** Opens two {@link GraphWindow}'s showing the two subgraphs of
        the best {@link SimilarityStatement}. */
    public void showBest() {
	SimilarityStatement best = findBest();
	System.out.println("THE BEST:");
	System.out.println(best.toString());
	GraphWindow gui1 = new GraphWindow(piece,100,150,seqPop());
	gui1.setAndShowHighlight(best.subgraph1);
	GraphWindow gui2 = new GraphWindow(piece,150,600,seqPop());
	gui2.setAndShowHighlight(best.subgraph2);

	if (best.identicalSubgraphs()) System.out.println("best has id sub");
    }

    /** Opens a {@link GraphList}'s showing the subgraphs of
        the population. */
    public void showPopulation() {
	SimilarityStatement best = findBest();
	boolean seqBool=seqPop();
	GraphList gl = new GraphList(population,25,200,best,seqBool);
    }
    /*
    public void showPopulation(Vector founds) {
	boolean seqBool=seqPop();
	GraphList gl = new GraphList(founds,25,200,(SimilarityStatement)founds.get(0),seqBool);
    }
    */

    /** Returns a {@link StatusInfo} object describing this generation. */
    public StatusInfo collectStatusInfo(int gen) {
	StatusInfo result= new StatusInfo();
	    

	result.generation = gen;
	result.popSize = population.size();
	
	double low=Double.MAX_VALUE,high=Double.MIN_VALUE;
	int lowIndex=-1; int highIndex=-1;
	double mean=0,stddev=0;
	// int sizeProblems=0;
	// int zerosize=0;
	int longest=0;
	double meansize=0;
	int maxoverlap=0, minoverlap=Integer.MAX_VALUE;;
	double meanoverlap=0;
	
	// SimilarityStatement tempDebug=null;
	for (int i=0; i<population.size(); i++) {
	    SimilarityStatement ss=(SimilarityStatement) population.get(i);
	    double x=ss.value();
	    if (x<low) {low=x; lowIndex=i;}
	    if (x>high){high=x;highIndex=i;}
	    mean+=x;
	    /*if (!ss.oksize()) {
	      sizeProblems++;
	      tempDebug=ss;
	      }*/
	    int siz = ss.size();
	    if (siz>longest) longest = siz;
	    meansize+=siz;
	    // if (ss.subgraph1.size()==0) zerosize++;
	    // if (ss.subgraph2.size()==0) zerosize++;
	    
	    int olap = ss.overlap();
	    
	    //check overlap med found-vektoren

	    if (olap>maxoverlap) maxoverlap=olap;
	    if (olap<minoverlap) minoverlap=olap;
	    meanoverlap+=olap;
	}
	meanoverlap=meanoverlap/population.size();
	meansize=meansize/population.size();
	mean = mean/population.size();
	for (int i=0; i<population.size(); i++) {
	    SimilarityStatement ss=(SimilarityStatement) population.get(i);
	    double x=ss.value();
	    stddev+=Math.pow(mean-x,2);
	}
	stddev = Math.sqrt(stddev/population.size());
	
	SimilarityStatement best = findBest();
	
	result.bestCopy = (SimilarityStatement) best.clone();
	result.bestFitness = low;
	result.worstFitness = high;
	result.meanFitness = mean;
	result.stddevFitness = stddev;
	
	result.bestSize = best.size();
	result.worstSize = ((SimilarityStatement)population.get(highIndex)).size();
	result.meanSize = meansize;
	result.longestLength = longest;
	result.minOverlap = minoverlap;
	result.maxOverlap = maxoverlap;
	result.meanOverlap = meanoverlap;
	
	int vcov = vertexCoverage();
	int ecov = edgeCoverage();
	double pieceSize=piece.size();
	double pieceEdges = piece.edges.size();
	
	result.vertexCoverage = vcov;
	result.vertexCovPercent = vcov/pieceSize;
	result.edgeCoverage = ecov;
	result.edgeCovPercent = ecov/pieceEdges;
	
	return result;
    }

    public void deprecatedStatusInfo(int gen) {
	System.out.println("STATUSINFO generation "+gen);
	System.out.println("\tpop size="+population.size());
	double low=Double.MAX_VALUE,high=Double.MIN_VALUE;
	int lowIndex=-1; int highIndex=-1;
	double mean=0,stddev=0;
	int sizeProblems=0;
	int zerosize=0;
	int longest=0;
	double meansize=0;
	int maxoverlap=0, minoverlap=Integer.MAX_VALUE;;
	double meanoverlap=0;

	SimilarityStatement tempDebug=null;
	for (int i=0; i<population.size(); i++) {
	    SimilarityStatement ss=(SimilarityStatement) population.get(i);
	    double x=ss.value();
	    if (x<low) {low=x; lowIndex=i;}
	    if (x>high){high=x;highIndex=i;}
	    mean+=x;
	    if (!ss.oksize()) {
		sizeProblems++;
		tempDebug=ss;
	    }
	    int siz = ss.size();
	    if (siz>longest) longest = siz;
	    meansize+=siz;
	    if (ss.subgraph1.size()==0) zerosize++;
	    if (ss.subgraph2.size()==0) zerosize++;
	    
	    int olap = ss.overlap();
	    if (olap>maxoverlap) maxoverlap=olap;
	    if (olap<minoverlap) minoverlap=olap;
	    meanoverlap+=olap;
	}
	meanoverlap=meanoverlap/population.size();
	meansize=meansize/population.size();
	mean = mean/population.size();
	for (int i=0; i<population.size(); i++) {
	    SimilarityStatement ss=(SimilarityStatement) population.get(i);
	    double x=ss.value();
	    stddev+=Math.pow(mean-x,2);
	}
	stddev = Math.sqrt(stddev/population.size());

	SimilarityStatement best = findBest();
	// System.out.println("\tBest fitness: "+((SimilarityStatement)population.get(lowIndex)).toString());
	System.out.println("\tBest fitness: "+low+" size="+(best.size()));
	System.out.println("\tWorst fitness: "+high+" size="+((SimilarityStatement)population.get(highIndex)).size());
	System.out.println("\tMean fitness: "+mean);
	System.out.println("\tFitness std.dev.:"+stddev);

	System.out.println("\tMean size: "+meansize);
	System.out.println("\tLongest: "+longest);
	//System.out.println("\tSize problems:"+sizeProblems);
	//System.out.println("\tZero size:"+zerosize);

	System.out.println("\tMin overlap="+minoverlap);
	System.out.println("\tMax overlap="+maxoverlap);
	System.out.println("\tMean overlap="+meanoverlap);

	int vcov = vertexCoverage();
	int ecov = edgeCoverage();
	double pieceSize=piece.size();
	double pieceEdges = piece.edges.size();

	System.out.println("\tVertex coverage = "+vcov+" ("+vcov/pieceSize+")");
	System.out.println("\tEdge coverage = "+ecov+" ("+ecov/pieceEdges+")");

	if (sizeProblems>0) {
	    System.out.println(tempDebug.toString());
	    System.out.println("sizeProblems>0");
	    System.exit(0);
	}
    }

    /** Returns the number of {@link SimilarityStatement}s which have
        identical {@link Subgraph}s. */
    public int countIdSub(Vector pop) {
	int result=0;
	for (int i=0; i<pop.size(); i++) {
	    SimilarityStatement ss = (SimilarityStatement) pop.get(i);
	    if (ss.identicalSubgraphs()) result++;
	}
	return result;
    }

    /** Adds a new {@link SimilarityStatement}s to the population of
        the next generation. Crossover is done by randomly selecting
        two {@link SimilarityStatement}s in the population and then
        selecting randomly one {@link Subgraph} from each {@link
        SimilarityStatement}. The selected {@link Subgraph}s are put
        into a new {@link SimilarityStatement} which is added to the
        population in the next generation. */
    public void crossover(Vector next) {
	int inx1 = rand.nextInt(next.size());
	int inx2 = inx1;
	while (inx2==inx1) {
	    inx2 = rand.nextInt(next.size());
	}
	SimilarityStatement ss1 = (SimilarityStatement) next.get(inx1);
	SimilarityStatement ss2 = (SimilarityStatement) next.get(inx2);
	Subgraph sub1;
	Subgraph sub2;
	if (rand.nextFloat()>0.5) sub1 = (Subgraph)ss1.subgraph1.clone();
	else sub1 = (Subgraph)ss1.subgraph2.clone();
	if (rand.nextFloat()>0.5) sub2 = (Subgraph)ss2.subgraph1.clone();
	else sub2 = (Subgraph)ss2.subgraph2.clone();

	if (sub1.equals(sub2)) {
	    // crossover(next);
	    return;
	}

	SimilarityStatement ss3=null;
	if (seqPop()) {
	    ss3= new SeqSimStmt((SeqSubgraph)sub1,(SeqSubgraph)sub2,(SeqSimMeasure)simMeasure,this);
	} else {
	    ss3= new NonSeqSimStmt((NonSeqSubgraph)sub1,(NonSeqSubgraph)sub2,(NonSeqSimMeasure)simMeasure,this);
	}
	
	try {
	    /* ensure subgraphs have same size */
	    int size1 = ss3.subgraph1.size();
	    int size2 = ss3.subgraph2.size();
	    if (size1>size2) {
		ss3.subgraph2.extend(size1-size2);
	    }
	    else if (size1<size2) {
		ss3.subgraph1.extend(size2-size1);
	    }
	    /* now sizes match */
	    if (!ss3.oksize()) {
		System.out.println("Crossover 1st error:"+size1+","+size2+" to "+ss3.subgraph1.size()+","+ss3.subgraph2.size());
	    }
	    ss3.mutate(mutationrate);
	    if (!ss3.oksize()) {
		System.out.println("Crossover 2nd error:"+ss3.subgraph1.size()+","+ss3.subgraph2.size());
		//System.exit(0);
		// crossover(next);
		return;
	    }
	    if (ss3.identicalSubgraphs()) {
		// crossover(next);
		return;
	    }
	    next.add(ss3);
	}
	catch (Exception e) {
	    // something went awry during mutation so ss3 is possibly unstable
	    // abandon it.
	    // System.out.println("crossover exception:"+e.getMessage());
	    // crossover(next);
	    return;
	}
    }
    /** Mutates a percentage of the population.
     */
    public void mutatePercentageOfPop(double percent) {
	SimilarityStatement best = findBest();
	
	int howmany = (int)(percent*population.size());
	for (int i=0; i<howmany; i++) {
	    
	    int chosen = rand.nextInt(population.size());
	    SimilarityStatement ss = (SimilarityStatement) population.get(chosen);
	    while (ss==best) {
		/*		System.out.println();
		System.out.println(" D E B U G D E B U G ss==best in mut%");
		System.out.println();*/
		chosen = rand.nextInt(population.size());
		ss = (SimilarityStatement) population.get(chosen);
	    }
	    SimilarityStatement copy=(SimilarityStatement)ss.clone();
	    /*	    if (seqPop()) {
		copy = (SeqSimStmt) ss.clone();
	    } else {
		copy = (NonSeqSimStmt) ss.clone();
		}*/
	    
	    try {
		copy.mutate(mutationrate); 
		population.set(chosen,copy);
	    }
	    catch (Exception e) {
		// mutation failed, forget about it
	    }
	}
    }
    /** Picks a random individual in the population. Mutates it and
	adds it to the population for the next generation.

    */
    public void mutation(Vector next) {
	int inx = rand.nextInt(next.size());
	SimilarityStatement ss = (SimilarityStatement) next.get(inx);
	SimilarityStatement ssmut = (SimilarityStatement) ss.clone();
	try {
	    ssmut.mutate(mutationrate);
	    if (!ssmut.oksize()) {
		System.out.println("mutation error:"+ssmut.toString());
		System.exit(0);
	    }
	    next.add(ssmut);
	}
	catch(Exception e) {
	    // something went awry during mutation so ssmut is possibly unstable,
	    // abandon it and try again
	    // System.out.println("mutation exception:"+e.getMessage());
	    // mutation(next);
	    return;
	}
    }

    /** Returns a clone of the most fit of two rendomly chosen {@link SimilarityStatement}.
     */
    public SimilarityStatement tournament() {
	// pick two contestants
	SimilarityStatement con1= (SimilarityStatement) population.get(rand.nextInt(population.size()));
	SimilarityStatement con2= (SimilarityStatement) population.get(rand.nextInt(population.size()));
	while (con2==con1) {
	    con2= (SimilarityStatement) population.get(rand.nextInt(population.size()));
	}

	double fitness1 = con1.value();
	double fitness2 = con2.value();

	SimilarityStatement winner=con2;
	if (fitness1<fitness2) winner=con1;
	// if equal fitness, pick the smallest one
	if (fitness1==fitness2) {
	    if (con1.size() <con2.size()) winner=con1;
	}
	/*
	// check for identical subgraphs, they should be weeded out
	if (con1.identicalSubgraphs()) winner = con2;
	if (con2.identicalSubgraphs()) winner = con1;
	*/
	return (SimilarityStatement) winner.clone();
    }

    int findIndexOfLowest(double[] d) {
	double low= Double.MAX_VALUE;
	int lowindex = -1;
	for (int i=0; i<d.length; i++) {
	    if (d[i]<low) {
		low=d[i];
		lowindex=i;
	    }
	}
	return lowindex;
    }

    /** Computes how many vertices in the mother piece that are included 
     * in a SimStmt in the population */ 
    public int vertexCoverage() {
	Vector containedVertices = new Vector();
	for (int i=0; i<population.size(); i++) {
	    SimilarityStatement ss = (SimilarityStatement) population.get(i);
	    for (int j=0; j<ss.subgraph1.vertices.size(); j++) {
		MusicVertex mv = (MusicVertex) ss.subgraph1.vertices.get(j);
		if (!containedVertices.contains(mv)) containedVertices.add(mv);
	    }
	    for (int j=0; j<ss.subgraph2.vertices.size(); j++) {
		MusicVertex mv = (MusicVertex) ss.subgraph2.vertices.get(j);
		if (!containedVertices.contains(mv)) containedVertices.add(mv);
	    }
	}
	return containedVertices.size();
    }

    /** Computes how many edges in the mother piece that are included 
     * in a SimStmt in the population */ 
    public int edgeCoverage() {
	Vector containedEdges = new Vector();
	for (int i=0; i<population.size(); i++) {
	    SimilarityStatement ss = (SimilarityStatement) population.get(i);
	    for (int j=0; j<ss.subgraph1.edges.size(); j++) {
		MusicEdge mv = (MusicEdge) ss.subgraph1.edges.get(j);
		if (!containedEdges.contains(mv)) containedEdges.add(mv);
	    }
	    for (int j=0; j<ss.subgraph2.edges.size(); j++) {
		MusicEdge mv = (MusicEdge) ss.subgraph2.edges.get(j);
		if (!containedEdges.contains(mv)) containedEdges.add(mv);
	    }
	}
	return containedEdges.size();
    }

    /** Sets the useCounter in each vertice. The useCounter tells how
	many Subgraphs of the SimilarityStatements in the population
	the vertice can be found in.
     */
    public void countUsageInfo() {
	piece.resetUsageInfo();
	for (int i=0; i<population.size(); i++) {
	    SimilarityStatement ss = (SimilarityStatement) population.get(i);

	    /* DEBUG
	      if(occurrence==null && occurrenceSearch()){
		System.out.println("occurrence is null!");
		System.exit(0);
	    }
	    if(ss==null){
		System.out.println("countUsageInfo: ss==null");
	    }else if(ss.subgraph1==null){
		System.out.println("ss.subgraph1==null");
	    }else if(ss.subgraph2==null){
		System.out.println("pop(i), i="+i);
		System.out.println("ss.subgraph2==null");
	    }
	    */
	    for (int j=0; j<ss.subgraph1.vertices.size(); j++) {
		MusicVertex mv = (MusicVertex) ss.subgraph1.vertices.get(j);
		mv.useCounter++;
	    }
	    for (int j=0; j<ss.subgraph2.vertices.size(); j++) {
		MusicVertex mv = (MusicVertex) ss.subgraph2.vertices.get(j);
		mv.useCounter++;
	    }
	}
    }

    public void appendVertexUseStatsToFile(String filename) {
	// count vertex use
	countUsageInfo();

	// append use info to file
	try {
	    FileWriter outfile=new FileWriter(filename,true);
	    outfile.write("# SimFinder vertex usage stats appended at "+new Date().toString()+"\n");
	    Vector v = (Vector) piece.vertices.clone();
	    sortByUsageCount(v);

	    for (int i=0; i<v.size(); i++) {
		MusicVertex mv= (MusicVertex) v.get(i);
		outfile.write(mv.useCounter+"\n");
	    }
	    outfile.write("\n");
	    outfile.flush();
	    outfile.close();
	} catch(Exception e) {
	    System.out.println("Exception caught in SimFinder.writeVertexUseStatsToFile(): "+e.getMessage());
	}
    }

    /* Takes a Vector of MusicVertex */
    private void sortByUsageCount(Vector v) {
	for (int i=0; i<v.size(); i++) {
	    for (int j=0; j<v.size()-1;j++) {
		MusicVertex mv1 = (MusicVertex) v.get(j);
		MusicVertex mv2 = (MusicVertex) v.get(j+1);
		if (mv1.useCounter<mv2.useCounter) {
		    //swap
		    v.set(j,mv2);
		    v.set(j+1,mv1);
		}
	    }
	}
    }

    /** for debugging */
    public void testPopulationIntegrity() {
	int result=0;
	for (int i=0; i<population.size(); i++) {
	    SimilarityStatement ss = (SimilarityStatement) population.get(i);
	    result+=ss.subgraph1.testIntegrity();
	    result+=ss.subgraph2.testIntegrity();
	}
	System.out.println("testPopulationIntegrity: "+result+" errors found.");
    }

    /** Lets the thread sleep for <tt>millisecs</tt> milliseconds. */
    public void sleep(long millisecs){
	try{Thread.sleep(millisecs);}catch(Exception e){ 
	    System.out.println(e.getMessage());
	}
    }
    /** Prints the String.*/
    void log(String s){System.out.println(s);}
    /** Prints the integer.*/
    void log(int s){System.out.println(s);}
    /** Prints the double. */
    void log(double s){System.out.println(s);}
    /** Prints the boolean. */
    void log(boolean s){System.out.println(s);}
    
    /** For debugging. */
    public void checkPopulation(String label){
	
	// System.out.println(label);
	// System.out.println("CheckPop: "+label);
	/*	
	if(occurrenceSearch() && (occurrence==null)){
	    System.out.println("occurrenceSearch() && occurrence==null "+label);
	    System.exit(0);
	}
	for(int i=0;i<population.size();i++){
	    SimilarityStatement s=(SimilarityStatement)population.get(i);
	    Subgraph sub1=s.subgraph1;
	    Subgraph sub2=s.subgraph2;
	    if((sub1==null) || (sub2==null)){
		System.out.println("sub1==null: "+(sub1==null)+" error "+label);
		System.out.println("sub2==null: "+(sub2==null)+" error "+label);
		new Exception("tt").printStackTrace();
		System.exit(0);
		}
		}
	*/
	// System.out.flush();
    }
    
}
