/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.*;
import java.util.Vector;

/** SimLib is a class not to be instantiated. Contains static
   instances of anonymous class extensions of SimilarityMeasure and
   ViewComparator. */
public class SimLib {

    /* A general method to compare compound vertices: given a
       similarity measure sm, if the vertices are of the same
       nonterminal, do their respective similarity measures subsume
       sm?*/
    static double compareCompoundsUnderMeasure(CompoundVertex cv1,
						      CompoundVertex cv2,
						      SimilarityMeasure sm) {
	if (cv1.productionRule.nt!=cv2.productionRule.nt) return 1.0;
	else {
	    if (simMeasurePitchSubsumption(sm,cv1.productionRule.measure) &&
		simMeasurePitchSubsumption(sm,cv2.productionRule.measure))
		return 0;
	}
	return 1.0;
    }
    static double compareCompoundsUnderViewComparator(CompoundVertex cv1,
							     CompoundVertex cv2,
							     ViewComparator vc) {
	if (cv1.productionRule.nt!=cv2.productionRule.nt) return 1.0;
	else {
	    ViewComparator vcrestrictive1 = cv1.productionRule.measure.getMostRestrictivePitchVC();
	    ViewComparator vcrestrictive2 = cv2.productionRule.measure.getMostRestrictivePitchVC();
	    if (viewComparatorPitchSubsumption(vc,vcrestrictive1) &&
		viewComparatorPitchSubsumption(vc,vcrestrictive2))
		return 0;
	}
	return 1.0;
    }

    /* Tells whether or not 'subsumed' is subsumed by 'subsumer' in
       the hierarchy of pitch-related similarity measures. The
       subsumption relation means that the subsumed simMeasure is more
       specific, and that the subsuming simMeasure is more general,
       thus having a larger set of perfect matches which encompasses,
       or subsumes, the set of perfect matches for the subsumed
       simMeasure. */
    static boolean simMeasurePitchSubsumption(SimilarityMeasure subsumer,
						     SimilarityMeasure subsumed) {
	ViewComparator vcsubsumer = subsumer.getMostRestrictivePitchVC();
	ViewComparator vcsubsumed = subsumed.getMostRestrictivePitchVC();
	
	return viewComparatorPitchSubsumption(vcsubsumer,vcsubsumed);
    }

    /* Tells whether or not 'subsumed' is subsumed by 'subsumer' in
       the hierarchy of pitch-related view comparators. The
       subsumption relation: see the note above for
       simMeasurePitchSubsumption. What we return here is actually
       'subsumes-or-is-equal-to'.*/
    static boolean viewComparatorPitchSubsumption(ViewComparator subsumer,
							 ViewComparator subsumed) {
	// the checks are hard coded this is very bad coding
	// style. improve as soon at possible by modeling the
	// subsumption hierarchy of pitch-related view comaparators
	if (subsumer==null && subsumed==null) return false;
	if (subsumer==subsumed) return true;
	if (subsumer==null) {
	    // anything is subsumed by null (is more specific than null)
	    if (subsumed==SimLibMeanDiff.absolutePitchMeanDiff) return true;
	    if (subsumed==SimLibMeanDiff.pitchIntervalMeanDiff) return true;
	    if (subsumed==SimLibMeanDiff.pitchClassMeanDiff) return true;
	    if (subsumed==SimLibMeanDiff.pitchAbsIntervalMeanDiff) return true;
	    if (subsumed==SimLibMeanDiff.pitchContourMeanDiff) return true;

	    if (subsumed==SimLibDiffCount.absPitchDiffCount) return true;
	    if (subsumed==SimLibDiffCount.pitchIntervalDiffCount) return true;
	    if (subsumed==SimLibDiffCount.diatonicIntervalDiffCount) return true;
	    if (subsumed==SimLibDiffCount.diatonicForwardIntervalDiffCount) return true;

	    if (subsumed==SimLibVertex.nsvcVertexPitch) return true;
	    if (subsumed==SimLibVertex.nsvcVertexPitchClass) return true;

	    if (subsumed==SimLibEdge.nsvcEdgePitchInterval) return true;
	    if (subsumed==SimLibEdge.nsvcEdgePitchContour) return true;
	}

	// mean difference view comparators
	if (subsumer==SimLibMeanDiff.pitchClassMeanDiff &&
	    subsumed==SimLibMeanDiff.absolutePitchMeanDiff) return true;
	if (subsumer==SimLibMeanDiff.pitchIntervalMeanDiff &&
	    subsumed==SimLibMeanDiff.absolutePitchMeanDiff) return true;
	if (subsumer==SimLibMeanDiff.pitchAbsIntervalMeanDiff &&
	    subsumed==SimLibMeanDiff.absolutePitchMeanDiff) return true;
	if (subsumer==SimLibMeanDiff.pitchContourMeanDiff &&
	    subsumed==SimLibMeanDiff.absolutePitchMeanDiff) return true;

	if (subsumer==SimLibMeanDiff.pitchAbsIntervalMeanDiff &&
	    subsumed==SimLibMeanDiff.pitchIntervalMeanDiff) return true;
	if (subsumer==SimLibMeanDiff.pitchContourMeanDiff &&
	    subsumed==SimLibMeanDiff.pitchIntervalMeanDiff) return true;

	// diff count view comparators
	if (subsumer==SimLibDiffCount.pitchIntervalDiffCount &&
	    subsumed==SimLibDiffCount.absPitchDiffCount) return true;
	if (subsumer==SimLibDiffCount.diatonicIntervalDiffCount &&
	    subsumed==SimLibDiffCount.absPitchDiffCount) return true;
	if (subsumer==SimLibDiffCount.diatonicForwardIntervalDiffCount &&
	    subsumed==SimLibDiffCount.absPitchDiffCount) return true;

	if (subsumer==SimLibDiffCount.diatonicIntervalDiffCount &&
	    subsumed==SimLibDiffCount.pitchIntervalDiffCount) return true;
	if (subsumer==SimLibDiffCount.diatonicForwardIntervalDiffCount &&
	    subsumed==SimLibDiffCount.pitchIntervalDiffCount) return true;

	if (subsumer==SimLibDiffCount.diatonicForwardIntervalDiffCount &&
	    subsumed==SimLibDiffCount.diatonicIntervalDiffCount) return true;

	// nsvc Vertex
	if (subsumer==SimLibVertex.nsvcVertexPitchClass &&
	    subsumed==SimLibVertex.nsvcVertexPitch) return true;

	// nsvc Edge
	if (subsumer==SimLibEdge.nsvcEdgePitchContour &&
	    subsumed==SimLibEdge.nsvcEdgePitchInterval) return true;

	return false;
    }

    static SeqSimStmt constructSeqSimStmt(MusicEdge me1, MusicEdge me2,
						 MotherGraph mother, 
						 SeqSimMeasure ssm) {
	SeqSubgraph ss1 = new SeqSubgraph("Temporary Evaluation SeqSubgraph 1",mother);
	SeqSubgraph ss2 = new SeqSubgraph("Temporary Evaluation SeqSubgraph 2",mother);
 	ss1.vertices.add(me1.from);
	ss1.vertices.add(me1.to);
	ss1.edges.add(me1);

 	ss2.vertices.add(me2.from);
	ss2.vertices.add(me2.to);
	ss2.edges.add(me2);

	SeqSimStmt result = new SeqSimStmt(ss1,ss2,ssm,null/*the SimFinder*/);
	return result;
    }

    /* Compound enabled */ 
    public static final SeqViewComparator seqGroupingStructure=
	new SeqViewComparator("seqGroupingStructure"){
		
		public ViewDifference seqCompare(SeqSimStmt sss){
		    GroupingStructure g=new GroupingStructure();
		    double g1=g.evaluateGroupings(sss.subgraph1, 
						  sss.myOwner.groupBaseGamma/*i SimFinderen*/);
		    double g2=g.evaluateGroupings(sss.subgraph2, 
						  sss.myOwner.groupBaseGamma);
		    double val=(g1+g2)/2.0;
		    ViewDifference result = new ViewDifference(val,this);
		    return result;
		}
	    };

    /* Compound enabled */ 
    public static final NonSeqViewComparator nonSeqGroupingStructure=
	new NonSeqViewComparator("nonSeqGroupingStructure"){
		
		public ViewDifference nonSeqCompare(NonSeqSimStmt nsss){
		    GroupingStructure g=new GroupingStructure();
		    double g1=g.evaluateGroupings(nsss.subgraph1,nsss.myOwner.groupBaseGamma);
		    double g2=g.evaluateGroupings(nsss.subgraph2,nsss.myOwner.groupBaseGamma);
		    double val=(g1+g2)/2.0;
		    ViewDifference result = new ViewDifference(val,this);
		    return result;
		}
	    };

    public static final NonSeqViewComparator nonSeqGroupingStructure2=
	new NonSeqViewComparator("nonSeqGroupingStructure"){
		
		public ViewDifference nonSeqCompare(NonSeqSimStmt nsss){
		    GroupingStructure g=new GroupingStructure();
		    double g1=g.evaluateGroupings(nsss.subgraph1, 0.2);
		    double g2=g.evaluateGroupings(nsss.subgraph2, 0.2);
		    double val=(g1+g2)/2.0;
		    ViewDifference result = new ViewDifference(val,this);
		    return result;
		}
	    };



    /* ---------- VIEWCOMPARATORS ---------- */
    public static final NSVCBagIR bagIRTrieDAP = new NSVCBagIR("bagIRTrieDAP","dap") {
	    
	    public ViewDifference nonSeqCompare(NonSeqSimStmt nsss){
		double res=0.0;
		int shortest=getSizeOfShortestSeqSubgraph(nsss);
		int bagLength=decideSize(shortest);
		double rating=bagIntersectionRatio(bagLength,nsss);
		int ratingBag1size=bag1size;
		int ratingBag2size=bag2size;
		int ratingMatchSize=matchSize;
		double ratingWhenLengthIs2=0.0;//bagIntersectionRating for size 2
		boolean rat2calculated=false;
		if(bagLength>2){
		    ratingWhenLengthIs2=bagIntersectionRatio(2,nsss);
		    rat2calculated=true;
		}
		double ratingVertices=0.0;
		ratingVertices=bagIntersectionRatioForVertices(nsss);
		double bagIRforSims=bagIntersectionRatioForSimEdges(nsss,"dap");
		//double sizePenalty=sizeValue(nsss.meanNestedSize());
		double sizePenalty=sizeValue(nsss.size());

		res=ratingVertices+ratingWhenLengthIs2;
		if(rating!=-1.0) {//not able to calculate (no follow edges in subgraph)
		    res+=rating;
		}
		if(bagIRforSims!=-1.0){
		    res+=bagIRforSims;
		}
		res+=sizePenalty;
		if(!rat2calculated)ratingWhenLengthIs2=-1.0;
		
		if(rating==-1.0 && bagIRforSims==-1.0) res+=100.0;

		
		return new NSVCBagIRDifference(res,this,
					       ratingMatchSize,ratingBag1size,ratingBag2size,
					       bagLength, sizePenalty,
					       rating, ratingWhenLengthIs2, 
					       ratingVertices, bagIRforSims
					       );
	    }
	};
    public static final NSVCBagIR bagIRTriePI = new NSVCBagIR("bagIRTriePI","pi") {
	    
	    public ViewDifference nonSeqCompare(NonSeqSimStmt nsss){
		double res=0.0;
		int shortest=getSizeOfShortestSeqSubgraph(nsss);
		int bagLength=decideSize(shortest);
		double rating=bagIntersectionRatio(bagLength,nsss);
		int ratingBag1size=bag1size;
		int ratingBag2size=bag2size;
		int ratingMatchSize=matchSize;
		double ratingWhenLengthIs2=0.0;//bagIntersectionRating for size 2
		boolean rat2calculated=false;
		if(bagLength>2){//
		    ratingWhenLengthIs2=bagIntersectionRatio(2,nsss);
		    rat2calculated=true;
		}

		double bagIRforSims=bagIntersectionRatioForSimEdges(nsss,"pi");

		//double sizePenalty=sizeValue(nsss.meanNestedSize());
		double sizePenalty=sizeValue(nsss.size());
		res=ratingWhenLengthIs2;
		if(rating!=-1.0) {//not able to calculate (no follow edges in subgraph)
		    res+=rating;
		}
		if(bagIRforSims!=-1.0){//not able to calculate
		    res+=bagIRforSims;
		}
		res+=sizePenalty;
		if(!rat2calculated)ratingWhenLengthIs2=-1.0;
		if(rating==-1.0 && bagIRforSims==-1.0) res+=100.0;
		double ratingVertices=-1.0;
		return new NSVCBagIRDifference(res,this,
					       ratingMatchSize,ratingBag1size,ratingBag2size,
					       bagLength, sizePenalty,
					       rating, ratingWhenLengthIs2, 
					       ratingVertices, bagIRforSims
					       );
	    }
	};
    public static final NSVCBagIR bagIRTrieDI = new NSVCBagIR("bagIRTrieDI","di") {
	    
	    public ViewDifference nonSeqCompare(NonSeqSimStmt nsss){
		double res=0.0;
		int shortest=getSizeOfShortestSeqSubgraph(nsss);
		int bagLength=decideSize(shortest);
		double rating=bagIntersectionRatio(bagLength,nsss);
		int ratingBag1size=bag1size;
		int ratingBag2size=bag2size;
		int ratingMatchSize=matchSize;
		double ratingWhenLengthIs2=0.0;//bagIntersectionRating for size 2
		boolean rat2calculated=false;
		if(bagLength>2){//
		    ratingWhenLengthIs2=bagIntersectionRatio(2,nsss);
		    rat2calculated=true;
		}

		double bagIRforSims=bagIntersectionRatioForSimEdges(nsss,"di");

		//double sizePenalty=sizeValue(nsss.meanNestedSize());
		double sizePenalty=sizeValue(nsss.size());

		res=ratingWhenLengthIs2;
		if(rating!=-1.0) {//not able to calculate (no follow edges in subgraph)
		    res+=rating;
		}
		if(bagIRforSims!=-1.0){//not able to calculate
		    res+=bagIRforSims;
		}
		res+=sizePenalty;
		if(!rat2calculated)ratingWhenLengthIs2=-1.0;
		if(rating==-1.0 && bagIRforSims==-1.0) res+=100.0;
		double ratingVertices=-1.0;
		return new NSVCBagIRDifference(res,this,
					       ratingMatchSize,ratingBag1size,ratingBag2size,
					       bagLength, sizePenalty,
					       rating, ratingWhenLengthIs2, 
					       ratingVertices, bagIRforSims
					       );
	    }
	};

    public static final NSVCBagIR bagIRTrieDFI = new NSVCBagIR("bagIRTrieDFI","dfi") {
	    
	    public ViewDifference nonSeqCompare(NonSeqSimStmt nsss){
		double res=0.0;
		int shortest=getSizeOfShortestSeqSubgraph(nsss);
		int bagLength=decideSize(shortest);
		double rating=bagIntersectionRatio(bagLength,nsss);
		int ratingBag1size=bag1size;
		int ratingBag2size=bag2size;
		int ratingMatchSize=matchSize;
		double ratingWhenLengthIs2=0.0;//bagIntersectionRating for size 2
		boolean rat2calculated=false;
		if(bagLength>2){//
		    ratingWhenLengthIs2=bagIntersectionRatio(2,nsss);
		    rat2calculated=true;
		}

		double bagIRforSims=bagIntersectionRatioForSimEdges(nsss,"dfi");

		//double sizePenalty=sizeValue(nsss.meanNestedSize());
		double sizePenalty=sizeValue(nsss.size());
		res=ratingWhenLengthIs2;
		if(rating!=-1.0) {//not able to calculate (no follow edges in subgraph)
		    res+=rating;
		}
		if(bagIRforSims!=-1.0){//not able to calculate
		    res+=bagIRforSims;
		}
		res+=sizePenalty;
		if(!rat2calculated)ratingWhenLengthIs2=-1.0;
		if(rating==-1.0 && bagIRforSims==-1.0) res+=100.0;
		double ratingVertices=-1.0;
		return new NSVCBagIRDifference(res,this,
					       ratingMatchSize,ratingBag1size,ratingBag2size,
					       bagLength, sizePenalty,
					       rating, ratingWhenLengthIs2, 
					       ratingVertices, bagIRforSims
					       );
	    }
	};
    
    
    /* IMPORTANT!! When placed in this same file, the similarity
       measures should be placed AFTER the ViewComparator definitions,
       otherwise the latter will not have been initialised at the time
       when the SimilarityMeasures are initialised, resulting in null
       pointer exceptions. */
		
    /* ---------- SIMILARITY MEASURES ---------- */

    //kaldes fra StartMusedataSimFinder
    public static final NonSeqSimMeasure nonSeqBagIRTrie_DiatonicAbsPitch_Grp = new NonSeqSimMeasure("nonSeqBagIRTrie_DiatonicAbsPitch_Grp") {
	    
	    public void initialize() {
		viewComparators.clear();
		viewComparators.add(bagIRTrieDAP);//defineret ovenfor
		viewComparators.add(nonSeqGroupingStructure);
	    }
	
	    public void evaluateNonSeq(NonSeqSimStmt nsss) {
		double absPitch=Double.MAX_VALUE;
		double group=Double.MAX_VALUE;
		Vector viewdiffs=evaluateAllViewComparators(nsss);
		for(int i=0;i<viewdiffs.size();i++){
		    ViewDifference vpr = (ViewDifference) viewdiffs.get(i);
		    if((vpr.origin.name).equals("bagIRTrieDAP")){
			absPitch=vpr.view;
		    }else if((vpr.origin.name).equals("nonSeqGroupingStructure")){
			group=vpr.view;
		    }else{
			System.out.println("SimLib: Unknown ViewComparator error:");
			System.out.println(vpr.origin.name);
			System.exit(0);
		    }
		}
		double result=absPitch;
		result+=group;
		SimMeasureResult smr = new SimMeasureResult(result,viewdiffs,this);
		nsss.smResult = smr;
	    }
	};

    public static final NonSeqSimMeasure nonSeqBagIRTrie_PitchInt_Grp = new NonSeqSimMeasure("nonSeqBagIRTrie_PitchInt_Grp") {
	    public void initialize() {
		viewComparators.clear();
		viewComparators.add(SimLib.bagIRTriePI);//defineret ovenfor
		viewComparators.add(nonSeqGroupingStructure);
	    }
	
	    public void evaluateNonSeq(NonSeqSimStmt nsss) {
		double pitchInt=Double.MAX_VALUE;
		double group=Double.MAX_VALUE;
		Vector viewdiffs=evaluateAllViewComparators(nsss);
		for(int i=0;i<viewdiffs.size();i++){
		    ViewDifference vpr = (ViewDifference) viewdiffs.get(i);
		    if((vpr.origin.name).equals("bagIRTriePI")){
			pitchInt=vpr.view;
		    }else if((vpr.origin.name).equals("nonSeqGroupingStructure")){
			group=vpr.view;
		    }else{
			System.out.println("SimLib: Unknown ViewComparator error:");
			System.out.println(vpr.origin.name);
			System.exit(0);
		    }
		}
		double result=pitchInt;
		result+=group;
		SimMeasureResult smr = new SimMeasureResult(result,viewdiffs,this);
		nsss.smResult = smr;
	    }
	};

    public static final NonSeqSimMeasure nonSeqBagIRTrie_DiatonicInt_Grp = new NonSeqSimMeasure("nonSeqBagIRTrie_DiatonicInt_Grp") {
	    
	    public void initialize() {
		viewComparators.clear();
		viewComparators.add(SimLib.bagIRTrieDI);//defineret ovenfor
		viewComparators.add(SimLib.nonSeqGroupingStructure);
	    }
	
	    public void evaluateNonSeq(NonSeqSimStmt nsss) {
		double di=Double.MAX_VALUE;
		double group=Double.MAX_VALUE;
		Vector viewdiffs=evaluateAllViewComparators(nsss);
		for(int i=0;i<viewdiffs.size();i++){
		    ViewDifference vpr = (ViewDifference) viewdiffs.get(i);
		    if((vpr.origin.name).equals("bagIRTrieDI")){
			di=vpr.view;
		    }else if((vpr.origin.name).equals("nonSeqGroupingStructure")){
			group=vpr.view;
		    }else{
			System.out.println("SimLib: Unknown ViewComparator error:");
			System.out.println(vpr.origin.name);
			System.exit(0);
		    }
		}
		double result=di;
		result+=group;
		SimMeasureResult smr = new SimMeasureResult(result,viewdiffs,this);
		nsss.smResult = smr;
	    }
	};

    public static final NonSeqSimMeasure nonSeqBagIRTrie_DFI_Grp = new NonSeqSimMeasure("nonSeqBagIRTrie_DFI_Grp") {
	    		
	    public void initialize() {
		viewComparators.clear();
		viewComparators.add(SimLib.bagIRTrieDFI);//defineret ovenfor
		viewComparators.add(SimLib.nonSeqGroupingStructure);
	    }
	
	    public void evaluateNonSeq(NonSeqSimStmt nsss) {
		double dfi=Double.MAX_VALUE;
		double group=Double.MAX_VALUE;
		Vector viewdiffs=evaluateAllViewComparators(nsss);
		for(int i=0;i<viewdiffs.size();i++){
		    ViewDifference vpr = (ViewDifference) viewdiffs.get(i);
		    if((vpr.origin.name).equals("bagIRTrieDFI")){
			dfi=vpr.view;
		    }else if((vpr.origin.name).equals("nonSeqGroupingStructure")){
			group=vpr.view;
		    }else{
			System.out.println("SimLib: Unknown ViewComparator error:");
			System.out.println(vpr.origin.name);
			System.exit(0);
		    }
		}
		double result=dfi;
		result+=group;
		SimMeasureResult smr = new SimMeasureResult(result,viewdiffs,this);
		nsss.smResult = smr;
	    }
	};

    public static double sizeValue(double d){return (1.0-(d*(0.01)));}


}
