/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Vector;
import java.io.*;

/** Class SimSegmenter embodies the segmentation algorithm. Its
    constructor accepts a host of parameters that control both how the
    SimSegmenter itself runs but also how it should use the
    SimFinder.*/
public class SimSegmenter {

    
    /** The current SimFinder used by this SimSegmenter. */
    public SimFinder sf;
    /** The MotherGraph being segmented by this SimSegmenter. */
    public MotherGraph mother;
    /** Parameter for the SimFinder. */
    public double crossover;
    /** Parameter for the SimFinder. */
    public double mutation;
    /** Parameter for the SimFinder (called initSubSize in the SimFinder). */
    public int initSize;
    /** Parameter for the SimFinder (populationSize) */
    public int popSize;
    /** Parameter for the SimFinder (reportFrequency)*/
    public int repFreq;
    /** Parameter for the SimFinder (maxSubsize). */
    public int maxSize=0;
    /** Parameter for the SimFinder. */
    public double freshBloodChance=0.05;
    /** Parameter for the SimFinder (groupBaseGamma). */
    public double groupBase;

    /** Name of the midi or MuseData file that 'mother' is constructed
        from. */
    public String file;
    /** How many generations a SimSegmenter should search for
        similarities using a given SimilarityMeasure before giving
        up. */
    public int simGens;
    /** How many generations a SimSegmenter should search for
        occurrences using a given SimilarityMeasure before giving
        up. */
    public int occGens;
    /** How many generations a SimSegmenter should keep searching for
        similarities using a given SimilarityMeasure even though it
        finds an acceptable similarity. If -1, runs through all occGens. */
    public int generationsAfterLimitReached; // if -1, runs through all occGens 
    /** How many generations the SimSegmenter should run the SimFinder
        between checks for SimilarityStatements that have come under
        the limit, i.e. for acceptable matches. */
    public int gensPerLimitCheck;
    
    /** The limit value that SimilarityStatements must come below to
        be acceptable. This varies between similarity measures. */
    public double limit=1.0;
    /** The current similarity measure used for searching. */
    public SimilarityMeasure theMeasure;
    /** The set of similarity measures ranked from specific to general. */
    public SimilarityMeasure[] measures;
    /** The current nonterminal number. */
    public int nonTerminalNumber=-1;
    
    /** A Vector of {@link Nonterminal}s already created. */
    public Vector nonterm=new Vector();
    /** A Vector of all {@link GraphProductionRule}s already
        created. */
    public Vector prodRules = new Vector();
    double meanNestedSizeLimit=3.0;
	
    /** Starts the SimSegmenter: Usage: java SimSegmenter <filename> */
    public static void main(String[] args) throws Exception{
	if(args.length!=1){
	    System.out.println("Usage: java SimSegmenter <filename>");
	    System.exit(0);
	}
	
	SimilarityMeasure[] measNoSeq={
	    SimLib.nonSeqBagIRTrie_DiatonicAbsPitch_Grp,
	    //SimLib.nonSeqBagIRTrie_PitchInt_Grp,
	    SimLib.nonSeqBagIRTrie_DiatonicInt_Grp,
	    SimLib.nonSeqBagIRTrie_DFI_Grp
	};
	
	SimilarityMeasure[] inversion = {SimLibDiffCount.seqDiatonicInversionAndLength_Grp};
	
	SimilarityMeasure[] measSeq = {
	    SimLibDiffCount.seqDiatonicAbsPitchAndLength_Grp,
	    //SimLibDiffCount.seqPitchIntervalAndLength_Grp,
	    SimLibDiffCount.seqDiatonicIntervalAndLength_Grp,
	    SimLibDiffCount.seqDFIAndLength_Grp
	};
	
	//klar til at kre koralen partwise
	SimSegmenter segm = new SimSegmenter(args[0],
					     false, // interpartEdges
					     true, // combineTies
					     false, // combineRests
					     measSeq,
					     0.01, // crossover
					     0.5, // mutation
					     5, // initsize
					     80, // popsize
					     10, // repFreq
					     "sf.out", // outfilename
					     500, // simGens
					     400, // occGens
					     2.0, // limit
					     0, // maxsize
					     0.1, //fresh blood chance - in mutation
					     -1, // gensAfterLimitReached
					     10, // gensPerLimitCheck
					     0.7 //groupBase-gamma
					     );
	segm.run();
	/* 
	segm.groupBase=0.6;
	segm.measures=measNoSeq;
	segm.limit=2.0;
	segm.simGens=150;
	segm.occGens=150;
	segm.initSize=4;
	segm.popSize=80;
	segm.run();
	*/
	segm.showAllProductionRules(segm.prodRules);
	segm.show();
    }


	/** 
	   @param filename The input file.
	   @param interpartEdges Whether the graph should be partwise
	   (interpartEdges = false) or not (interpartEdges = true).
	   @param combineTies Make only one note (vertice) in stead of
	   two tied.
	   @param combineRests Make one (longer) rest (vertice) in
	   stead of two (shorter).

	   @param meas the array of {@link SimilarityMeasure}s. The
	   order of the measures determines the order in which the
	   search takes place.
	   @param cross the crossover parameter used in the GA.
	   @param mutthe mutation parameter used in the GA.
	   @param initsz the initial size for the subgraphs used in the GA.
	   @param popsz the population size used in the GA.
	   @param repfq report frequency used in the GA.
	   @param outfile the output file used in the GA.
	   @param simgen the number of generations used in the GA to
	   find best matching graphs.
	   @param occgen the number of generations used to find an
	   occurrence of an already found {@link Subgraph}.
	   @param lim the threshold. {@link SimilarityStatement}s with
	   a fitness below this limit are accepted (ant their
	   subgraphs substituted with a compound).
	   @param maxsiz it is possoble to specify a maximal size for
	   the subgraphs used in the GA.
	   @param freshblood used in the GA.
	   @param gensAfterLimitReached runs this number of
	   generations even though the mest found aslready is below
	   the limit (threshold lim) specified.
	   @param gensLimitCheck run this number of generations before
	   checking if the limit is reached.
	   @param gamma the grouping structire value
	   @throws Exception if the SimFinder is not able to be constructed. */
    public SimSegmenter(String filename,
			boolean interpartEdges,
			boolean combineTies,
			boolean combineRests,
			SimilarityMeasure[] meas,
			double cross,
			double mut,
			int initsz,
			int popsz,
			int repfq,
			String outfile,
			int simgen,
			int occgen,
			double lim,
			int maxsiz,
			double freshblood,
			int gensAfterLimitReached,
			int gensLimitCheck,
			double gamma
			) throws Exception{
	Vector prodRules = new Vector(); // of GraphProductionRule
	Vector nonterm = new Vector(); // of Nonterminal

	if (filename.endsWith("md") ||
	    filename.endsWith("md2") ||
	    filename.endsWith("MD") ||
	    filename.endsWith("MD2")) {
	    // call musedata constructor
	    mother = new MotherGraph(filename,interpartEdges,combineTies,combineRests);
	} else if (filename.endsWith("mid") ||
		   filename.endsWith("MID")) {
	    // call midi constructor
	    mother = new MotherGraph(filename,interpartEdges);
	} else {
	    throw new Exception("Unrecognised file: "+filename+" -- filename must end with 'md', 'md2' or 'mid'");
	}
	
	measures=meas;
	crossover=cross;
	mutation=mut;
	initSize=initsz;
	popSize=popsz;
	repFreq=repfq;
	file=outfile;
	simGens=simgen;
	occGens=occgen;
	limit=lim;
	maxSize=maxsiz;
	freshBloodChance=freshblood;
	generationsAfterLimitReached=gensAfterLimitReached;
	gensPerLimitCheck=gensLimitCheck;
	groupBase=gamma;
    }

    /** Runs the Simsegmenting algorithm. */
    public void run() {
	int maxThemes=20;
	
        boolean newTheme=true;
	for(int i=0;i<measures.length;i++){
	    while(newTheme && nonTerminalNumber<maxThemes){
		theMeasure=measures[i];
		nonTerminalNumber++;
 		printHeader("NEW SEARCH "+nonTerminalNumber
			    +" with measure: "+ theMeasure.name);
		sleep();
		newTheme=false;
		sf=getSF();
		runSF();

		SimilarityStatement result=sf.findBest();
		// is the best found result good enough?
		if(result.subgraph1.overlap(result.subgraph2)==0 &&
		   result.value()<limit &&
		   result.meanNestedSize()>3.0
		   ) {
		    /* Found a good SimilarityStatement ( a motif), so
                       there might be others (derivations). Search for
                       more themes when restarting the while-loop*/
		    newTheme=true;
		    Vector forbidden=new Vector();

		    // clone the subgraphs that were found to be similar
		    Subgraph occGraph1=prepareSubgraphForOccSearch(result.subgraph1);
		    Subgraph occGraph2=prepareSubgraphForOccSearch(result.subgraph2);

		    /* ------- COMPOUND SUBSTITUTION BEGIN -------*/
		    /* Replace sub1 and sub2 with compounds, and add a
                       subgraph containing each of these compounds to
                       the list of forbidden subgraphs (we don't want
                       to find occurrences containing the compounds
                       just substituted). */
		    Nonterminal nt = new Nonterminal("nt "+nonTerminalNumber,theMeasure);
		    nonterm.add(nt);

		    Subgraph sub1=substituteCompoundAndGetItsSubgraph(sf.piece, 
								      result.subgraph1,
								      nt,
								      theMeasure,
								      result.smResult,
								      null);
		    Subgraph sub2=substituteCompoundAndGetItsSubgraph(sf.piece, 
								      result.subgraph2,
								      nt,
								      theMeasure,
								      result.smResult,
								      null);
		    //System.out.println("sub1.size="+sub1.size());
		    //System.out.println("sub2.size="+sub2.size());
		    forbidden.add(sub1);
		    forbidden.add(sub2);
		    // sub1 and sub2 contain each only a single vertex
		    // -- the compound produced by the substitution
		    // This way of getting the compounds is an ugly hack
		    // TODO:improve in some way
		    CompoundVertex cv1=(CompoundVertex)sub1.vertices.get(0);
		    CompoundVertex cv2=(CompoundVertex)sub2.vertices.get(0);


		    /* BEGIN OCCURRENCE SEARCH */
		    for(int j=i;j<measures.length;j++){
			boolean occ1=true; // search for occ.s of subgraph 1
			boolean occ2=false; // search for occ.s of subgraph 2
			int countOccSearches=0;
			theMeasure=measures[j];
			while(occ1 || occ2){
			    countOccSearches++;
			    sleep();
			    
			    if(occ1){
				printHeader("OCCURRENCE SEARCH "+countOccSearches
					       +"[occ1] with measure="+theMeasure.name
					       +", "+nt.toString());
				occ1=occurrenceSearch(occGraph1,forbidden,nt,cv1.productionRule);
				if (!occ1) System.out.println("---Forgetting about occ1");
			    }
			    if(occ2){
				printHeader("OCCURRENCE SEARCH "+countOccSearches
					       +"[occ2] with measure="+theMeasure.name
					       +", "+nt.toString());
				occ2=occurrenceSearch(occGraph2,forbidden,nt,cv2.productionRule);
				if (!occ2) System.out.println("---Forgetting about occ2");
			    }
			}
		    }// end for (occ search)
		} // end if best is good enough
	    } // end while we should find a new theme
	} // end for all measures
    }
    
    /** Shows the population in the last generation. */
    public void show(){
	//new GraphWindow(mother,200,200,sf.seqPop());
	sf.showPopulation();
    }
    
    
    private SimFinder getSF(){
	SimFinder result=null;
	try{
	    result = new SimFinder(mother,popSize,initSize,
				   crossover,mutation,repFreq,
				   theMeasure,null/*occSearchGraph*/,maxSize,
				   freshBloodChance, groupBase);
	}catch(Exception e){
	    
	}
	if(result==null && initSize>1){
	    initSize--;
	    System.out.println("Trying with smaller initSize:"+initSize);
	    result=getSF();
	}
	return result;
    }
    
    private void runSF(){
	try{
	    if (generationsAfterLimitReached==-1) {
		// don't care about limit, run all generations anyway
		sf.run(simGens,file);
		return;
	    } else {
		int totalGens=0;
		while (totalGens+gensPerLimitCheck<simGens) {
		    sf.run(gensPerLimitCheck,file);
		    totalGens+=gensPerLimitCheck;

		    // check if limit reached
		    SimilarityStatement res=sf.findBest();
		    if(res.value()<limit && 
		       res.subgraph1.overlap(res.subgraph2)==0 &&
		       res.meanNestedSize()>meanNestedSizeLimit){
			System.out.println("best="+res.value());
			System.out.println("FOUND THEME!");
			// limit reached, run a little more and return
			sf.run(generationsAfterLimitReached,file);
			return;
		    } else {
		    // limit not reached. Carry on.
		    System.out.println("best="+res.value());
		}

		}
		if (totalGens<simGens) {
		    sf.run(simGens-totalGens,file);
		}
	    }

	}catch(Exception e){
	    //execution stopped: could make new SimStmts with size initSize
	    System.out.println(e.getMessage());
	    initSize--;
	    if(initSize>1){
		System.out.println("Trying with smaller initSize:"+initSize);
		sf=getSF();
		runSF();
	    }else{
		System.out.println("Run not possible");
		sf.showPopulation();
	    }
	}
    }

    private void runOccSF() throws Exception {
	if (generationsAfterLimitReached==-1) {
	    // don't care about limit, run all generations anyway
	    sf.run(occGens,file);
	    return;
	} else {
	    int totalGens=0;
	    while (totalGens+gensPerLimitCheck<occGens) {
		sf.run(gensPerLimitCheck,file);
		totalGens+=gensPerLimitCheck;
		
		// check if limit reached
		SimilarityStatement res=sf.findBest();
		if(res.value()<limit && 
		   res.subgraph1.overlap(res.subgraph2)==0 &&
		   res.meanNestedSize()>meanNestedSizeLimit){
		    // limit reached, run a little more and return
		    sf.run(generationsAfterLimitReached,file);
		    System.out.println("FOUND OCCURRENCE!");
		    return;
		} else {
		    // limit not reached. Carry on.
		    System.out.println("best="+res.value());
		}
	    }
	    if (totalGens<occGens) {
		sf.run(occGens-totalGens,file);
	    }
	}
    }

    void sleep(){
	try{Thread.sleep(1000);}catch(Exception e){}
    }

    private Subgraph prepareSubgraphForOccSearch(Subgraph s){
	Subgraph res=null;
	if(theMeasure instanceof SeqSimMeasure){
	    res=s.deepClone().sequentialClone();
	}else if(theMeasure instanceof NonSeqSimMeasure){
	    res=(Subgraph)s.deepClone().clone();
	}else{
	    System.out.println("SimSegmenter: unknown measure");
	    System.exit(0);
	}
	if(res==null){
	    
	    System.out.println("SimSegmenter.prepare...: this graphs clone was null:");
	    System.out.println(s.toString());
	}
	return res;
    }

    private boolean occurrenceSearch(Subgraph occGraph, Vector forbidden,
				     Nonterminal nonterm, GraphProductionRule originalMotifPR) {
	try{
	    if(occGraph==null){
		System.out.println("SimSegmenter.occurrenceSearch: occGraph==null");
		System.exit(0);
	    }

	    // don't use getSF() for this simfinder, we want to catch
	    // any exceptions from the simfinder here
	    sf = new SimFinder(sf.piece,popSize,initSize,
			       crossover,mutation,repFreq,
			       theMeasure,occGraph,maxSize,
			       freshBloodChance,groupBase);
	    sf.forbiddenSubgraphs=forbidden;
	    
	    runOccSF();
	    SimilarityStatement res = sf.findBest();

	    // does the best found statement contain an occurrence of
	    // the subgraph 'occGraph'
	    if(res.value()<limit && 
	       res.subgraph1.overlap(res.subgraph2)==0 &&
	       res.meanNestedSize()>2.0){
		// found an occurrence!! substitute it and add to forbidden-list
		forbidden.add(substituteCompoundAndGetItsSubgraph(sf.piece, 
								  res.subgraph2,
								  nonterm,
								  theMeasure,
								  res.smResult,
								  originalMotifPR));
		return true;
	    }else{
		return false;
	    }
	}catch(Exception e){ 
	    //could not initialise with the given length (occ.size())
	    return false;//skip Occurrence Search
	}
	//return false;
    }
    
    private Subgraph substituteCompoundAndGetItsSubgraph(MotherGraph piece,
							 Subgraph sub,
							 Nonterminal nt,
							 SimilarityMeasure sm,
							 SimMeasureResult simMeasResult,
							 GraphProductionRule originalMotifPR) {

	CompoundVertex cv= piece.compoundSubstitution(sub,nt,sm,simMeasResult);
	cv.productionRule.original=originalMotifPR;
	if (originalMotifPR!=null) {
	    originalMotifPR.derivations.add(cv.productionRule);
	}
	prodRules.add(cv.productionRule);
	
	Subgraph result=null;
	if (sm instanceof SeqSimMeasure) {
	    result=new SeqSubgraph("Forbidden subgraph ("+nt.toString()+")",piece);
	} else if(sm instanceof NonSeqSimMeasure){
	    result=new NonSeqSubgraph("Forbidden subgraph ("+nt.toString()+")",piece);
	} else{ 
	    System.out.println("SimSegmenter.substComp...: unknown measure"); 
	    System.exit(0);
	}
	result.vertices.add(cv);

	return result;
    }
    private void printHeader(String h) {
	System.out.println();
	System.out.println("########################################");
	System.out.println("# "+h);
	System.out.println("########################################");
	System.out.println();
    }

    public void showAllProductionRules(Vector pr) {
	/*	ProdRuleWindow[] ruleWindows = new ProdRuleWindow[pr.size()];
	printHeader("Showing all production rules:");
	for (int i=0; i<pr.size(); i++) {
	    GraphProductionRule gpr = (GraphProductionRule) pr.get(i);
	    System.out.println("\t"+gpr.toString());
	    ruleWindows[i]=new ProdRuleWindow(gpr,50+i*25,300+i*35,sf.seqPop());
	    }*/
	ProdRuleList prl= new ProdRuleList(pr,200,200);
    }
    
}
