/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Random;


/** A SimilarityStatement is a tentative assertion that two {@link
Subgraph}s are similar according to some {@link SimilarityMeasure}. It
can thus be evaluated using the SimilarityMeasure connected to it, and
when evaluated, a SimilarityStatement thus has a {@link
SimMeasureResult} telling how good a SimilarityStatement it
is. SimilarityStatement is an abstract class, because it must be
either sequential or non-sequential.  */
public abstract class SimilarityStatement {

    SimFinder myOwner=null;
    
    /* class representing statements about the similarity of two subgraphs */
    static Random rand = new Random();

    /** One subgraph. */
    public Subgraph subgraph1;
    /** The other subgraph. */
    public Subgraph subgraph2;
    /** The SimilarityMeasure used to evaluate this SimilarityStatement. */
    protected SimilarityMeasure simMeasure=null;
    /** The result of evaluating this SimilarityStatement using the
        attached SimilarityMeasure. */
    public SimMeasureResult smResult = null;

    /** The amount of overlap between subgraph1 and subgraph2. */
    public int olap=0;

    /*
    public SimilarityStatement(Subgraph s1, Subgraph s2) {
	subgraph1 = s1;
	subgraph2 = s2;
	}*/

    public SimilarityStatement(Subgraph s1, Subgraph s2, SimFinder sf) {
	subgraph1 = s1;
	subgraph2 = s2;
	myOwner=sf;
    }

    // public SimilarityStatement() {    }

    /** Returns the size of the compared subgraphs.*/
    public int size() {
	return subgraph1.size();
    }
    
    /** Returns the maximal nested size of the two {@link Subgraph}s. */
    public int maxNestedSize() {
	return Math.max(subgraph1.nestedSize(),subgraph2.nestedSize());
    }

    /** Returns the mean nested size of the two {@link Subgraph}s. */
    public double meanNestedSize(){
	return ((double)subgraph1.nestedSize()+(double)subgraph2.nestedSize())/2.0;
    }

    /** Returns the mean size of the two subgraphs, not counting rests. */
    public int sizeNoRests(){
	int rests1=subgraph1.nestedNumberOfRests();
	int rests2=subgraph2.nestedNumberOfRests();
	return (subgraph1.size()*2-rests1-rests2)/2;
    }

    /* ved brug af denne sizer fr simStmt'et ikke gavn af at indlemme
       ekstra pauser i enderne!  
       ikke compound safe! */
    /** @deprecated Not used anymore. */
    public int sizeNoOutmostRests(){
	int rests1l=subgraph1.numberOfLeftmostRests();//ikke compound safe!
	int rests1r=subgraph1.numberOfRightmostRests();
	int size1=subgraph1.size()-rests1l-rests1r;

	int rests2l=subgraph2.numberOfLeftmostRests();
	int rests2r=subgraph2.numberOfRightmostRests();
	int size2=subgraph2.size()-rests2l-rests2r;
	
	return (size1+size2)/2;//NB heltalsdivision
    }

    public SimilarityMeasure getSimilarityMeasure() {
	return simMeasure;
    }

    public void  setSimilarityMeasure(SimilarityMeasure sm) {
	simMeasure=sm;
    }

    /** For debugging. */
    public boolean oksize() {
	int s1=subgraph1.vertices.size();
	int s2=subgraph2.vertices.size();
	if (s1!=s2) {
	    return false;
	}
	return true;
    }

    /** If this SimilarityStatement has not been evaluated (or has
        been changed, e.g. mutated, since the last evaluation, it is
        re-evaluated by a call to the relevant SimilarityMeasure. The
        value is returned. */
    public double value() {
	//double overlapPenaltyBase=1.6;
	if (smResult==null) {
	    simMeasure.evaluate(this);
	    smResult.measureNoOverlap=smResult.measure;
	    if(overlap()>0){
		smResult.measure=1000000;
	    }
	    //smResult.measure=smResult.measure*Math.pow(overlapPenaltyBase,overlap());
	}
	return smResult.measure;
    }
    
    public String toString() {
	value(); // to ensure this SimStmt has been evaluated 
	String result = "SimStmt of\n";
	result += "\tSize = "+size()+"\n";
	result += "\tOverlap="+overlap()+"\n";
	result += smResult.toStringWithViewDifferences();
	return result;
    }

    /** Includes toStrings of the two compared subgraphs. */
    public String toStringWithSubgraphs() {
	// String result = toString();
	// debug
	String result = toString();

	result += "Subgraph1: \n"+subgraph1.toString()+"\n";
	result += "Subgraph2: \n"+subgraph2.toString()+"\n";
	return result;
    }

    public String gnuplotHeader() {
	String result = "";
	for (int i=0; i<simMeasure.viewComparators.size(); i++) {
	    ViewComparator vp = (ViewComparator) simMeasure.viewComparators.get(i);
	    result+= vp.name+" ";
	}
	result += simMeasure.name;
	return result;
    }

    public String toGnuplotString() {
	// dummy is used to call value() because otherwise simMeasure
	// might be null
	double dummy = value();
	String result = "";
	for (int i=0; i< smResult.viewDifferences.size(); i++) {
	    ViewDifference vp = (ViewDifference) smResult.viewDifferences.get(i);
	    result+= vp.view+" ";
	}
	result += dummy+" ";
	return result;
    }

    /** Constructs a one-line description of this SimilarityStatement
        which is useful in GUI lists. */
    public String toListString() {
	String result = "";
	result += "Size="+size();
	result += " Overlap="+overlap();
	result += " Value="+value();
	return result;
    }

    /** Mutation is defined separately for sequential and
        non-sequential similarity statements (see {@link SeqSimStmt}
        and {@link NonSeqSimStmt}). */
    public abstract void mutate(double mutrate) throws Exception;

    public abstract Object clone();/* {
	SimilarityStatement ss = new SimilarityStatement((MusicGraph)subgraph1.clone(), (MusicGraph)subgraph2.clone());
	ss.absolutePitch=absolutePitch;
	ss.pitchContour = pitchContour;
	ss.pitchInterval = pitchInterval;
	ss.absoluteLength = absoluteLength;

	ss.valueAssigned=true;
	ss.olap=overlap();
	return ss;
	}*/


    /** Returns true if subgraph1 equals subgraph2 (i.e. they have
        maximal possible overlap). */
    public boolean identicalSubgraphs() {
	return subgraph1.equals(subgraph2);
    }

    /** Tells whether or not this SimilarityStatement has been
        evaluated and assigned a value. */
    public boolean valueAssigned() { return smResult!=null; }

    /** Returns how many vertices and edges the two subgraphs have in
        common. */
    public int overlap() {
	olap=subgraph1.overlap(subgraph2);
	int moreOverlap=0;
	if(myOwner!=null){
	    for(int fgr=0;fgr<myOwner.forbiddenSubgraphs.size();fgr++){
		Subgraph sg=(Subgraph)myOwner.forbiddenSubgraphs.get(fgr);
		if(!myOwner.occurrenceSearch()){
		    moreOverlap+=subgraph1.overlap(sg);
		}
		moreOverlap+=subgraph2.overlap(sg);
	    }
	    olap+=moreOverlap;
	}else{
	    /* This is normal when we construct a simstmt in the
	       SimLibs to use another simMeasure inside a
	       viewcomparator. */
	    
	    // System.out.println("SimilarityStatement:myOwner is null");
	    //System.out.println("size="+size());
	    //System.exit(0);
	}
	return olap;
    }
}
