/* Copyright (C) 2003 Sren Tjagvad Madsen & Martin Elmer Jrgensen
 * email: {tjagvad,elmer}@daimi.au.dk */

import java.util.Vector;
import jm.music.data.*;
import jm.util.*;

/** A Subgraph is basically a subset of vertices and a subset of edges
    of some {@link MotherGraph}. The Subgraph class is abstract
    because in the SimFinder system, a subgraph is either sequential
    ({@link SeqSubgraph}) or non-sequential ({@link
    NonSeqSubgraph}). The two kinds do have some common properties and
    methods, though, and these are described in this abstract Subgraph
    class.*/

public abstract class Subgraph extends MusicGraph {

    /** The MotherGraph that owns the vertices and edges pointed to by
        this Subgraph. */
    protected MotherGraph mothergraph=null;

    /* Subgraph SAFE */
    /** Returns true if this Subgraph is sequential with only strong
	 FOLLOW edges. The method to check this: 1) this subgraph can
	 have only one starting vertex. 2) all descendants(vertices)
	 must have exactly one descendant, 3) except the last vertex,
	 which must have exactly zero descendants. */
    public boolean isStrongSequential() {
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    if (me.isSimultaneousEdge()) {
		// System.out.println("simEdge");
		return false;
	    }
	    if (me.isWeak()) {
		// System.out.println("weakEdge");
		return false;
	    }
	}

	sortVerticesByStarttime();

	int[] ins = new int[vertices.size()];
	int[] outs = new int[vertices.size()];

	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    for (int j=0; j<mv.inEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.inEdges.get(j);
		if (me.isFollowEdge() && edges.contains(me)) {
		    ins[i]++;
		}
	    }
	    for (int j=0; j<mv.outEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.outEdges.get(j);
		if (me.isFollowEdge() && edges.contains(me)) {
		    outs[i]++;
		}
	    }
	}
	if (ins[0]!=0) {
	    // System.out.println("ins[0]!=0");
	    return false;
	}
	if (outs[outs.length-1]!=0) {
	    // System.out.println("outs[last]!=0");
	    return false;
	}
	for (int i=1; i<ins.length-1; i++) {
	    if (ins[i]!=1 || outs[i]!=1) {
		// System.out.println("multi edge in middle");
		return false;
	    }
	}
	return true;

	/*	Vector sources = getStrongSources();
	if (sources.size()!=1) return false;
	MusicVertex mv = (MusicVertex) sources.get(0);

	for (int i=1; i<vertices.size(); i++) {
	    // find out how many of mv's outEdges are in this graph
	    int outs=0;
	    MusicVertex next = null;
	    for (int j =0; j<mv.outEdges.size(); j++) {
		MusicEdge me= (MusicEdge) mv.outEdges.get(j);
		if (me.isStrongFollowEdge()) {
		    if (this.edges.contains(me)) {
			outs++;
			next=me.to;
		    }
		}
		
	    }
	    // all descendants must have exactly one descendant
	    if (i<vertices.size()-1) {
		if (outs!=1) return false;
	    }
	    // except the last one, must have no descendants
	    else if (i==vertices.size()-1) { 
		if (outs!=0) return false;
	    }
	    mv=next;
	    }
	    return true;
	*/
    }

    public String toString() {
	String result = super.toString();
	return result;
    }

    /** Returns how many times this Subgraph may be extended 
        to the left. */
    public abstract int getLeftExtensibility();
    /** Returns how many times this Subgraph may be extended 
        to the right. */
    public abstract int getRightExtensibility();

    /** Extends this subgraph by randomly adding to it a {@link
        MusicVertex} which is connected to this subgraph in its mother
        graph. The edge connecting them is also added. */
    public abstract void extend(int times) throws Exception;

    /** Shortens this subgraph by randomly choosing and removing a
        {@link MusicVertex} and the edge(s) that connected it to the
        rest of this subgraph. Only vertices that do not split this
        graph may be removed. */
    public abstract void shorten(int times) throws Exception;

    /** Slides this Subgraph either left or right a number of times. */
    public abstract void slide(int times) throws Exception;

    /** Returns a Vector of {@link MusicVertex} which follow upon any
        of the music vertices in the input vector, i.e. there is a
        FOLLOWS {@link MusicEdge} from the vertex in the input vector
        to the one in the output vector.  */
    public Vector getAllSons(Vector vert /* of MusicVertex */) {
	Vector result = new Vector();
	for (int i=0; i<vert.size(); i++) {
	    MusicVertex mv = (MusicVertex) vert.get(i);
	    for (int j=0; j<mv.outEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.outEdges.get(j);
		if (me.isStrongFollowEdge() && edges.contains(me)) {
		    if (!result.contains(me.to))
			result.add(me.to);
		}
	    }
	}
	return result;
    }

    /** Returns a Vector of Vector of {@link MusicEdge}; the Vector of
        MusicEdges at index i in the returned Vector contains the
        proper follow in-edges of MusicVertex i in this Subgraph;
        proper follow in-edges are follow in-edges to the given vertex
        that are also contained in this Subgraph. */
    public Vector getAllProperFollowInEdges(boolean includeWeakEdges) {
	Vector inEdgeVectors = new Vector();

	/* find subgraph 1's edges */	
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    Vector ownedInEdges = new Vector();
	    for (int j=0; j<mv.inEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.inEdges.get(j);
		if (me.isFollowEdge() && edges.contains(me)) {
		    if (!includeWeakEdges && me.isStrongFollowEdge()) {
			ownedInEdges.add(me);
		    } else if (includeWeakEdges) {
			ownedInEdges.add(me);
		    } 
		}
	    }
	    inEdgeVectors.add(ownedInEdges);
	}
	return inEdgeVectors;
    } 
    /** Returns a Vector of Vector of {@link MusicEdge}; the Vector of
        MusicEdges at index i in the returned Vector contains the
        proper follow out-edges of MusicVertex i in this Subgraph;
        proper follow out-edges are follow out-edges to the given
        vertex that are also contained in this Subgraph. */
    public Vector getAllProperFollowOutEdges(boolean includeWeakEdges) {
	Vector outEdgeVectors = new Vector();

	/* find subgraph 1's edges */	
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    Vector ownedOutEdges = new Vector();
	    for (int j=0; j<mv.outEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.outEdges.get(j);
		if (me.isFollowEdge() && edges.contains(me)) {
		    if (!includeWeakEdges && me.isStrongFollowEdge()) {
			ownedOutEdges.add(me);
		    } else if (includeWeakEdges) {
			ownedOutEdges.add(me);
		    }
		}
	    }
	    outEdgeVectors.add(ownedOutEdges);
	}
	return outEdgeVectors;
    } 

    /** Returns a Vector of Vector of {@link MusicEdge}; the Vector of
        MusicEdges at index i in the returned Vector contains the
        proper in-edges of MusicVertex i in this Subgraph; proper
        in-edges are in-edges to the given vertex that are also
        contained in this Subgraph. */
    public Vector getAllProperInEdges(boolean includeWeakEdges) {
	Vector inEdgeVectors = new Vector();
	/* find subgraph 1's edges */	
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    Vector ownedInEdges = new Vector();
	    for (int j=0; j<mv.inEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.inEdges.get(j);
		if (edges.contains(me)) {
		    if (includeWeakEdges) {
			ownedInEdges.add(me);
		    }
		    else if (!includeWeakEdges && !me.isWeak()) {
			ownedInEdges.add(me);
		    }
		}
	    }
	    inEdgeVectors.add(ownedInEdges);
	}
	return inEdgeVectors;
    } 
    /** Returns a Vector of Vector of {@link MusicEdge}; the Vector of
        MusicEdges at index i in the returned Vector contains the
        proper out-edges of MusicVertex i in this Subgraph; proper
        out-edges are out-edges to the given vertex that are also
        contained in this Subgraph. */
    public Vector getAllProperOutEdges(boolean includeWeakEdges) {
	Vector outEdgeVectors = new Vector();

	/* find subgraph 1's edges */	
	for (int i=0; i<vertices.size(); i++) {
	    MusicVertex mv = (MusicVertex) vertices.get(i);
	    Vector ownedOutEdges = new Vector();
	    for (int j=0; j<mv.outEdges.size(); j++) {
		MusicEdge me = (MusicEdge) mv.outEdges.get(j);
		if (edges.contains(me)) {
		    if (includeWeakEdges) {
			ownedOutEdges.add(me);
		    }
		    else if (!includeWeakEdges && !me.isWeak()) {
			ownedOutEdges.add(me);
		    }
		}
	    }
	    outEdgeVectors.add(ownedOutEdges);
	}
	return outEdgeVectors;
    } 

    /** Returns the Score object that the MotherGraph of this Subgraph
        originated from. */
    public Score getScore() {
	if(mothergraph==null) {
	    return null;
	}else{
	    return mothergraph.getScore();
	}
    }

    /** Returns the MotherGraph of this Subgraph. */
    public MotherGraph getMotherGraph() { return mothergraph; }

    /** For debugging. (converts this subgraph to a Score and writes
        it to a midi file) */
    public void printTheScore(String label){
	Score s=convertToScore();
	s.setTempo(100);
	Write.midi(s,label);
	
    }

    /** Returns a Vector of all strong FOLLOW edges in this Subgraph. */
    public Vector getStrongFollowEdges() {
	Vector result= new Vector();
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    if (me.isStrongFollowEdge()) result.add(me);
	}
	return result;
    }
    /** Returns a Vector of all FOLLOW edges in this Subgraph. */
    public Vector getFollowEdges() {
	Vector result= new Vector();
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    if (me.isFollowEdge()) result.add(me);
	}
	return result;
    }
    /** Returns a Vector of all strong SIMULTANEOUS edges in this Subgraph. */
    public Vector getStrongSimultaneousEdges() {
	Vector result= new Vector();
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    if (me.isSimultaneousEdge()) result.add(me);
	}
	return result;
    }
    /** Returns a Vector of all SIMULTANEOUS edges in this Subgraph. */
    public Vector getSimultaneousEdges() {
	Vector result= new Vector();
	for (int i=0; i<edges.size(); i++) {
	    MusicEdge me = (MusicEdge) edges.get(i);
	    if (me.isStrongSimultaneousEdge()) result.add(me);
	}
	return result;
    }

    
   public abstract double getMothergraphCoverage();

    //public abstract void roundGraph();
    public abstract int numberOfLeftmostRests();
    public abstract int numberOfRightmostRests();

    /*
    public Vector splitVerticesIntoParts(){
	return null;
    }
        

    public MusicVertex findNextMusicVertexInSamePart(MusicVertex mv){
	return null;
    }
    */

    /** Removes a given {@link MusicVertex} mv from this Subgraph
        along with any MusicEdges that connected mv to the rest of
        this Subgraph. */
    public void removeVertexAndConnectedEdges(MusicVertex mv) {
	for (int i=0; i<mv.inEdges.size(); i++) {
	    edges.remove((MusicEdge) mv.inEdges.get(i));
	}
	for (int i=0; i<mv.outEdges.size(); i++) {
	    edges.remove((MusicEdge) mv.outEdges.get(i));
	}
	vertices.remove(mv);
    }

    /** Adds a given {@link MusicVertex} mv and a {@link MusicEdge} me
        to this Subgraph along with any MusicEdges in the MotherGraph
        that connect mv to the rest of this Subgraph. */
    public void addVertexAndNecessaryEdges(MusicVertex mv, MusicEdge me) {
	vertices.add(mv);
	edges.add(me);
	/* Ensure there are no edges in mothergraph between mv and
         * some other vertex in this subgraph that are not included in
         * this subgraph*/
	for (int i=0; i<mv.inEdges.size(); i++) {
	    MusicEdge med = (MusicEdge) mv.inEdges.get(i);
	    if (vertices.contains(med.from)) {
		if (!edges.contains(med)) {
		    edges.add(med);
		}
	    }
	}
	for (int i=0; i<mv.outEdges.size(); i++) {
	    MusicEdge med = (MusicEdge) mv.outEdges.get(i);
	    if (vertices.contains(med.to)) {
		if (!edges.contains(med)) {
		    edges.add(med);
		}
	    }
	}
	sortVerticesByStarttime();
    }

    /** If mv is outside this subgraph, and there is an edge e in the
       mothergraph connecting mv to some vertex in this subgraph, this
       method returns e, otherwise null. */
    public MusicEdge findEdgeConnectingToVertex(MusicVertex mv) {
	if (vertices.contains(mv)) return null;
	
	for (int j=0; j<mv.inEdges.size(); j++) {
	    MusicEdge me = (MusicEdge) mv.inEdges.get(j);
	    if (vertices.contains(me.from)) return me;
	    if (vertices.contains(me.to)) return me;
	}
	for (int j=0; j<mv.outEdges.size(); j++) {
	    MusicEdge me = (MusicEdge) mv.outEdges.get(j);
	    if (vertices.contains(me.from)) return me;
	    if (vertices.contains(me.to)) return me;
	}
    	return null;
    }
}
